// Opcode implementations. 
// Complex opcodes might be implemented in other files
// (e.g., destructure and restructure in grammar.c),
// but instruction decoding should still be done here.

#include "grammar.h"
#include "code.h"

DEFINE_bool(lotsofcopies, false, "Cause additional copying of ASTs (debugging).");
DECLARE_bool(disassemble);
DECLARE_bool(profile);
DECLARE_int32(parsedebug);

static void xxprint(Type *t, void *v);
static void xxprint_fields(Type *t, void **v, int *first);
static ListNode* copylistnode(ListNode *src, ListNode *next, Type *t, int n);

Fn*
currentfn(Thread *t)
{
	if(t->fp_ == nil)
		return nil;
	return ((Fn**)t->fp_)[FrameOffsetFN/sizeof(void*)];
}

// Decode a Loc (%l) from the instruction stream
// and return the pointer that it describes.
void*
Thread::loc()
{
	int i, d, n;
	char *p;
	
	// assert(((int)pc_.i & 3) == 0);
	i = *pc_.i++;
	if(i == ~0)
		return nil;
	p = nil;
	switch((i&7)>>1){
	default:
	case LocFrame>>1:
		p = (char*)fp_;
		d = i>>3;
		break;
	case LocGlobal>>1:
		p = (char*)gp_;
		d = i>>3;
		break;
	case LocEscape>>1:
		n = (i>>3)&0xFF;
		p = (char*)fp_ + FrameOffsetEP;
		while(n-- >= 0)
			p = *(char**)p;
		d = i>>11;
	}
	p += d;
	if(i&LocIndirect){
		p = *(char**)p;
		p += *pc_.i++;
	}

	return p;
}

// Return the pointer for the given var.
void*
var2ptr(Thread *t, Var *v)
{
	void *base;

	base = nil;
	if(v == nil)
		panic("var2ptr");
	if(v->f == globalframe){
		base = t->gp_;
	}else if(v->escapes == 0){
		base = t->fp_;
	}else{
		panic("var2ptr escapes - not implemented");
	}
	
	return (char*)base + v->off;
}

// Type-preserving binary operators.
#define binoptype(name, type, ctype, op) \
	void \
	name ## type(Thread *t) \
	{ \
		ctype *dst, *src1, *src2; \
		dst = static_cast<ctype*>(t->loc()); \
		src1 = static_cast<ctype*>(t->loc()); \
		src2 = static_cast<ctype*>(t->loc()); \
		*dst = *src1 op *src2; \
	}

#define binop(name, op) \
	binoptype(name, i, int, op) \
	binoptype(name, f, double, op)

#define binopi(name, op) \
	binoptype(name, i, int, op)

binop(xadd, +)
binop(xsub, -)
binop(xmul, *)
binop(xdiv, /)
binopi(xmod, %)
binopi(xlsh, <<)
binopi(xrsh, >>)
binopi(xor, |)
binopi(xand, &)
binopi(xxor, ^)

// Shouldn't be special, but is, thanks to C.
void
xmodf(Thread *t)
{
	double *dst, *src1, *src2;

	dst = static_cast<double*>(t->loc());
	src1 = static_cast<double*>(t->loc());
	src2 = static_cast<double*>(t->loc());
	*dst = fmod(*src1, *src2);
}

// Type-preserving unary operators.
#define unoptype(name, type, ctype, op) \
	void \
	name ## type(Thread *t) \
	{ \
		ctype *dst, *src; \
		dst = static_cast<ctype*>(t->loc()); \
		src = static_cast<ctype*>(t->loc()); \
		*dst = op *src; \
	}

#define unop(name, op) \
	unoptype(name, i, int, op) \
	unoptype(name, f, double, op)

#define unopi(name, op) unoptype(name, i, int, op)

unop(xneg, -)
unop(xpos, +)
unopi(xtwid, ~)
unopi(xnot, !)

// Comparisons
#define cmpoptype(name, type, ctype, op) \
	void \
	name ## type(Thread *t) \
	{ \
		int *dst; \
		ctype *src1, *src2; \
		dst = static_cast<int*>(t->loc()); \
		src1 = static_cast<ctype*>(t->loc()); \
		src2 = static_cast<ctype*>(t->loc()); \
		*dst = *src1 op *src2; \
	}

#define cmpop(name, op) \
	cmpoptype(name, i, int, op) \
	cmpoptype(name, f, double, op)

#define cmpopi(name, op) cmpoptype(name, i, int, op)

cmpop(xlt, <)
cmpop(xgt, >)
cmpop(xle, <=)
cmpop(xge, >=)
cmpop(xeq, ==)
cmpop(xne, !=)

// String comparisons
#define cmpopstring(name, op) \
	void \
	name(Thread *t) \
	{ \
		int cmp, *dst; \
		Name *src1, *src2; \
		dst = static_cast<int*>(t->loc()); \
		src1 = static_cast<Name*>(t->loc()); \
		src2 = static_cast<Name*>(t->loc()); \
		if(src1 == src2) \
			cmp = 0; \
		else \
			cmp = strcmp(*src1,*src2); \
		*dst = cmp op 0; \
	}

cmpopstring(xlts, <)
cmpopstring(xgts, >)
cmpopstring(xles, <=)
cmpopstring(xges, >=)

// Load immediate
#define loadoptype(type, ctype) \
	void \
	xload ## type(Thread *t) \
	{ \
		ctype *pp; \
		pp = static_cast<ctype*>(t->loc()); \
		*pp = *t->pc_.type++; \
	}

loadoptype(i, int)
loadoptype(f, double)
loadoptype(p, void*)

// One of a kind

void
xmkframe(Thread *t)
{
	Fn *fn;
	char **fpp;

	fpp = (char**)t->loc();
	fn = *(Fn**)t->loc();
	*fpp = (char*)mkfnframe(fn);
}

// If you change this, change run.c:/^callfn too.
void
xcall(Thread *t)
{
	Fn *fn;
	void **fp;
	
	fn = *(Fn**)t->loc();
	fp = *(void***)t->loc();

	profilecall(currentfn(t), fn->link);

	if(fn->builtin){
		fn->builtin((char*)fp - fn->frame->argsize);
		return;
	}

	if(--t->depth_ < 0)
		panic("call depth - %s", fn->name);
	fp[0] = t->pc_.c;
	fp[1] = t->fp_;
	t->pc_.c = (Code**)fn->code->bp;
	t->fp_ = fp;
	if(t->upcall_)
		t->upcall_++;
}

void
xret(Thread *t)
{
	void **fp;

	t->depth_++;
	fp = (void**)t->fp_;
	t->pc_.c = (Code**)fp[0];
	t->fp_ = fp[1];
	freefnframe(fp);
	if(t->upcall_ && --t->upcall_ == 0)
		longjmp(t->stop_, 1);
}

// Tail call: equivalent to xcall xret, less stack usage.
void
xretcall(Thread *t)
{
	Fn *fn;
	void **fp;
	void **oldfp;

	fn = *(Fn**)t->loc();
	fp = *(void***)t->loc();

	if(FLAGS_profile)
		profilecall(currentfn(t), fn->link);

	oldfp = (void**)t->fp_;
	fp[-1] = oldfp[-1];	/* return value pointer */
	fp[0] = oldfp[0];	/* return pc */
	fp[1] = oldfp[1];	/* return fp */
	t->fp_ = fp;
	freefnframe(oldfp);

	if(fn->builtin){
		fn->builtin((char*)fp - fn->frame->argsize);
		xret(t);
		return;
	}

	t->pc_.c = (Code**)fn->code->bp;
}

void
xjmp(Thread *t)
{
	int d;

	d = *t->pc_.i++;
	t->pc_.b += d;
}

void
xcjmp(Thread *t)
{
	int c, d;

	c = *(int*)t->loc();
	d = *t->pc_.i++;
	if(c)
		t->pc_.b += d;
}

void
xncjmp(Thread *t)
{
	int c, d;

	c = *(int*)t->loc();
	d = *t->pc_.i++;
	if(!c)
		t->pc_.b += d;
}

void
xstop(Thread *t)
{
	longjmp(t->stop_, 1);
}


// Array allocator (separate because it needs an array count).
Array*
newarray(int n, int sizeof1)
{
	Array *a;

	a = (Array*)galloc(sizeof *a + n*sizeof1);
	a->n = n;
	a->sizeof1 = sizeof1;
	a->v = a+1;
	return a;
}

void
xnewa(Thread *t)
{
	Array **pa;
	int n, m;
	
	pa = (Array**)t->loc();
	n = *(int*)t->loc();
	m = sizeoftype((Type*)*t->pc_.p++);
	*pa = newarray(n, m);
}

// Allocaters
static Map*
_xnewm(Type *type)
{
	Map *m;
	
	assert(type->op == TypeMap);
	m = (Map*)galloc(sizeof *m);
	m->h = mkhash<void*,void*>();
	m->sizeof1 = sizeoftype(type->left);
	return m;
}
static ListNode*
_xnewl(Type *type)
{
	ListNode *l;
	
	assert(type->op == TypeList);
	l = (ListNode*)galloc(sizeof(void*) + sizeoftype(type->left));
	return l;
}

void*
newstruct(Type *type)
{
	void *v, **p;

	assert(type && type->op == TypeStruct);
	v = (void*)galloc(sizeoftype(type));
	// Tag type pointer
	if(type->super){
		p = (void**)((char*)v + tagtypeoffset(type));
		*p = type;
	}
	return v;
}

void
xnew(Thread *t)
{
	void **vp;
	Type *type;
	
	vp = (void**)t->loc();
	type = (Type*)*t->pc_.p++;

	if(type->op == TypeArray)
		panic("xnew Array");
	else if(type->op == TypeList)
		*vp = _xnewl(type);
	else if(type->op == TypeMap)
		*vp = _xnewm(type);
	else if(type->op == TypeStruct)
		*vp = newstruct(type);
	else
		*vp = (void*)galloc(sizeoftype(type));
}

void
xnewcopy(Thread *t)
{
	void **dst, *src;
	Type *type, *tag;
	
	dst = (void**)t->loc();
	src = *(void**)t->loc();
	type = (Type*)*t->pc_.p++;

	assert(type->op == TypeStruct);
	if(type->super){
		tag = *(Type**)((char*)src + tagtypeoffset(type));
		*dst = newstruct(tag);
		copystruct(*dst, src, tag, tag);
	}else{
		*dst = newstruct(type);
		copystruct(*dst, src, type, type);
	}
}

void
xalloc(Thread *t)
{
	void **vp;
	int n;
	
	vp = (void**)t->loc();
	n = *t->pc_.i++;
	*vp = (void*)galloc(n);
}

// Map iterators
void
xnewmapiter(Thread *t)
{
	MapIter *i;
	Map *m;

	i = (MapIter*)t->loc();
	m = *(Map**)t->loc();

	if(m == nil)
		memset(i, 0, sizeof(i) + m->sizeof1);
	else{
		memset(i, 0, sizeof(i) + m->sizeof1);
		i->sizeof1 = m->sizeof1;
		hashiterstart(&i->iter, m->h);
	}
}

void
xmapiternext(Thread *t)
{
	int *dst;
	MapIter *i;

	dst = (int*)t->loc();
	i = (MapIter*)t->loc();

	*dst = hashiternextkv(&i->iter, &i->kv);
	if(*dst){
		i->k = i->kv.key;
		memmove(i+1, i->kv.value, i->sizeof1);
	}
}

void
xmapitertuple(Thread *t)
{
	void *dst;
	MapIter *i;

	dst = (void*)t->loc();
	i = (MapIter*)t->loc();
	if(i->iter.h)
		memmove(dst, &i->k, sizeof(void*) + i->sizeof1);
}


// Printing
static void
xxprint_fields(Type *t, void **v, int *first)
{
	ExtStruct *le;
	char *p;
	char *base;
	
	assert(t->op == TypeStruct);
	base = (char*)*v;

	// Super fields
	if(t->super){
		xxprint_fields(t->super, v, first);
		p = (char*)*v;
	}else{
		p = base + structdataoffset(t);
	}
	// Self fields
	for(List<Var*> *lv=t->vars; lv; lv=lv->next){
		Var *fv = lv->x;
		Type *ft = fv->type;
		if(fv->attrfn || fv->isbuiltinattr)
			continue;
		if(!*first)
			print(", ");
		*first = 0;
		print("%s=", fv->name);
		xxprint(ft, p);
		p += sizeoftype(ft);
	}
	*v = p;
	// Extension fields
	if(t->extensible){
		p = base + extptroffset(t);
		le = *(ExtStruct**)p;
		for(; le; le=le->next){
			assert(le->type && le->type->extends);
			if(le->type->extends == t){
				void *ev = le;
				xxprint_fields(le->type, &ev, first);
			}
		}
	}
}

static void
xxprint(Type *t, void *v)
{
	char *p;
	void *vv;
	int i, first;
	Array *a;
	ListNode *ln;
	Type *base;

	switch(t->op){
	case TypeBool:
		print("%s", *(int*)v ? "true" : "false");
		break;
	case TypeInt:
		print("%d", *(int*)v);
		break;
	case TypeFloat:
		print("%f", *(double*)v);
		break;
	case TypePtr:
		vv = *(void**)v;
		if(vv == nil) {
			print("?[%T / nil]", t);
		} else if(isptrast(t)) {
			print("?[%T / %p]", t, vv);
		} else {
			print("&");
			xxprint(t->left, vv);
		}
		break;
	case TypeTuple:
		p = (char*)v;
		print("(");
		for(List<Type*> *l=t->types; l; l=l->next){
			if(l!=t->types)
				print(", ");
			xxprint(l->x, p);
			p += sizeoftype(l->x);
		}
		print(")");
		break;
	case TypeString:
		p = *(char**)v;
		if(p)
			print("\"%s\"", p);
		else
			print("<nil>");
		break;
	case TypeStruct:
		vv = v;
		first = 1;
		base = rootstruct(t);
		print("%s", base->name);
		if(base->discriminated){
			Type **ptag = (Type**)((char*)v+tagtypeoffset(base));
			Type *tag = *ptag;
			if(tag)
				print(".%s", tag->name);
			else
				print(".<nil>");
			print("{");
			xxprint_fields(tag, &vv, &first);
			print("}");
		}
		else{
			print("{");
			xxprint_fields(t, &vv, &first);
			print("}");
		}
		break;
	case TypeArray:
		a = *(Array**)v;
		print("array [");
		if(a){
			p = (char*)a->v;
			for(i=0; i<a->n; i++){
				if(i > 0)
					print(", ");
				xxprint(t->left, p);
				p += a->sizeof1;
			}
		}
		print("]");
		break;
	case TypeList:
		ln = *(ListNode**)v;
		print("list [");
		for(; ln; ln=ln->next){
			xxprint(t->left, ln->v);
			if(ln->next)
				print(", ");
		}
		print("]");
		break;
	case TypeSubst:
		xxprint(t->left, v);
		break;
	default:
		print("?[%T / %p]", t, *(char**)v);
		break;
	}
}

void
xprint(Thread *t)
{
	Type *type;
	void *v;

	type = (Type*)*t->pc_.p++;
	v = (void*)t->loc();
	xxprint(type, v);
	print("\n");
}

int linkfncb;
int linkfncode;
int linkfncalls;
void
linkfnstats(void)
{
	fprint(2, "linkfn: %d buffers, %d code, %d calls\n", linkfncb, linkfncode, linkfncalls);
}

extern void registertrace(Fn*);
void
xlinkfn(Thread *t)
{
	Fn **dst;
	void *ep;
	Fn *fn;
	Fn *newfn;
	Loc eploc;
static int first;
if(first == 0){
	first = 1;
//	atexit(linkfnstats);
}

	dst = (Fn**)t->loc();
	ep = *(void**)t->loc();
	fn = (Fn*)*t->pc_.p++;
	memset(&eploc, 0, sizeof eploc);
	eploc.mode = LocFrame;
	eploc.off = FrameOffsetEP;

	Typecheck tc((Frame*)~0, (Scope*)~0);
	newfn = mkfn(fn->line, fn->name, fn->fnargs, fn->rettype, fn->stmts);
	newfn->type = fn->type;
	newfn->outer = fn->outer;
	newfn->link = fn->link;
	newfn->frame = fn->frame;

	assert(newfn->link->link == newfn->link);

	newfn->code = mkcodebuf();
	Compile *cs = new Compile(&tc);
	newfn->cs = cs;
	cs->codebuf_ = newfn->code;
	cs->emitloadptr(fn->line, &eploc, ep);
	eploc.off = FrameOffsetFN;
	cs->emitloadptr(fn->line, &eploc, fn);
	cs->emitC(fn->line, xjmp);
	// XXX what if it reallocs right here?
	cs->emiti(fn->code->bp - newfn->code->p - 4);
	delete cs;
	cs = nil;
	
	profileclosure(currentfn(t), newfn);

	linkfncb += newfn->code->ep - newfn->code->bp;
	linkfncode += newfn->code->p - newfn->code->bp;
	linkfncalls++;

	if(FLAGS_disassemble){
		fprint(2, "--linkfn %p\n", newfn);
		disasm(newfn->code->bp, newfn->code->p);
		fprint(2, "--\n");
	}
	*dst = newfn;
}

// Load effective address
void
xlea(Thread *t)
{
	void **dst;
	void *src;
	
	dst = (void**)t->loc();
	src = (void*)t->loc();
	*dst = src;
}

// Load effective address of extension structure
void
xleax(Thread *t)
{
	ExtStruct **dst, **src, **lp, *l;
	Type *xt;

	dst = (ExtStruct**)t->loc();
	src = (ExtStruct**)t->loc();
	xt = (Type*)*t->pc_.p++;
	// Search for extension structure.
	for(lp=src; (l=*lp); lp=&l->next){
		if(l->type == xt){
			// Move to front
			if(lp != src){
				*lp = l->next;
				l->next = *src;
				*src = l;
			}
			*dst = l;
			return;
		}
	}
	// Not found: allocate new extension structure.
	l = (ExtStruct*)galloc(sizeoftype(xt));
	l->next = *src;
	l->type = xt;
	*src = l;
	*dst = l;
}

// Copy a possibly discriminated/extensible structure.
void
copystruct(void *dst, const void *src, Type *dtype, Type *stype)
{
	Type *base;
	void **p;
	int sz, doff, soff;

	base = rootstruct(dtype);
	assert(base && base == rootstruct(stype));

	// First, copy all fields shared by both types (including reserved data)
	sz = sizeoftype(dtype);
	doff = structdataoffset(dtype);
	soff = structdataoffset(stype);
	memcpy((char*)dst + doff, (char*)src + soff, sz - doff);
	if(base->reserved)
		memcpy(dst, src, base->reserved);
	// If discriminated, make sure the dst's type ptr is set
	if(base->discriminated){
		p = (void**)((char*)dst + tagtypeoffset(dtype));
		*p = dtype;
	}
	// If extensible, copy the shared extension structs
	if(base->extensible){
		ExtStruct *sx = *(ExtStruct**)((char*)src + extptroffset(stype));
		ExtStruct *dx = *(ExtStruct**)((char*)dst + extptroffset(dtype));
		copyxstructs(dx, sx, dtype, stype);
	}
}

void
copyxstructs(ExtStruct *dst, const ExtStruct *src, Type *dtype, Type *stype)
{
	ExtStruct **pdst;
	Type *t;
	int sz;

	pdst = &dst;
	for(; src; src=src->next){
		t = src->type;
		assert(t);
		// Skip extensions which `src' doesn't share with `dst'.
		if(t->extends != dtype && !isverticalsubtype(t->extends, dtype))
			continue;
		sz = sizeoftype(t);
		*pdst = (ExtStruct*)galloc(sz);
		memcpy(*pdst, src, sz);
		pdst = &(*pdst)->next;
	}
	*pdst = nil;
}

void
xcopystruct(Thread *t)
{
	Type *dtype, *stype;
	void *dst, *src;

	dst = (void*)t->loc();
	src = (void*)t->loc();
	dtype = (Type*)*t->pc_.p++;
	stype = (Type*)*t->pc_.p++;
	copystruct(dst, src, dtype, stype);
}

// Simple data copying
void
xmov4(Thread *t)
{
	int *dst, *src;
	
	dst = (int*)t->loc();
	src = (int*)t->loc();
	if(src == nil){
		*dst = 0;
		return;
	}
	*dst = *src;
}

void
xmovn(Thread *t)
{
	int *dst, *src, n;
	
	n = *t->pc_.i++;
	dst = (int*)t->loc();
	src = (int*)t->loc();
	if(src == nil){
		memset(dst, 0, n);
		return;
	}
	if(dst == nil)
		panic("nil in xmovn");
	memmove(dst, src, n);
}

// Builtin function call.
void
xbuiltin(Thread *t)
{
	void (**fp)(void*);

	fp = (void(**)(void*))t->loc();
	(*fp)(t->fp_);
}

Fn*
mkbuiltinfn(Name name, void (*impl)(void*), Type *type)
{
	Fn *fn;
	int n;

	fn = mkfn(ZL, name, nil, nil, nil);
	fn->name = name;
	fn->type = type;
	fn->builtin = impl;
	fn->frame = mkFrame();
	fn->frame->size = FrameReserved;
	n = sizeoftype(typeptr(type->left));
	if(type->op == TypeForall)
		type = type->left;
	assert(type->op == TypeFn);
	for(List<Type*> *l=type->types; l; l=l->next)
		n += sizeoftype(l->x);
	fn->frame->argsize = n;
	return fn;
}

// AST structuring.
static void
fixslotorder(void **slots, int n, int *order)
{
	void **tmp;
	int i;

	if(order == nil)
		return;
	tmp = (void**)emalloc(n * sizeof tmp[0]);
	for(i=0; i < n; i++)
		tmp[i] = slots[order[i]];
	memmove(slots, tmp, n * sizeof tmp[0]);
	free(tmp);
}

void
xdestructure(Thread *t)
{
	int *dst = (int*)t->loc();
	Ast *src = *(Ast**)t->loc();
	Type *typ = (Type*)*t->pc_.p++;
	Ast *ast = (Ast*)*t->pc_.p++;
	int n = *t->pc_.i++;
	void *view  = (void*)*t->pc_.p++;
	int *order = (int*)*t->pc_.p++;
	void ***slots = (void***)t->loc();
	List<Var*> *canonicals = (List<Var*>*)*t->pc_.p++;

	if(src == nil){
		*dst = (ast == nil);
		return;
	}
	// Can happen via astsplit()
	if(src->op == AstString){
		*dst = 0;
		return;
	}

	// This check is necessary when the destructure also
	// performs a horizontal downcast.  For example:
	//    a: ptr A.expr;
	//    a ~ B.expr{...}; // B <: A
	// If the test isn't necessary, t == nil.

	if(typ != nil && !issubtypesyntax(typ, src->_specifictype)){
		*dst = 0;
		return;
	}
	fixslotorder((void**)slots, n, order);
	*dst = destructure(canonicals, ast, src, n, slots);
}

// Ugly hook so zeta code can cleanup an Ast post-restructure.
typedef struct PostXForm PostXForm;
struct PostXForm
{
	Thread *t;
	Fn *fn;
	View *view;
};

// FIXME: is this name comparison ok? (looks ok at first glance)
static Ast*
dopostxform(Ast *ast, PostXForm *xf)
{
	struct {
		Ast *in;
		Ast **out;
	} f;
	Ast *out;

	if(ast->symname != xf->view->reptype->name)
		return ast;
	f.in = ast;
	f.out = &out;
	callfn(xf->t, xf->fn, &f);
	return out;
}

void
xrestructure(Thread *t)
{
	Ast **dst;
	Ast *ast;
	Ast *context;
	View *view;
	void **slots;
	Ast **contextp;
	int n, *order;
	PostXForm xf;

	dst = (Ast**)t->loc();
	contextp = (Ast**)t->loc();
	if(contextp)
		context = *contextp;
	else
		context = nil;
	ast = (Ast*)*t->pc_.p++;
	n = *t->pc_.i++;
	view  = (View*)*t->pc_.p++;
	order = (int*)*t->pc_.p++;
	slots = (void**)t->loc();

	fixslotorder(slots, n, order);
	if(view && view->postxform){
		xf.t = t;
		xf.fn = *(Fn**)var2ptr(t, view->postxform);
		xf.view = view;
		*dst = restructure(ast, n, slots, dopostxform, &xf);
	}
	else
		*dst = restructure(ast, n, slots, (Ast*(*)(Ast*,void*))nil, (void*)nil);
	ast = *dst;
	if(context){
		assert(ast->_parent == NoParentYet && !ast->isshared);
		ast->_parent = context;
	}else if(!ast->isshared && ast->_parent == NoParentYet){
		// could set ast->_parent = nil here (instead of NoParentYet), but astparent does it lazily now.
		if(FLAGS_lotsofcopies)
			ast->_parent = nil;
	}
}

void
xrestructslot(Thread *t)
{
	Ast **dst, *link;
	Type *type;
	
	dst = (Ast**)t->loc();
	type = (Type*)*t->pc_.p++;
	link = *(Ast**)t->loc();
	*dst = restructslot(type, link);
}

void
xsamestructure(Thread *t)
{
	Ast *a, *b;
	int *dst;

	dst = (int*)t->loc();
	a = *(Ast**)t->loc();
	b = *(Ast**)t->loc();
	*dst = samestructure(a, b);
}

static int
checkslotorder(Ast *in, Ast *out, int n, int *order)
{
	Ast **inslots, **outslots, **tmp;
	int i, k, tmpn;

	// Gather both sets of slots into a list.
	inslots  = (Ast**)emalloc(n*2 * sizeof inslots[0]);
	outslots = inslots + n;
	tmpn = n;
	tmp  = inslots;
	astcollectslots(in, &tmpn, &tmp);
	assert(tmpn == 0);
	tmpn = n;
	tmp  = outslots;
	astcollectslots(out, &tmpn, &tmp);
	if(tmpn != 0){
		fprint(2, "in: %#A\n", in);
		fprint(2, "out: %#A\n", out);
	}
	assert(tmpn == 0);

	// Compare, using the 'sep' field to link out-->in.
	// order[] maps old indexes --> new indexes.
	for(i=0; i < n; i++)
		order[i] = -1;
	for(i=0; i < n; i++){
		out = outslots[i];
		assert(out);
		for(k=0; k < n; k++){
			in = inslots[k];
			assert(in);
			if(out == in || out->sep == in){
				order[k] = i;
				break;
			}
		}
	}
	for(i=0; i < n; i++)
		assert(order[i] >= 0);
	free(inslots);

	// Return non-zero if the order changed.
	for(i=0; i < n; i++)
		if(order[i] != i)
			return 1;
	return 0;
}

void
astviewxform(Thread *t, PatternXForm *xf)
{
	struct {
		Ast *in;
		Ast **out;
	} f;
	Ast *out, **past;
	int n, **order;
	View *v;
	Var *var;
	Fn *fn;

	past  = (Ast**)(xf->codebuf->bp + xf->off);
	n     = *(int*)(xf->codebuf->bp + xf->off + sizeof(void*));
	v   = *(View**)(xf->codebuf->bp + xf->off + sizeof(void*) + sizeof(int));
	order = (int**)(xf->codebuf->bp + xf->off + 2*sizeof(void*) + sizeof(int));
	assert(v == xf->view);

	// Resolve the xform functions.
	if(v->xform == nil){
		var = lookup(globalscope, v->xformname, 0);
		if(var == nil)
			panic("couldn't find view xform function '%s'", var->name);
		if(var->type->op != TypeFn){
		Badfn:
			panic("bad type for xform function '%s' [%T]", var->name, var->type);
		}
		if(listlen(var->type->types) != 1)
			goto Badfn;
		Type *p = replacegramtype(v->parseme, v->parseme->grammar->wildroottype);
		Type *r = replacegramtype(v->reptype, v->reptype->grammar->wildroottype);
		assert(p);
		assert(r);
		Type *expected = typefn2(typetuple(list1(nil, typeptr(p))), typeptr(r));
		if(!issubtype(var->type, expected))
			goto Badfn;
		v->xform = var;
	}
	if(v->postxform == nil && v->postxformname){
		var = lookup(globalscope, v->postxformname, 0);
		if(var == nil)
			panic("couldn't find view postxform function '%s'", var->name);
		if(var->type->op != TypeFn){
		Badpost:
			panic("bad type for postxform function '%s' [%T]", var->name, var->type);
		}
		if(listlen(var->type->types) != 1)
			goto Badpost;
		Type *r = replacegramtype(v->reptype, v->reptype->grammar->wildroottype);
		assert(r);
		Type *expected = typefn2(typetuple(list1(nil, typeptr(r))), typeptr(r));
		if(!issubtype(var->type, expected))
			goto Badpost;
		v->postxform = var;
	}

	// Transform this ast by the view xform function.
	fn = *(Fn**)var2ptr(t, v->xform);
	f.in = *past;
	out = (Ast*)0xfefefefe;
	f.out = &out;
	t->rundepth_++;     // necessary because this is *not* called from run()
	callfn(t, fn, &f);
	t->rundepth_--;

	// Compute the slot ordering relative to the original order and
	// if it changed, patch the new ordering into the instr stream.
	if(n > 0){
		*order = (int*)emalloc(n * sizeof(int));
		if(!checkslotorder(*past, out, n, *order)){
			free(*order);
			*order = nil;
		}
	}

	// Patch the transformed ast into the instruction stream.
	*past = out;

	if(out == nil)
		error(xf->line, "view transformation failed for {%A}", f.in);

	if(FLAGS_parsedebug)
		fprint(2, "xformed pattern:\n  {%A}\n  {%A}\n %#A\n", f.in, out, out);
}

void
xastparent(Thread *t)
{
	Ast **dst, *src;
	Type *typ;

	dst = (Ast**)t->loc();
	src = *(Ast**)t->loc();
	typ = (Type*)*t->pc_.p++;
	if(typ == nil)
		// invoked by ast.parent
		*dst = astrealparent(src);
	else
		// invoked by astparent()
		*dst = asttypedparent(src, typ);
}

// Convert from a string
void
xatoi(Thread *t)
{
	int *dst;
	char *src;
	
	dst = (int*)t->loc();
	src = *(char**)t->loc();
	if(src == nil)
		*dst = 0;
	else
		*dst = strtol(src, (char**)nil, 0);
}

void
xatof(Thread *t)
{
	double *dst;
	char *src;
	
	dst = (double*)t->loc();
	src = *(char**)t->loc();
	if(src == nil)
		*dst = 0;
	else
		*dst = strtod(src, (char**)nil);
}

// Convert from an int
void
xitoa(Thread *t)
{
	char **dst;
	int src;
	
	dst = (char**)t->loc();
	src = *(int*)t->loc();
	*dst = nameprint("%d", src);
}

void
xitof(Thread *t)
{
	double *dst;
	int src;
	
	dst = (double*)t->loc();
	src = *(int*)t->loc();
	*dst = (double)src;
}

// Convert from a bool
void
xbtoa(Thread *t)
{
	char **dst;
	int src;
	
	dst = (char**)t->loc();
	src = *(int*)t->loc();
	*dst = src ? nameof("true") : nameof("false");
}

void
xbtof(Thread *t)
{
	double *dst;
	int src;
	
	dst = (double*)t->loc();
	src = *(int*)t->loc();
	*dst = src ? 1.0 : 0.0;
}

// Convert from a float
void
xftoa(Thread *t)
{
	char **dst;
	double src;
	
	dst = (char**)t->loc();
	src = *(double*)t->loc();
	*dst = nameprint("%f", src);  // XXX should print exactly
}

void
xftoi(Thread *t)
{
	int *dst;
	double src;
	
	dst = (int*)t->loc();
	src = *(double*)t->loc();
	*dst = (int)src;
}


// AST to string
void
xasttoa(Thread *t)
{
	char **dst;
	Ast *ast;

	dst = (char**)t->loc();
	ast = *(Ast**)t->loc();
	*dst = asttoa(ast);
}

// AST to list
void
xasttolist(Thread *t)
{
	ListNode **dst;
	Ast *ast;

	dst = (ListNode**)t->loc();
	ast = *(Ast**)t->loc();
	*dst = asttolistnode(ast);
}

// AST to option
void
xasttooption(Thread *t)
{
	Ast **dst;
	Ast *ast;

	dst = (Ast**)t->loc();
	ast = *(Ast**)t->loc();
	if(ast->nright == 0)
		*dst = nil;
	else if(ast->nright == 1)
		*dst = astright(ast, 0);
	else
		panic("illegal option(ast) conversion");
}

// String concatenation
void
xadds(Thread *t)
{
	char **dst;
	char *src1, *src2;
	
	dst = (char**)t->loc();
	src1 = *(char**)t->loc();
	src2 = *(char**)t->loc();
	
	if(src1 == nil)
		*dst = src2;
	else if(src2 == nil)
		*dst = src1;
	else
		*dst = nameprint("%s%s", src1, src2);
}

// Non-zero floating point.
void
xnzf(Thread *t)
{
	int *dst;
	double *src;
	
	dst = (int*)t->loc();
	src = (double*)t->loc();
	*dst = *src != 0;
}

// Non-zero string.
void
xnzs(Thread *t)
{
	int *dst;
	char *src;
	
	dst = (int*)t->loc();
	src = *(char**)t->loc();
	*dst = src != nil && *src != 0;
}

// String index - creates new string.
void
xstringindex(Thread *t)
{
	char **dst;
	char *src;
	int index;
	
	dst = (char**)t->loc();
	src = *(char**)t->loc();
	index = *(int*)t->loc();
	
	// strlen: not so efficient.
	if(src == nil || index < 0 || (int)strlen(src) <= index)
		panic("bad string index: %d in %s", index, src);
	*dst = nameofn(src+index, 1);
}

// Load effective address of array entry.
void
xleaa(Thread *t)
{
	Array *a;
	int index;
	void **dst;
	
	dst = (void**)t->loc();
	a = *(Array**)t->loc();
	index = *(int*)t->loc();
	
	if(a == nil || index < 0 || index >= a->n)
		panic("array index out of bounds");
	*dst = (char*)a->v + a->sizeof1*index;
}

// Load effective address of map entry.
void
xleam(Thread *t)
{
	Map *m;
	void *index, **dst, *v;
	
	dst = (void**)t->loc();
	m = *(Map**)t->loc();
	index = *(void**)t->loc();	// assumes pointer, int same size

	if(m == nil)
		panic("map is nil");
	v = hashget(m->h, index);
	if(v == nil){
		v = (void*)galloc(m->sizeof1);
		hashput(m->h, index, v);
	}
	*dst = v;
}

// Load effective address of associative list entry
void
xleaassoc(Thread *t)
{
	AssocList *l;
	void *index, **dst;
	
	dst = (void**)t->loc();
	l = *(AssocList**)t->loc();
	index = *(void**)t->loc();	// assumes pointer, int same size

	for(; l; l=l->next)
		if(l->k == index){
			*dst = &l->v;
			return;
		}
	*dst = nil;
}

// Load effective address of list[n]
void
xlealistn(Thread *t)
{
	ListNode *l;
	void **dst;
	int i;
	
	dst = (void**)t->loc();
	l = *(ListNode**)t->loc();
	i = *(int*)t->loc();

	if(i < 0)
		panic("bad list index %d", i);
	for(; l && i; l=l->next, i--)
		;
	if(!l || i)
		panic("bad list index %d", i);
	*dst = &l->v;
}

// Copy a block of memory of a given type.
static void
typemove(void *dst, void *src, Type *t, int n)
{
	assert(t);
	// Do discriminated and extensible structs specially.
	if(t->op == TypeStruct && (rootstruct(t)->discriminated || t->extensible))
		copystruct(dst, src, t, t);
	else
		memmove(dst, src, n);
}

// Copy a ListNode, but set next to next.
static ListNode*
copylistnode(ListNode *src, ListNode *next, Type *t, int n)
{
	ListNode *l;

	l = (ListNode*)galloc(sizeof(void*) + n);
	l->next = next;
	typemove(l->v, src->v, t, n);
	return l;
}

void
xaddl(Thread *t)
{
	ListNode **dst, *src1, *src2, **lp;
	ListNode *l;
	Type *type;
	int n;
	
	type = (Type*)*t->pc_.p++;  // element type
	dst = (ListNode**)t->loc();
	src1 = *(ListNode**)t->loc();
	src2 = *(ListNode**)t->loc();
	
	if(src1 == nil)
		*dst = src2;
	else if(src2 == nil)
		*dst = src1;
	else{
		n = sizeoftype(type);
		lp = dst;
		for(l=src1; l; l=l->next){
			*lp = copylistnode(l, nil, type, n);
			lp = &(*lp)->next;
		}
		*lp = src2;
	}
}

// Array length
void
xlengtha(Thread *t)
{
	int *dst, len;
	Array *a;
	
	dst = (int*)t->loc();
	a = *(Array**)t->loc();
	if(a)
		len = a->n;
	else
		len = 0;
	*dst = len;
}

// List length
void
xlengthl(Thread *t)
{
	int *dst;
	ListNode *l;
	
	dst = (int*)t->loc();
	l = *(ListNode**)t->loc();
	for(*dst=0; l; l=l->next)
		(*dst)++;
}

// String length
void
xlengths(Thread *t)
{
	int *dst;
	char *s;
	
	dst = (int*)t->loc();
	s = *(char**)t->loc();
	*dst = s ? strlen(s) : 0;
}

// Expand an array into a list
void
xarraytolist(Thread *t)
{
	ListNode **dst, **lp;
	Array *a;
	Type *type;
	int i;

	type = (Type*)*t->pc_.p++;  // element type
	dst = (ListNode**)t->loc();
	a = *(Array**)t->loc();

	if(!a || a->n == 0)
		*dst = nil;
	else {
		assert(a->sizeof1 == sizeoftype(type));
		lp = dst;
		for(i=0; i < a->n; i++){
			*lp = (ListNode*)galloc(sizeof(void*) + a->sizeof1);
			typemove((*lp)->v, (char*)a->v + i*a->sizeof1, type, a->sizeof1);
			lp = &(*lp)->next;
		}
	}
}

// Flatten a list into an array
void
xlisttoarray(Thread *t)
{
	Array **dst, *a;
	ListNode *src, *l;
	Type *type;
	int i, n, m;

	type = (Type*)*t->pc_.p++;  // element type
	dst = (Array**)t->loc();
	src = *(ListNode**)t->loc();

	for(n=0, l=src; l; l=l->next)
		n++;
	m = sizeoftype(type);
	a = newarray(n, m);
	for(l=src, i=0; l; l=l->next, i++)
		typemove((char*)a->v + i*m, l->v, type, m);
	*dst = a;
}

// listmap() primitive
void
xlistmap(Thread *t)
{
	ListNode *src, **dst;
	Type *type, *in, *out;
	Fn *fn;
	void *arg, **ret;
	int min, mout;

	type = (Type*)*t->pc_.p++;  // tuple(in element, out element)
	dst = (ListNode**)t->loc();
	src = *(ListNode**)t->loc();
	fn  = *(Fn**)t->loc();

	assert(type->op == TypeTuple);
	assert(listlen(type->types) == 2);

	in  = listnth(type->types, 0);
	out = listnth(type->types, 1);
	min  = sizeoftype(in);
	mout = sizeoftype(out);
	arg = emalloc(min + sizeof(void*));
	ret = (void**)((char*)arg + min);

	for(; src; src=src->next){
		*dst = (ListNode*)galloc(sizeof(void*) + mout);
		memmove(arg, src->v, min);
		*ret = (*dst)->v;
		callfn(t, fn, arg);
		dst = &(*dst)->next;
	}
	free(arg);
}

// Zero memory
void
xzero(Thread *t)
{
	void *v;
	Type *type, *base;
	int sz, off;
	
	v = (void*)t->loc();
	type = (Type*)*t->pc_.p++;

	sz = sizeoftype(type);
	if(type && type->op == TypeStruct){
		off = structdataoffset(type);
		base = rootstruct(type);
		memset((char*)v + off, 0, sz - off);
		if(base->reserved)
			memset(v, 0, base->reserved);
	}
	else
		memset(v, 0, sizeoftype(type));
}

void
xzeron(Thread *t)
{
	void *v;
	int n;
	
	v = (void*)t->loc();
	n = *t->pc_.i++;
	memset(v, 0, n);
}

// Reverse list
void
xreversel(Thread *t)
{
	ListNode **dst, *src;
	Type *type;
	int n;
	
	dst = (ListNode**)t->loc();
	src = *(ListNode**)t->loc();
	type = (Type*)*t->pc_.p++;

	*dst = nil;
	n = sizeoftype(type);
	for(; src; src=src->next)
		*dst = copylistnode(src, *dst, type, n);
}

// String slice
void
xslices(Thread *t)
{
	char **dst, *src;
	int *phi, lo, hi;
	
	dst = (char**)t->loc();
	src = *(char**)t->loc();
	if(src == nil)
		src = "";
	lo = *(int*)t->loc();
	phi = (int*)t->loc();
	if(phi == nil)
		hi = strlen(src);
	else
		hi = *phi;
	if(lo < 0 || hi < lo || hi > (int)strlen(src))
		panic("bad slice: [%d:%d] of %s", lo, hi, src);
	*dst = nameofn(src+lo, hi-lo);
}

// Array slice
void
xslicea(Thread *t)
{
	static Array a0;
	Array **dst, *src, *a;
	int *phi, lo, hi;
	
	dst = (Array**)t->loc();
	src = *(Array**)t->loc();
	if(src == nil)
		src = &a0;
	lo = *(int*)t->loc();
	phi = (int*)t->loc();
	if(phi == nil)
		hi = src->n;
	else
		hi = *phi;
	if(lo < 0 || hi < lo || hi > src->n)
		panic("bad slice: [%d:%d] of %d-element array", lo, hi, src->n);
	a = (Array*)galloc(sizeof *a);
	*a = *src;
	a->n = hi-lo;
	a->v = (char*)a->v + lo * a->sizeof1;
	*dst = a;
}

// Subtype checking
void
xissubtype(Thread *t)
{
	int *dst;
	Type **p1, **p2, *t1, *t2;

	dst = (int*)t->loc();
	p1 = (Type**)t->loc();
	p2 = (Type**)t->loc();
	// bit of a hack - if 'dst' is nil, one of these pointers
	// is bad and shouldn't be dereferenced
	if(!*dst)
		return;
	// t2 <: t1?
	t1 = *p1;
	t2 = *p2;
	assert(t1 && t1->op == TypeStruct);
	assert(t2 && t2->op == TypeStruct);
	*dst = isverticalsubtype(t1, t2);
}

// Grammar cast checking.
// An Ast can be dynamically cast to a type 't'
// if its specific type is <: 't'.
void
xcancastast(Thread *t)
{
	int *dst;
	Type *typ;
	Ast *ast;

	dst = (int*)t->loc();
	typ = (Type*)*t->pc_.p++;
	ast = *(Ast**)t->loc();

	// Must be a supertype of the most specific grammar rule of this parse tree.
	if(ast == nil || ast->op == AstString)
		*dst = 0;
	else
		*dst = issubtypesyntax(typ, ast->_specifictype);
}

// List <--> ListNode conversions
// These only work when the list element is a pointer
ListNode*
listtolistnode(List<void*> *list)
{
	ListNode **lp, *n, *first;
	List<void*> *l;

	first = nil;
	lp = &first;
	for(l=list; l; l=l->next){
		n = (ListNode*)galloc(sizeof *n);
		*(void**)n->v = l->x;
		*lp = n;
		lp = &n->next;
	}
	return first;
}

List<void*>*
listnodetolist(Arena *a, ListNode *node)
{
	List<void*> *l;
	ListNode *ln;

	l = ZList;
	for(ln=node; ln; ln=ln->next)
		l = listappend(a, l, *(void**)ln->v);
	return l;
}

// Extract the tagname of a discriminated struct.
void
xtagstring(Thread *t)
{
	Name *dst;
	char *src;
	Type *base, *type;
	
	dst = (Name*)t->loc();
	src = *(char**)t->loc();
	base = (Type*)*t->pc_.p++;

	type = *(Type**)(src + tagtypeoffset(base));
	*dst = type->name;
}

// Call the parse builtin
void
xparse(Thread *t)
{
	Name text, file, *p;
	Type *type;
	Ast **dst;

	dst = (Ast**)t->loc();
	type = (Type*)*t->pc_.p++;
	text = *(Name*)t->loc();
	p    = (Name*)t->loc();
	file = p ? *p : Name::null;
	*dst = parsebuiltin(type, text, file);
}

// Attribute evaluation
typedef struct AttrX AttrX;
typedef struct AttrPending AttrPending;

struct AttrX {
	int state;
	int data;
};

struct AttrPending {
	Var *attr;
	Ast *ast;
};
AttrPending pending[10000];
int npending;

void*
computeattr(Ast *ast, Var *attr)
{
	int i;
	AttrX *data;
	AttrBlock *block;
	Type *host;
	Fn *fn;
	AttrPending *p;
	
	if(ast == nil)
		panic("dereferencing nil ast in xattr for '%s'", attr->name);

	assert(attr->isbuiltinattr || attr->attrfn);
	assert(attr->host);

	// Determine which block this attribute is in.
	// Recall that there is a 4 level heirarchy:
	//   Ast > G > G.sym > G.rule
	// and that both "Ast" and "G" share "gram_attrs".
	host = attr->host;
	if(!host->super || !host->super->super)
		block = &ast->gram_attrs;
	else if(!host->super->super->super)
		block = &ast->sym_attrs;
	else
		block = &ast->rule_attrs;

	// Lazily resize
	if(block->size <= attr->off){
		block->attrs = erealloc(block->attrs, host->attrsize);
		memset((char*)block->attrs + block->size, 0, host->attrsize - block->size);
		block->size = host->attrsize;
	}
	data = (AttrX*)((char*)block->attrs + attr->off);

	// Compute
	switch(data->state){
	case 0:	// not computed
		data->state = 1;
		if(npending >= nelem(pending)){
			fprint(2, "attribute stack overflow\n");
			goto showstack;
		}
		p = &pending[npending++];
		p->attr = attr;
		p->ast = ast;
		// Builtin
		if(attr->isbuiltinattr){
			if(attr->name == Name::prev){
				astprevnext(ast, (Ast**)&data->data, nil);
			}
			else if(attr->name == Name::next)
				astprevnext(ast, nil, (Ast**)&data->data);
			else
				panic("unknown builtin attribute '%s'", attr->name);
		}
		// Normal (user defined)
		else{
			struct {
				Ast *in;
				void *out;
			} args;
			fn = *(Fn**)var2ptr(running, attr->attrfn);
			if(fn == nil)
				panic("no fn defined for attribute %T :: %s", attr->host, attr->name);
			assert(fn->frame->argsize == 2*sizeof(void*));
			args.in = ast;
			args.out = &data->data;
			if(debugfd)
				traceenterattr(ast, attr->name, attr->host);
			callfn(running, fn, &args);
			if(debugfd)
				traceleaveattr();
		}
		npending--;
		data->state = 2;
		break;

	case 1:	// being computed
		{
			Ast *prev, *next;
			astprevnext(ast, &prev, &next);
			fprint(2, "circular computation of attribute %T.%s, for %p:\n"
			      "%#lA\n"
			      "parent = %p (%s)\n"
			      "prev   = %p (%s)\n"
			      "next   = %p (%s)\n"
			      "full parent tree is:\n"
			      "%#lA\n",
			      attr->host, attr->name, ast, ast,
			      ast->_parent, ast->_parent ? (char*)ast->_parent->symname : "<nil>",
			      prev, prev ? (char*)prev->symname : "<nil>",
			      next, next ? (char*)next->symname : "<nil>",
			      ast->_parent);
		showstack:
			fprint(2, "attribute computation stack:\n");
			for(i=npending-1; i>=0; i--){
				enum { L = 80 - 12 - 1 - 8 };
				char buf[L+5];
				p = &pending[i];
				snprint(buf, L+2, "%#A", p->ast);
				if(strlen(buf) > L)
					strcpy(buf+L-3, "...");
				fprint(2, "%12s %p %s\n", p->attr->name,
					p->ast, buf+1);
			}
			panic("attribute");
		}

	case 2: // already computed
		if(!attr->isbuiltinattr && debugfd)
			tracereadattr(ast, attr->name, attr->host);
		break;
	}
	return &data->data;
}


void
xattr(Thread *t)
{
	void *dst = (void*)t->loc();
	Ast *ast = *(Ast**)t->loc();
	Var *attr = (Var*)*t->pc_.p++;
	void *src = computeattr(ast, attr);
	if(dst)
		memmove(dst, src, sizeoftype(attr->type));
}

void
xnop(Thread *t)
{
}

