#include <sys/socket.h>
#include "a.h"

int debugfd;

// ========================================================
// Message buffers

struct debugmsg
{
	char *buf;
	char *pos;
	char *end;
	int allocated;
};

static struct debugmsg mkasttrace;
static struct debugmsg attrtrace;

// Assert that there are at least size bytes of data left in msg.
static void
dmcheck(struct debugmsg *msg, int size)
{
	if(msg->end - msg->pos < size) {
		panic("debug protocol error: message terminated early");
	}
}

// Ensure that there are at least size bytes free at the end of msg.
// The cursor must be at the end of the message.
static void
dmensure(struct debugmsg *msg, int size)
{
	int resize = 0;

	assert(msg->end == msg->pos);
	assert(size < 10*1024*1024);

	if(msg->allocated == 0) {
		resize = 1;
		msg->allocated = 1;
	}
	while(msg->allocated - (msg->end - msg->buf) < size) {
		resize = 1;
		msg->allocated *= 2;
	}

	if(resize) {
		char *newbuf, *newpos, *newend;
		newbuf = static_cast<char*>(realloc(msg->buf, msg->allocated));
		newpos = newbuf + (msg->pos - msg->buf);
		newend = newbuf + (msg->end - msg->buf);
		msg->buf = newbuf;
		msg->pos = newpos;
		msg->end = newend;
	}
}

// Test if the cursor is at the end of msg.
static int
dmiseom(struct debugmsg *msg)
{
	return msg->pos == msg->end;
}

// Assert that the cursor is at the end of msg.
static void
dmeom(struct debugmsg *msg)
{
	if(!dmiseom(msg)) {
		panic("debug protocol error: extra data in message");
	}
}

// Move the cursor to the beginning of msg.
static void
dmrewind(struct debugmsg *msg)
{
	msg->pos = msg->buf;
}

// Clear the contents of msg and reset the cursor.
static void
dmreset(struct debugmsg *msg)
{
	msg->pos = msg->end = msg->buf;
}

// Initialize msg.
static void
dmclear(struct debugmsg *msg)
{
	msg->buf = msg->pos = msg->end = 0;
	msg->allocated = 0;
}

// ========================================================
// Sending and receiving message buffers

// Receive len bytes from fd, appending them to msg.
static void
dmrecv(struct debugmsg *msg, int fd, int len)
{
	int got;

	dmensure(msg, len);
	while(len > 0) {
		got = recv(fd, msg->pos, len, 0);
		assert(got <= len);
		if(got == 0)
			panic("Debug client unexpectedly disconnected");
		if(got == -1)
			panic("Error receiving debug message");
		msg->end += got;
		msg->pos = msg->end;
		len -= got;
	}
}

// Send the contents of msg, starting at the cursor, to fd.
static void
dmsend(struct debugmsg *msg, int fd)
{
	int msglen;

	dmrewind(msg);
	while(msg->pos < msg->end) {
		msglen = send(fd, msg->pos, msg->end-msg->pos, 0);
		if(msglen == -1)
			panic("Error sending debug response");
		msg->pos += msglen;
	}
}

// ========================================================
// Message buffer wire encoding

#define DMGETPUT(name, typ)			\
	static typ				\
	dmget##name(struct debugmsg *msg)	\
	{					\
		typ *v;				\
		dmcheck(msg, sizeof(typ));	\
		v = (typ*)msg->pos;		\
		msg->pos += sizeof(typ);	\
		return *v;			\
	}					\
	static void				\
	dmput##name(struct debugmsg *msg, typ v) \
	{					\
		dmensure(msg, sizeof(typ));	\
		*(typ*)msg->pos = v;		\
		msg->end += sizeof(typ);	\
		msg->pos = msg->end;		\
	}

DMGETPUT(bool, char)
DMGETPUT(byte, signed char)
DMGETPUT(int, int)
DMGETPUT(pointer, void*)

#undef DMGETPUT

static const void*
dmgetbuf(struct debugmsg *msg, int *len)
{
	int buflen;
	const void *buf;

	buflen = dmgetint(msg);
	if(buflen == -1)
		return nil;
	if(len)
		*len = buflen;
	dmcheck(msg, buflen);
	buf = msg->pos;
	msg->pos += buflen;
	return buf;
}

static void
dmputbuf(struct debugmsg *msg, const void *buf, int buflen)
{
	if (buf == nil) {
		assert(buflen == 0);
		dmputint(msg, -1);
		return;
	}
	dmputint(msg, buflen);
	dmensure(msg, buflen);
	if(buflen > 0)
		memcpy(msg->pos, buf, buflen);
	msg->end += buflen;
	msg->pos = msg->end;
}

static const char*
dmgetstring(struct debugmsg *msg)
{
	int len;
	const char *str;
	str = static_cast<const char*>(dmgetbuf(msg, &len));
	if(!str)
		return nil;
	if(*(str+len-1) != 0)
		panic("debug protocol error: unterminated string");
	return str;
}

static void
dmputstring(struct debugmsg *msg, const char *str)
{
	if(!str) {
		dmputbuf(msg, nil, 0);
	} else {
		int len = strlen(str)+1;
		dmputbuf(msg, str, len);
	}
}

static void
dmputname(struct debugmsg *msg, Name name)
{
	// This could be optimized for better space usage.  This is
	// only worthwhile if the in-memory attribute traces are
	// getting too large.
	dmputstring(msg, name);
}

// ========================================================
// Debug protocol

static int
debugmessage(struct debugmsg *msg, struct debugmsg *out)
{
	const char* cmd = dmgetstring(msg);
	int i;

	if(strcmp(cmd, "getAst") == 0) {
		Ast *ast = static_cast<Ast*>(dmgetpointer(msg));
		int hastext = 0;
		dmeom(msg);
		switch(ast->op) {
		case AstString:
			dmputstring(out, "string");
			hastext = 1;
			break;
		case AstRule:
			dmputstring(out, "rule");
			break;
		case AstSlot:
			dmputstring(out, "slot");
			hastext = 1;
			break;
		case AstListSlot:
			dmputstring(out, "listslot");
			hastext = 1;
			break;
		case AstMerge:
			dmputstring(out, "merge");
			break;
		default:
			panic("debugmessage a: unknown ast op", ast->op);
		}
		dmputstring(out, ast->symname);
		if(hastext) {
			dmputbool(out, 1);
			dmputstring(out, ast->text);
		} else {
			dmputbool(out, 0);
		}
		dmputbool(out, ast->isshared);
		dmputint(out, ast->traceid);
		dmputpointer(out, ast->_parent);
		dmputpointer(out, ast->copiedfrom);
		dmputname(out, ast->zline.file);
		dmputint(out, ast->zline.line);
		dmputint(out, ast->nright);
		for(i=0; i<ast->nright; ++i) {
			dmputpointer(out, ast->_right[i]);
		}
	} else if (strcmp(cmd, "astTrace") == 0) {
		int useFile = dmgetbool(msg);
		Name file = (useFile ? nameof(dmgetstring(msg)) : Name::null);
		int useSymname = dmgetbool(msg);
		Name symname = useSymname ? nameof(dmgetstring(msg)) : Name::null;
		dmeom(msg);
		dmrewind(&mkasttrace);
		while(!dmiseom(&mkasttrace)) {
			Ast *ast = static_cast<Ast*>(dmgetpointer(&mkasttrace));
			if (useFile && file != ast->zline.file)
				continue;
			if (useSymname && symname != ast->symname)
				continue;
			dmputpointer(out, ast);
		}
	} else if (strcmp(cmd, "attrTrace") == 0) {
		dmeom(msg);
		dmputbuf(out, attrtrace.buf, attrtrace.end-attrtrace.buf);
	} else if (strcmp(cmd, "listFiles") == 0) {
		Arena arena;
		memset(&arena, 0, sizeof arena);
		Hash<Name,Ast*> *files = mkhasha<Name,Ast*>(&arena);
		dmeom(msg);
		dmrewind(&mkasttrace);
		while(!dmiseom(&mkasttrace)) {
			Ast *ast = static_cast<Ast*>(dmgetpointer(&mkasttrace));
			if (!hashget(files, ast->zline.file)) {
				hashput(files, ast->zline.file, ast);
				dmputname(out, ast->zline.file);
			}
		}
		aclear(&arena);
	} else if (strcmp(cmd, "close") == 0) {
		dmeom(msg);
		return 1;
	} else {
		panic("Unknown debug message '%s'", cmd);
	}
	return 0;
}

// ========================================================
// Debugger interface

void
debuginit(void)
{
	char *zdb_fd = getenv("ZDB_FD");
	char *end;
	if(zdb_fd) {
		debugfd = strtol(zdb_fd, &end, 10);
		if (*zdb_fd == '\0' || *end != '\0')
			panic("Invalid ZDB_FD environment");
	} else {
		debugfd = 0;
	}
}

void
debugbreak(void)
{
	int done;
	int msglen;
	struct debugmsg msg, out;

	dmclear(&msg);
	dmclear(&out);

	assert(debugfd != 0);
	for(done = 0; !done;) {
		// Get length
		dmreset(&msg);
		dmrecv(&msg, debugfd, sizeof(int));
		dmrewind(&msg);
		msglen = dmgetint(&msg);

		/// Get message
		dmreset(&msg);
		dmrecv(&msg, debugfd, msglen);
		dmrewind(&msg);

		// Process message
		dmreset(&out);
		done = debugmessage(&msg, &out);

		// Send response
		dmreset(&msg);
		dmputint(&msg, out.end-out.buf);
		dmsend(&msg, debugfd);
		dmsend(&out, debugfd);
	}
}

// ========================================================
// Trace recording

void
tracemkast(Ast *ast)
{
	static int counter;
	assert(debugfd);
	dmputpointer(&mkasttrace, ast);
	ast->traceid = counter++;
}

void
traceenterattr(Ast *ast, Name name, Type *host)
{
	assert(debugfd);
	dmputbyte(&attrtrace, 0);
	dmputpointer(&attrtrace, ast);
	dmputname(&attrtrace, name);
	// XXX Should provide interface to types
	dmputname(&attrtrace, host->name);
	Type *tag = ast->_specifictype;
	while(tag) {
		dmputbool(&attrtrace, 1);
		dmputname(&attrtrace, tag->name);
		tag = tag->super;
	}
	dmputbool(&attrtrace, 0);
}

void
traceleaveattr()
{
	assert(debugfd);
	dmputbyte(&attrtrace, 1);
}

void
tracereadattr(Ast *ast, Name name, Type *host)
{
	assert(debugfd);
	dmputbyte(&attrtrace, 2);
	dmputpointer(&attrtrace, ast);
	dmputname(&attrtrace, name);
	dmputname(&attrtrace, host->name);
	Type *tag = ast->_specifictype;
	while(tag) {
		dmputbool(&attrtrace, 1);
		dmputname(&attrtrace, tag->name);
		tag = tag->super;
	}
	dmputbool(&attrtrace, 0);
}
