// Grammar construction
// Abstract syntax construction
// String conversions
#include "grammar.h"

// Symbol names are unique for every grammar (and rule names
// are unique for every symbol), so there are no duplicate tags.
// Thus, we can ignore the "scope" parameter in this file.
#define lookuptag(a, b)   lookuptag(a, b, nil)

// For debugging the grammar: don't addwildcard WILD and LISTWILD rules.
// (Breaks everything involving parse patterns!)
DEFINE_bool(noslots, false, "Turn off auto-slots; breaks everything!");

// Debugging modular grammar stuff
DEFINE_int32(modulargramdebug, 0, "Modular grammars debug level.");

Line GramLine;           // hack for findgrammar

// Use instead of cfgrule and cfgrulesym in this file
typedef CfgAction(AstToken) AstAction;

static CfgRule<AstToken>* addrule(Type*, Name, AstAction*, CfgPrec<AstToken>*, CfgSym<AstToken>*, CfgSym<AstToken>**, int, int*);
static CfgRule<AstToken>* addrulesym(Type*, AstAction*, CfgPrec<AstToken>*, CfgSym<AstToken>*, CfgSym<AstToken>*, ...);

// ========================================================
// Symbol names

// Reserved symbol names
static Name symreserved[] = {
	nameof("EOF"),      // cfg special names
	nameof("error"),
	nameof(".any"),     // wildcards
	nameof("@.any"),
	nameof("merge"),    // for matching merges
	nameof("slot"),     // for matching slots
};

int
isreservedsymname(Name name)
{
	int i;
	static int init;

	if(!init){
		init = 1;
		// Convert reserved words from char* to Name.
		for(i=0; i<nelem(symreserved); i++)
			symreserved[i] = nameof(symreserved[i]);
	}

	if(!name || !*name)
		return 0;
	// L.*, R.*, and list syms
	if(strncmp(name, "L.", 2) == 0 || strncmp(name, "R.", 2) == 0)
		return 1;
	if(name[strlen(name)-1] == '@')
		return 1;
	for(i=0; i<nelem(symreserved); i++)
		if(symreserved[i] == name)
			return 1;
	return 0;
}

static int
checkreserved(Line line, Name name)
{
	if(isreservedsymname(name)){
		error(line, "symbol name '%s' is reserved", name);
		return 1;
	}
	return 0;
}

// ========================================================
// Grammar Names

Type*
findgrammar(Name name)
{
	Type *t = lookuptype(globalscope, name, 0);
	if(name && (t == nil || t->op != TypeSyntax))
		error(GramLine, "couldn't find grammar '%s'\n", name);
	return t;
}

// ========================================================
// Symbol construction

// Add a new non-terminal symbol to the grammar.
// If a symbol with the given name already exists, do nothing.
// Either way, return the CfgSym<AstToken>* for name.  
// If isnew != NULL, set *isnew to whether the CfgSym<AstToken>* is newly created.
//
// If this is the root grammar, use name as the CfgSym<AstToken> name directly.
// Otherwise, name the CfgSym<AstToken> "grammar.name".
static CfgSym<AstToken>*
addsym(Type *tgram, Name name, int *isnew)
{
	Gram *g = tgram->grammar;
	CfgSym<AstToken> *sym;
	Type *tsym;

	if((tsym = lookuptag(tgram, name)) != nil){
		if(isnew)
			*isnew = 0;
		assert(tsym->cfgsym);
		return tsym->cfgsym;
	}

	if(g->roottype != tgram)
		sym = cfgnewsym(&g->cfg, nameprint("%s.%s", tgram->name, name), nil);
	else
		sym = cfgnewsym(&g->cfg, name, nil);
	if(isnew)
		*isnew = 1;
	addsymtype(tgram, sym, name);
	return sym;
}

// Add a new literal string to the lexer and to the grammar.
// If a symbol for the given literal already exists, do nothing.
// Either way, return the CfgSym<AstToken>* for the literal string.
// 
// Uses L.str as the name for the CfgSym<AstToken>.
static CfgSym<AstToken>*
addstr(Type *tgram, char *str)
{
	Gram *g = tgram->grammar;
	CfgSym<AstToken> *sym;

	Name n = nameof(str);
	if((sym = hashget(g->literals, n)) == nil){
		sym = cfgnewsym(&g->cfg, nameprint("L.%s", str), nil);
		hashput(g->literals, n, sym);
		addtodfa(g->lexer, n, RegexpLiteral, --g->dfaprec, sym);
	}
	return sym;
}

// Add a new regular expression to the lexer and to the grammar.
// If a symbol for the given literal already exists, do nothing.
// Either way, return the CfgSym<AstToken>* for the regexp.
// The flags are for the lexer.  
// 
// Uses R.regexp as the name for the CfgSym<AstToken>.
static CfgSym<AstToken>*
addregexp(Type *tgram, char *regexp, int flags)
{
	Gram *g = tgram->grammar;
	CfgSym<AstToken> *sym;

	Name n = nameof(regexp);
	if((sym = hashget(g->regexps, n)) == nil){
		sym = cfgnewsym(&g->cfg, nameprint("R.%s", n), nil);
		sym->reflags = flags;
		hashput(g->regexps, n, sym);
		addtodfa(g->lexer, n, RegexpEgrep|flags, --g->dfaprec, sym);
	}
	return sym;
}

// Add sym: .any rule if needed.
static void
addwildcard(CfgSym<AstToken> *sym)
{
	Gram *g;
	if(sym->slotted || FLAGS_noslots)
		return;
	sym->slotted = 1;
	g = (Gram*)sym->gram;
	addrulesym(_cfgsymtype(sym), action, nil, sym, g->any, nil);
}

// Create the internal term for a list of `one' syms
// separated by `sep'.  Base is the base name: the created
// symbol will be named `base@'. 
// Note that foo@ is used only internally, and cannot be 
// substituted for a any (its raison d'être).
// For example, to create foo[,]@, base is "foo[,]",
// one is the sym for "foo", and sep is the sym for ",".
// To create foo@, base is "foo", one is the again the
// sym for "foo", and sep is nil.
static CfgSym<AstToken>*
gtermlistsym(Type *tgram, Name base, CfgSym<AstToken> *one, CfgSym<AstToken> *sep)
{
	int isnew;
	CfgSym<AstToken> *sym, *listsym;
	CfgPrec<AstToken> *prec;
	Gram *g;
	Type *tsym;

	assert(one);
	g = tgram->grammar;
	sym  = addsym(tgram, nameprint("%s@", base), &isnew);
	tsym = _cfgsymtype(sym);
	if(isnew){
		// base@: one
		// base@: base@ [sep] one
		addrulesym(tsym, action, nil, sym, one, nil);
		if(sep){
			prec = cfglookprec(&g->cfg, cescapestring(sep->name+2, -1, '"'));
			addrulesym(tsym, action, prec, sym, sym, sep, one, nil);
		}
		else
			addrulesym(tsym, action, nil, sym, sym, one, nil);
		
		// base@: @one
		listsym = addsym(tgram, nameprint("@%s", one->name), &isnew);
		addrulesym(tsym, action, nil, sym, listsym, nil);

		// base@: @.any
		addrulesym(tsym, action, nil, sym, g->anylist, nil);
	}
	return sym;
}

// Translate a grammar term like "foo" or bar or bar[,]?
// into a CfgSym<AstToken>*.
static CfgSym<AstToken>*
gtermtosym(Type *tgram, GTerm *gterm)
{
	int isnew;
	Name base, name;
	CfgSym<AstToken> *sym, *sub, *sep, *one;

	if(gterm == nil)
		return nil;

	switch(gterm->op){
	case GTermName:
		sym = addsym(tgram, gterm->name, nil);
		addwildcard(sym);
		return sym;

	case GTermString:
		return addstr(tgram, gterm->str);

	case GTermRegexp:
		return addregexp(tgram, gterm->str, gterm->flags);

	case GTermOpt:
		sym = addsym(tgram, nameprint("%s?", gterm->name), &isnew);
		if(isnew){
			// name?: (empty)
			// name?: name
			sub = addsym(tgram, gterm->name, nil);
			addrulesym(_cfgsymtype(sym), action, nil, sym, nil);
			addrulesym(_cfgsymtype(sym), action, nil, sym, sub, nil);
			addwildcard(sym);
		}
		return sym;

	case GTermStar:
	case GTermPlus:
		if(gterm->str)
			base = nameprint("%s[%s]", gterm->name, gterm->str);
		else
			base = gterm->name;
		if(gterm->op == GTermStar)
			name = nameprint("%s*", base);
		else
			name = nameprint("%s+", base);
		sym = addsym(tgram, name, &isnew);
		if(isnew){
			// base*: base@
			// base*: (empty)
			//    or
			// base+: base@
			if(gterm->str)
				sep = addstr(tgram, gterm->str);
			else
				sep = nil;
			one = addsym(tgram, gterm->name, nil);
			sub = gtermlistsym(tgram, base, one, sep);
			// Make the non-empty rule first, which matches comments in grammar.h.
			addrulesym(_cfgsymtype(sym), action, nil, sym, sub, nil);
			if(gterm->op == GTermStar)
				addrulesym(_cfgsymtype(sym), action, nil, sym, nil);
			addwildcard(sym);
		}
		return sym;

	default:
		panic("bad gterm op %d", gterm->op);
		return nil;
	}
}


// ========================================================
// Rule construction
// Naming of CfgRules<AstToken> works like CfgSym<AstToken> naming; names get
// prefixed with "grammar.sym.".

static CfgRule<AstToken>*
addrule(Type *tleft, Name name, AstAction *fn, CfgPrec<AstToken> *prec,
        CfgSym<AstToken> *left, CfgSym<AstToken> **right, int nright, int *isnew)
{
	int oldnrule;
	CfgRule<AstToken> *r;
	Gram *g = tleft->grammar;
	Type *tgram = tleft->super;

	if(isnew)
		*isnew = 0;
	// Skip parsing rules.
	if(left->name && left->name[0] == '$')
		return nil;
	// Add a rule. It gets a type and a wildcard.
	addwildcard(left);
	oldnrule = g->cfg.nrule;
	r = cfgrule(&g->cfg, fn, prec, left, right, nright);
	if(name){
		if(g->roottype != tgram)
			r->name = nameprint("%s.%s.%s", tgram->name, tleft->name, name);
		else
			r->name = name;
	}
	else{
		name = r->name;
	}
	// Is this a new rule?
	if(g->cfg.nrule != oldnrule){
		if(isnew)
			*isnew = 1;
		addruletype(tleft, r, name);
	}
	return r;
}

static CfgRule<AstToken>*
addrulesym(Type *tleft, AstAction *fn, CfgPrec<AstToken> *prec,
           CfgSym<AstToken> *left, CfgSym<AstToken> *right0, ...)
{
	va_list arg;
	CfgRule<AstToken> *r;
	Gram *g = tleft->grammar;

	// Add a rule. It gets a type and a wildcard.
	addwildcard(left);
	va_start(arg, right0);
	r = vcfgrulesym(&g->cfg, fn, prec, left, right0, arg);
	va_end(arg);
	addruletype(tleft, r, r->name);
	return r;
}

// ========================================================
// Grammar construction

static Gram*
mkgram(Name name, Scope *scope)
{
	Gram *g;

	g = (Gram*)cfgnewgram<AstToken>(sizeof *g);
	g->cfg.valsize = sizeof(AstToken);
	g->name = name;

	g->lexer = mkdfa<CfgSym<AstToken> >();
	g->literals = mkhash<Name, CfgSym<AstToken>*>();
	g->regexps = mkhash<Name, CfgSym<AstToken>*>();
	g->sharedsyms = mkhash<CfgSym<AstToken>*, intptr>();
	g->sharedrules = mkhash<CfgRule<AstToken>*, intptr>();

	addgramtype(g, name, scope);

	// Reserved symbols
	g->any     = addsym(g->roottype, nameof(".any"),  nil);
	g->anylist = addsym(g->roottype, nameof("@.any"), nil);
	g->eof     = addsym(g->roottype, nameof("EOF"),   nil);
	g->error   = addsym(g->roottype, nameof("error"), nil);
	g->slot    = addsym(g->roottype, nameof("slot"),  nil);
	g->merge   = addsym(g->roottype, nameof("merge"), nil);

	return g;
}

static void
addgramterm(Gram *g, Type *tgram, PGram *p)
{
	CfgSym<AstToken> *left, **terms;
	CfgPrec<AstToken> *prec, *hi, *lo;
	CfgGram<AstToken> *gram;
	CfgRule<AstToken> *rule;
	Name symname, rulename;
	int i, n, isnew;
	GTerm *t;
	View *view;
	CfgPrecKind pkind;
	Type *tsym;

	gram = &g->cfg;
	terms = nil;

	List<GTerm*> *l;
	List<Name> *ln;

	switch(p->op){
	case PGramRule:
		// Convert all rule symbols
		n = listlen(p->terms);
		terms = (CfgSym<AstToken>**)emalloc(n*sizeof terms[0]);
		for(i=0, l=p->terms; i<n; i++, l=l->next){
			t = l->x;
			if(checkreserved(t->line, t->name))
				goto done;
			terms[i] = gtermtosym(tgram, t);
		}
		if(checkreserved(p->left->line, p->left->name))
			break;
		left = gtermtosym(tgram, p->left);
		tsym = lookuptag(tgram, p->left->name);
		// Check if this name exists.
		if(p->name && lookuptag(_cfgsymtype(left), p->name) != nil){
			error(p->line, "rule name clash at '%s'", p->name);
			break;
		}
		// Check the prec.
		prec = cfglookprec(gram, p->prec);
		if(p->prec && !prec)
			error(p->line, "unknown precedence level '%s'", p->prec);
		// Add the rule, and warn about duplicates
		rule = addrule(tsym, p->name, action, prec, left, terms, n, &isnew);
		if(!isnew)
			warn(p->line, "duplicate rule '%s' (ignoring)", rule->name);
		//fprint(2, "NEW RULE: %s :: %R\n", tgram->name, rule);
		break;

	case PGramIgnore:
		// XXX ignored terms are not extension specific
		for(l=p->terms; l; l=l->next){
			t = l->x;
			switch(t->op){
			case GTermName:
				error(t->line, "cannot ignore %s", t->str);
				break;
			case GTermString:
				addtodfa(g->lexer, t->str, RegexpLiteral, 0, Ignore);
				g->ignores = listappend(nil, g->ignores, nameprint("L.%s", t->str));
				break;
			case GTermRegexp:
				addtodfa(g->lexer, t->str, RegexpEgrep|t->flags, 0, Ignore);
				g->ignores = listappend(nil, g->ignores, nameprint("R.%s", t->str));
				break;
			case GTermOpt:
			case GTermStar:
			case GTermPlus:
				error(t->line, "cannot ignore compound terms");
				break;
			}
		}
		break;

	case PGramAssoc:
		// TODO: make these names scoped to an extension
		for(ln=p->names; ln; ln=ln->next){
			if((prec = cfglookprec(gram, ln->x)) != nil){
				if(prec->assoc && prec->assoc != p->assoc)
					error(p->line, "conflicting associativity for '%s'\n", ln->x);
			}
			else{
				prec = cfgnewprec(gram, ln->x, CfgPrecPriority);
				prec->assoc = (CfgAssoc)p->assoc;
			}
		}
		break;

	case PGramPriority:
	case PGramPrefer:
		// TODO: make these names scoped to an extension
		pkind = (p->op == PGramPriority) ? CfgPrecPriority : CfgPrecPrefer;
		if(listlen(p->names) < 2)
			warn(p->line, "meaningless predecence directive");
		hi = nil;
		for(ln=p->names; ln; ln=ln->next, hi=lo){
			if((lo = cfglookprec(gram, ln->x)) != nil){
				if(lo->kind != pkind)
					error(p->line, "conflicting precedence kind for '%s'", lo->name);
			}
			else if((lo = cfgnewprec(gram, ln->x, pkind)) == nil)
				error(p->line, "conflicting precedence kind for '%s'", lo->name);
			if(!cfglinkprec(hi, lo))
				error(p->line, "circular precedence directive");
		}
		break;

	case PGramShare:
		for(List<List<Name>*> *ll=p->lnames; ll; ll=ll->next){
			ln = ll->x;
			assert(ln && ln->x);
			// Hacky: terms is list<list<Name>>, inner list is 'sym' or 'sym,rule'
			symname  = ln->x;
			rulename = ln->next ? ln->next->x : Name::null;
			CfgSym<AstToken> *sym = gtermtosym(tgram, mkGTermName(symname));
			if(rulename){
				Type *t;
				// Rules must exist
				t = lookuptag(_cfgsymtype(sym), rulename);
				if(t == nil){
					error(p->line, "no rule named '%s.%s'", symname, rulename);
					break;
				}
				rule = t->cfgrule;
				assert(rule);
				hashput(g->sharedrules, rule, 1L);
			}
			else{
				hashput(g->sharedsyms, sym, 1L);
			}
		}
		break;

	case PGramView:
		// p->name  = view symbol ("viewtype")
		// p->name2 = concrete symbol ("reptype")
		// Viewname can't be an existing symbol
		if(lookuptag(tgram, p->name)){
			error(p->line, "bad view; symbol '%s' already exists", p->name);
			break;
		}
		if((tsym = lookuptag(tgram, p->name2)) == nil){
			error(p->line, "no symbol '%s'", p->name2);
			break;
		}
		// Ugly construction of a view.
		view = mkgramview(tgram, p->name, tsym);
		for(List<List<Name>*> *ll=p->lnames; ll; ll=ll->next){
			ln = ll->x;
			assert(listlen(ln) >= 2);
			// Hacky: terms is list<list<Name>>, inner list is propname # values
			Name prop    = ln->x;
			List<Name> *values = ln->next;
			if(strcmp(prop, "pattern") == 0){
				view->parseme = lookuptag(tgram, values->x);
				assert(view->parseme);
				assert(values->next == nil);
			}
			else if(strcmp(prop, "slotin") == 0){
				List<CfgSym<AstToken>*> *l;
				assert(listlen(values) >= 2);
				symname = values->x;
				l = hashget(view->slotins, symname);
				for(values=values->next; values; values=values->next){
					Name valuename = values->x;
					Type *t;
					if((t = lookuptag(tgram, valuename)) == nil){
						error(p->line, "no symbol '%s'", valuename);
						break;
					}
					l = listappend(nil, l, t->cfgsym);
				}
				hashput(view->slotins, symname, l);
			}
			else if(strcmp(prop, "slotout") == 0){
				assert(listlen(values) == 2);
				Name key = values->x;
				Name out = values->next->x;
				if(lookuptag(tgram, key) == nil){
					error(p->line, "no symbol '%s'", key);
					break;
				}
				if(lookuptag(tgram, out) == nil){
					error(p->line, "no symbol '%s'", key);
					break;
				}
				hashput(view->slotouts, key, out);
			}
			else if(strcmp(prop, "xform") == 0){
				view->xformname = values->x;
				assert(values->next == nil);
			}
			else if(strcmp(prop, "postxform") == 0){
				view->postxformname = values->x;
				assert(values->next == nil);
			}
			else
				error(p->line, "bad property '%s'", prop);
		}
		break;
	}

done:
	free(terms);
}

Type*
addgrammar(Typecheck *tc, StmtGrammar *stmt)
{
	Gram *gram;   // the CFG for this family
	Type *type;   // the new grammar's type
	Arena a;

	memset(&a, 0, sizeof a);
	GramLine = stmt->line_;
	gram = nil;
	type = nil;

	// Unnamed grammar extension.  Auto-generate with a mangled
	// name so we don't stomp on the global namespace.
	// (Could be more clever and take the grammar name from the filename.)
	if(!stmt->name_){
		Name rootname = stmt->extends_->x;
		assert(rootname);
		int i=0;
		do{
			i++;
			stmt->name_ = nameprint("%s_#%d", rootname, i);
		}while(lookuptype(tc->scope_, stmt->name_, 0) != nil);
	}
	// The grammar name must be unique.
	else{
		if(lookuptype(tc->scope_, stmt->name_, 0) != nil){
			error(stmt->line_, "redefinition of type '%s'\n", stmt->name_);
			goto out;
		}
	}

	// Obtain the grammar's CFG and make a new type.
	if(!stmt->extends_){
		// New grammar family.
		gram = mkgram(stmt->name_, tc->scope_);
		type = gram->roottype;
	}
	else{
		// Extending an old family.
		int olderrors = errors;
		List<Type*> *extends = listmap(nil, stmt->extends_, findgrammar);
		if(errors > olderrors)
			goto out;
		// All grammars must be in this family.
		for(List<Type*> *l=extends; l; l=l->next){
			Type *t = l->x;
			if(!gram)
				gram = t->grammar;
			else if(t->grammar != gram){
				error(stmt->line_, "mixed grammar family '%s' in extends\n", t->name);
				goto out;
			}
		}
		assert(gram);
		type = addgramtype(gram, stmt->name_, tc->scope_);
		// Remember which types we extend, for lazy type expansions.
		type->types = extends;
		// Connect the DAG of grammar extensions.
		// The inherited types will be copied lazily.
		for(List<Type*> *l=extends; l; l=l->next){
			Type *t = l->x;
			t->supergrammars = listappend(nil, t->supergrammars, type);
		}
	}

	// Add the new symbols and rules to this grammar.
	for(List<PGram*> *l=stmt->grams_; l; l=l->next)
		addgramterm(gram, type, l->x);

	// This grammar is visible in the current scope.
	tc->scope_->grams = listinsert((Arena*)nil, tc->scope_->grams, type);

	// Install type G, and also (G+?) if this is a root.
	declaretype(tc, stmt->line_, type->name, &type);
	if(!stmt->extends_){
		gram->wildroottype = typeunion(list2(&a, type, typedontcare()), gram);
		declaretype(tc, stmt->line_, nameprint("%s?", type->name), &gram->wildroottype);
	}

	if(FLAGS_modulargramdebug)
		fprint(2, "Added grammar: %s\n", stmt->name_);
	if(FLAGS_modulargramdebug > 1)
		fprint(2, "%G\n", gram);

out:
	aclear(&a);
	return type;
}

int
grammarfmt(Fmt *fmt)
{
	Gram *gram;
	CfgSym<AstToken> *s;
	CfgRule<AstToken> *r;
	CfgPrec<AstToken> *lo, *hi;
	char *sep;
	int i,k;

	gram = va_arg(fmt->args, Gram*);
	if(gram == nil)
		return fmtprint(fmt, "<nil>");
	// Header
	fmtprint(fmt, "grammar %s {\n", gram->name);
	// Ignored symbols
	if(gram->ignores){
		fmtprint(fmt, "  %%ignore ");
		sep = "";
		for(List<Name> *l=gram->ignores; l; l=l->next){
			fmtprint(fmt, "%s%s", sep, l->x);
			sep = ", ";
		}
		fmtprint(fmt, ";\n");
	}
	// Shared symbols
	{
		fmtprint(fmt, "  %%shared ");
		sep = "";
		Hashiter<CfgSym<AstToken>*,intptr> it;
		Hashkv<CfgSym<AstToken>*,intptr> kv;
		for(hashiterstart(&it, gram->sharedsyms); hashiternextkv(&it, &kv); ){
			s = kv.key;
			fmtprint(fmt, "%s%s", sep, s->name);
			sep = ", ";
		}
		hashiterend(&it);
		fmtprint(fmt, ";\n");
	}
	// Shared rules
	{
		fmtprint(fmt, "  %%shared ");
		sep = "";
		Hashiter<CfgRule<AstToken>*,intptr> it;
		Hashkv<CfgRule<AstToken>*,intptr> kv;
		for(hashiterstart(&it, gram->sharedrules); hashiternextkv(&it, &kv); ){
			r = kv.key;
			fmtprint(fmt, "%s%s.%s", sep, r->left->name, r->name);
			sep = ", ";
		}
		hashiterend(&it);
		fmtprint(fmt, ";\n");
	}
	// Grammar associativities
	fmtprint(fmt, "  %%left ");
	sep = "";
	for(i=0; i < gram->cfg.nprec; i++){
		hi = gram->cfg.prec[i];
		if(hi->assoc == CfgAssocLeft){
			fmtprint(fmt, "%s%s", sep, hi->name);
			sep = " ";
		}
	}
	fmtprint(fmt, ";\n");
	fmtprint(fmt, "  %%right ");
	sep = "";
	for(i=0; i < gram->cfg.nprec; i++){
		hi = gram->cfg.prec[i];
		if(hi->assoc == CfgAssocRight){
			fmtprint(fmt, "%s%s", sep, hi->name);
			sep = " ";
		}
	}
	fmtprint(fmt, ";\n");
	fmtprint(fmt, "  %%nonassoc ");
	sep = "";
	for(i=0; i < gram->cfg.nprec; i++){
		hi = gram->cfg.prec[i];
		if(hi->assoc == CfgAssocNone){
			fmtprint(fmt, "%s%s", sep, hi->name);
			sep = " ";
		}
	}
	fmtprint(fmt, ";\n");
	// Grammar precedence levels
	for(i=0; i < gram->cfg.nprec; i++){
		hi = gram->cfg.prec[i];
		for(List<CfgPrec<AstToken>*> *l=hi->belowme; l; l=l->next){
			lo = l->x;
			fmtprint(fmt, "  %%prec %s > %s;\n", hi->name, lo->name);
		}
	}
	// Grammar rules
	for(i=0; i < gram->cfg.nsym; i++){
		s = gram->cfg.sym[i];
		for(k=0; k < s->nrule; k++){
			r = s->rule[k];
			Type *t = gramsupertype(_cfgruletype(r));
			fmtprint(fmt, "  [%s] %R", t ? (char*)t->name : "", r);
			if(r->prec)
				fmtprint(fmt, "    [%s]", r->prec->name);
			fmtprint(fmt, "\n");
		}
	}
	fmtprint(fmt, "}");
	return 0;
}

// ========================================================
// Abstract syntax construction
//
// XXX comment here describing Ast structure and fields

// Return the kind of list this sym is: a star or a plus?
int
astlistkind(CfgSym<AstToken> *sym)
{
	int l;
	
	l = strlen(sym->name);
	if(l == 0)
		return 0;
	switch(sym->name[l-1]){
	case '+':
	case '*':
	case '@':
		return sym->name[l-1];
	default:
		return 0;
	}
}

int nmkast;
// Allocate an Ast with the given number of Ast* pointers
// on its right hand side.
static Ast*
mkast(CfgSym<AstToken> *sym, int nright)
{
	Ast *ast;
	Type *t;

	assert(sym);
	t = _cfgsymtype(sym);
	assert(sym->name);

	ast = (Ast*)emalloc(sizeof *ast + nright*sizeof astright(ast, 0));
	ast->_right = (Ast**)(ast+1);
	ast->nright = nright;
	ast->sym = sym;
	ast->symname = t ? t->name : sym->name;   // un-obfuscate
	ast->listkind = astlistkind(sym);
	ast->_parent = NoParentYet;
	if(debugfd)
		tracemkast(ast);
nmkast++;
	return ast;
}

// Create the AST for a given rule and right-hand symbols.
Ast*
mkastrule(CfgRule<AstToken> *r, Ast **right)
{
	Ast *ast;
	Type *t, *tgram, *tsym, *trule, *tt;
	int i, n;

	assert(r);

	if((ast = findsharedast(r, right)) != nil)
		return ast;
	n = r->nsavedright;
	ast = mkast(r->left, n);
	ast->op = AstRule;
	ast->rule = r;
	memmove(ast->_right, right, n * sizeof ast->_right[0]);
	for(i=0; i<n; i++){
		if(right[i] == nil)
			panic("mkastrule: restructure nil Ast");
		if(right[i]->hasslots)
			ast->hasslots = 1;
	}

	// Parsing this Ast requires the grammar that is the least upper
	// bound of this origin grammar for this rule and the required
	// grammar for all subtrees.  We need to save 'tsym' and 'trule'
	// because CfgSym<AstToken>::name and CfgRule<AstToken>::name may be obfuscated for
	// scoping reasons (see notes in grammartyping.c).
	trule = _cfgruletype(r);
	assert(trule);
	tsym  = trule->super;
	assert(tsym);
	tgram = tsym->super;
	assert(tgram);
	if(tgram->name == nameof("COutput")){
		fprint(2, "%s : %s\n", r->name, r->fmtname);
	}
	for(i=0; i<n; i++){
		if(right[i]->op == AstRule || right[i]->op == AstMerge){
			t = right[i]->_specifictype;
			assert(t);
			assert(t->super);
			// See notes in dat.h for struct Ast
			// (AstMerge has a type like G.sym."merge").
			t = t->super->super;
			assert(t);
			// Perf note: if there is a bottleneck, this is probably it.
			tt = joingrammartypes(tgram, t);
			if(tt == nil){
				fprint(2, "cannot join %T and %T\n", tgram, t);
				fprint(2, "%A\n", right[i]);
			}
			tgram = tt;
			assert(tgram);
		}
	}
	// Now that we have the specific grammar, narrow the type
	// to the rule used for this Ast.
	ast->_specifictype = lookuptag(lookuptag(tgram, tsym->name), trule->name);
	assert(ast->_specifictype);
	// Add to the shared memo table.
	memosharedast(ast);
	return ast;
}

// Create an Ast node representing the (ambiguous) merge of a and b.
Ast*
mkastmerge(Ast *a, Ast *b)
{
	Ast *ast;
	Type *tgram, *ta, *tb;

	assert(a->sym == b->sym);
	ast = mkast(a->sym, 2);
	ast->op = AstMerge;
	ast->_right[0] = a;
	ast->_right[1] = b;
	ast->zline = a->zline;
	ast->hasslots = (a->hasslots || b->hasslots);
	// Determine the most specific merge type.
	ta = gramsupertype(a->_specifictype);
	tb = gramsupertype(b->_specifictype);
	assert(ta && tb);
	tgram = joingrammartypes(ta, tb);
	assert(tgram);
	ast->_specifictype = lookuptag(lookuptag(tgram, a->symname), Name::merge);
	assert(ast->_specifictype);
	return ast;
}

// Create an Ast node representing a literal or regexp string.
Ast*
mkaststring(CfgSym<AstToken> *sym, Name text)
{
	static Hash<Name,Ast*> *hlitsyms;
	static Hash<Name,Hash<Name,Ast*>*> *hregsyms;

	// Each AstString has a unique Ast*.  The unique Ast*s
	// are hashed by symname.  Literals are directly hashed,
	// and regexps use a double hash.
	assert(sym);
	assert(sym->name);
	assert(sym->name[0]);
	assert(sym->name[1] == '.');
	assert(text);

	if(!hlitsyms)
		hlitsyms = mkhash<Name,Ast*>();
	if(!hregsyms)
		hregsyms = mkhash<Name,Hash<Name,Ast*>*>();

	Hash<Name,Ast*> *h = nil;
	int islit = (sym->name[0] == 'L');
	Ast *ast = nil;
	if(islit)
		ast = hashget(hlitsyms, sym->name);
	else{
		h = hashget(hregsyms, sym->name);
		if(h)
			ast = hashget(h, text);
	}
	if(ast){
		assert(ast->text == text);
		return ast;
	}

	ast = mkast(sym, 0);
	ast->op = AstString;
	ast->sym = sym;
	ast->symname = sym->name;   // "L.___" or "R.___"
	ast->text = text;
	ast->_specifictype = typeaststring();
	ast->isshared = 1;
	ast->_parent = nil;
	
	if(islit)
		hashput(hlitsyms, sym->name, ast);
	else{
		if(!h){
			h = mkhash<Name,Ast*>();
			hashput(hregsyms, sym->name, h);
		}
		hashput(h, text, ast);
	}
	return ast;
}

// FIXME: what if type = "G..any"?
// Create an Ast node representing a slot.  Type may be any
// kind of type (Grammar, Symbol, or Rule); using a Grammar
// type creates an ".any" slot.  The specifictype of the
// returned Ast will end in ".slot".
Ast*
mkastslot(Type *type, Name name)
{
	Ast *ast;

	assert(type);
	assert(type->cfgsym);

	ast = mkast(type->cfgsym, 0);
	ast->op = AstSlot;
	ast->text = name;
	ast->hasslots = 1;
	ast->_specifictype = type;
	return ast;
}

Ast*
mkastlistslot(Type *type, Name name)
{
	Ast *ast;

	ast = mkastslot(type, name);
	ast->op = AstListSlot;
	return ast;
}


// ========================================================
// Ast sharing
// This is user-enabled with the %share directive.
// Parents are meaningless for shared nodes.
// TODO: check all places parents are set
// TODO: check copy & rebuild

static Hash<CfgSym<AstToken>*, List<Ast*>*> *sharedsyms;
static Hash<CfgRule<AstToken>*, List<Ast*>*> *sharedrules;
static Hash<CfgSym<AstToken>*, Hash<Ast*,Ast*>*> *sharedoneright;

static void
allocsharedhashes(void)
{
	sharedsyms = mkhash<CfgSym<AstToken>*,List<Ast*>*>();
	sharedrules = mkhash<CfgRule<AstToken>*,List<Ast*>*>();
	sharedoneright = mkhash<CfgSym<AstToken>*, Hash<Ast*,Ast*>*>();
}

enum {
	ShareSym  = 1<<0,
	ShareRule = 1<<1,
};

static int
shouldshareast(CfgSym<AstToken> *sym, CfgRule<AstToken> *rule, Ast **right)
{
	Gram *gram;
	int i, sharing;

	// Check if this Ast should be shared.
	// Don't share when there are slots.
	assert(sym);
	assert(rule);

	// We get ast->_right directly, so don't do anything
	// with right[] which could be unsafe!
	// Simply checking for nil and hasslots is safe.
	for(i=0; i < rule->nsavedright; i++)
		if(!right[i] || right[i]->hasslots)
			return 0;

	gram = (Gram*)sym->gram;
	sharing = 0;
	if(hashget(gram->sharedsyms, sym))
		sharing |= ShareSym;
	if(hashget(gram->sharedrules, rule))
		sharing |= ShareRule;

	return sharing;
}

Ast*
findsharedast(CfgRule<AstToken> *r, Ast **right)
{
	Ast *ast;
	int i, sharing;

	// Check if this Ast should be shared.
	if((sharing = shouldshareast(r->left, r, right)) == 0)
		return nil;

	if(!sharedsyms)
		allocsharedhashes();

	// Search for an existing Ast.
	// Special case: one rhs element which is shared.
	if(r->nsavedright == 1 && right[0]->isshared){
		Hash<Ast*,Ast*> *h;
		if((h = hashget(sharedoneright, r->left)) == nil)
			return nil;
		return hashget(h, right[0]);
	}

	// General case: search a list (potentially slow!)
	List<Ast*> *all;
	if(sharing & ShareSym)
		all = hashget(sharedsyms, r->left);
	else
		all = hashget(sharedrules, r);

	for(List<Ast*> *l=all; l; l=l->next){
		ast = l->x;
		if(ast->rule != r)
			continue;
		for(i=0; i < r->nsavedright; i++)
			if(!samestructure(astright(ast, i), right[i]))
				goto Next;
		assert(ast->isshared);
		return ast;
	Next:;
	}

	// Not found.
	return nil;
}

void
memosharedast(Ast *ast)
{
	Ast *r;
	int sharing;

	// Check if this Ast should be shared.
	// OK to pass ast->_right[] because we only check 'hasslots'.
	if((sharing = shouldshareast(ast->sym, ast->rule, ast->_right)) == 0)
		return;

	if(!sharedsyms)
		allocsharedhashes();

	// Memo
	ast->isshared = 1;
	ast->_parent = nil;

	// Special case: one rhs element which is shared.
	if(ast->nright == 1){
		r = astright(ast, 0);
		if(r && r->isshared){
			Hash<Ast*,Ast*> *h;
			h = hashget(sharedoneright, ast->sym);
			if(h == nil){
				h = mkhash<Ast*,Ast*>();
				hashput(sharedoneright, ast->sym, h);
			}
			hashput(h, r, ast);
		}
	}

	// General case: add to the list.
	if(sharing & ShareSym){
		List<Ast*> *l = hashget(sharedsyms, ast->sym);
		l = listinsert((Arena*)nil, l, ast);
		hashput(sharedsyms, ast->sym, l);
	}
	else{
		List<Ast*> *l = hashget(sharedrules, ast->rule);
		l = listinsert((Arena*)nil, l, ast);
		hashput(sharedrules, ast->rule, l);
	}
}


// ========================================================
// String conversions

static void _asttoa(Ast*, Fmt*);
static void _asttos(Ast*, Fmt*, int, int);

// Convert Ast to a string.  Returns a Name pointer.
Name
asttoa(Ast *ast)
{
	char *s;
	Name t;
	Fmt fmt;
	int len;

	fmtstrinit(&fmt);
	_asttoa(ast, &fmt);
	s = fmtstrflush(&fmt);
	if(s[0]){
		len = strlen(s);
		if(s[len-1] == ' ')
			s[len-1] = 0;
	}
	t = nameof(s);
	free(s);
	return t;
}

// Convert Ast to a string, writing to fmt.
// Just emits the string representation of all the leaves,
// separated by spaces.
static void
_asttoa(Ast *ast, Fmt *fmt)
{
	int i, j;
	const char *slot;
	CfgSym<AstToken> *sym;

	if(ast == nil || ast == NoParentYet)
		return;
	switch(ast->op){
	default:
		fmtprint(fmt, "[op.%#ux]", ast->op);
		break;
	case AstString:
		fmtstrcpy(fmt, ast->text);
		fmtrune(fmt, ' ');
		break;
	case AstRule:
		j = 0;
		for(i=0; i<ast->rule->nright; i++){
			sym = ast->rule->right[i];
			if(strncmp("L.", sym->name, 2) == 0){
				fmtstrcpy(fmt, sym->name+2);
				fmtrune(fmt, ' ');
			}else
				_asttoa(astright(ast, j++), fmt);
		}
		assert(j == ast->nright);
		break;
	case AstSlot:
	case AstListSlot:
		slot = (ast->op == AstSlot) ? "slot" : "listslot";
		if(ast->text && ast->text[0])
			fmtprint(fmt, "\\%s::%s.%s", ast->text, ast->symname, slot);
		else
			fmtprint(fmt, "\\_::%s.%s", ast->symname, slot);
		break;
	case AstMerge:
		fmtstrcpy(fmt, "[M| ");
		_asttoa(astright(ast, 0), fmt);
		fmtstrcpy(fmt, "]");
		break;
	}	
}

// Convert Ast to single-line S-expression notation.
char*
asttos(Ast *ast)
{
	char *s;
	Fmt fmt;

	fmtstrinit(&fmt);
	_asttos(ast, &fmt, 0, 0);
	s = fmtstrflush(&fmt)+1;
	if(s[0])
		s++;
	return s;
}
	
// Convert Ast to an S-expression (Lisp-like) representation,
// writing to fmt.  If newlines != 0, insert line breaks for each
// new complex clause, indented by indent levels.  All output
// starts with a leading space.
static void _asttoslist(Ast*, Fmt*, int, int);
static char dots[100];
void
_asttos(Ast *ast, Fmt *fmt, int newlines, int indent)
{
	int i;
	Ast *r;
	const char *sep;
	
	// Initialize "infinite" stream of . . . . .
	if(dots[0] == 0){
		for(i=0; i<nelem(dots)-1; i++)
			dots[i] = ". "[i%2];
	}

	if(ast == nil || ast == NoParentYet){
		fmtprint(fmt, " nil");
		return;
	}
	switch(ast->op){
	default:
		fmtprint(fmt, " op.%#ux", ast->op);
		break;
	case AstString:
		fmtprint(fmt, " '%s", ast->text);
		break;
	case AstRule:
		fmtprint(fmt, " (%R", ast->rule);
		for(i=0; i<ast->nright; i++){
			r = astright(ast, i);
			// special case for printing lists
			if(r->op == AstRule && r->sym->name[strlen(ast->sym->name)-1] == '@')
				_asttoslist(r, fmt, newlines, indent);
			else{
				if(newlines)
					fmtprint(fmt, "\n%.*s", indent*2, dots);
				_asttos(r, fmt, newlines, indent+1);
			}
		}
		fmtprint(fmt, ")");
		break;
	case AstSlot:
		sep = ast->text ? "(" : "";
		if(ast->sym == nil)
			fmtprint(fmt, " %s!nil.slot", sep);
		else
			fmtprint(fmt, " %s%s.slot", sep, ast->sym->name);
		if(ast->text)
			fmtprint(fmt, " '%s)", ast->text);
		break;
	case AstListSlot:
		sep = ast->text ? "(" : "";
		if(ast->sym == nil)
			fmtprint(fmt, " %s!nil.listslot", sep);
		else
			fmtprint(fmt, " %s%s.listslot", sep, ast->sym->name);
		if(ast->text)
			fmtprint(fmt, " '%s)", ast->text);
		break;
	case AstMerge:
		if(ast->sym)
			fmtprint(fmt, " (%s.merge", ast->sym->name);
		else
			fmtstrcpy(fmt, "(merge");
		for(i=0; i<2; i++){
			if(newlines)
				fmtprint(fmt, "\n%.*s", indent*2, dots);
			_asttos(astright(ast, i), fmt, newlines, indent+1);
		}
		fmtstrcpy(fmt, ")");
		break;
	}	
}

// Format body of list into fmt.
static void
_asttoslist(Ast *ast, Fmt *fmt, int newlines, int indent)
{
	if(ast->nright > 1)
		_asttoslist(astright(ast, 0), fmt, newlines, indent);
	if(ast->nright > 0){
		if(newlines)
			fmtprint(fmt, "\n%.*s", indent*2, dots);
		_asttos(astright(ast, ast->nright-1), fmt, newlines, indent+1);
	}
}

// %A - format Ast as infix string (leaves separated by spaces)
// %#A - format Ast as one-line S-expression.
// %#lA - format Ast as multi-line S-expression.
int
astfmt(Fmt *fmt)
{
	Ast *ast;
	
	ast = va_arg(fmt->args, Ast*);
	if(fmt->flags&FmtSharp)
		_asttos(ast, fmt, fmt->flags&FmtLong, 0);
	else
		_asttoa(ast, fmt);
	return 0;
}


// ========================================================
// Ast lists

static Ast *_mkastlist(CfgSym<AstToken>*, List<Ast*>*, Ast*);

// Given a list of Ast, and a foo*+ template, create a
// (foo*+ foo@) containing it.
Ast*
mkastlist(Ast *tmplate, List<Ast*> *l)
{
	CfgSym<AstToken> *root, *sym;
	CfgRule<AstToken> *rule;
	Ast *right[1];
	int kind;

	root = tmplate->sym;
	kind = astlistkind(root);
	assert(kind);

	if(l == nil){
		// Up to the caller to check emptiness.
		// Empty rule is rules[2] (matches the comments in grammar.h).
		assert(kind == '*');
		assert(root->rule[2]->nright == 0);
		return mkastrule(root->rule[2], nil);
	}

	// Non-empty is rule[1] (same as grammar.h, note that .any is rule[0]).
	rule = root->rule[1];
	assert(rule->nright == 1);

	sym = rule->right[0];
	assert(sym->name[strlen(sym->name)-1] == '@');

	right[0] = _mkastlist(sym, l, nil);
	return mkastrule(rule, right);
}

// Construct a tree of foo@ lists (where listsym is a foo@ sym).
static Ast*
_mkastlist(CfgSym<AstToken> *listsym, List<Ast*> *l, Ast *prev)
{
	CfgSym<AstToken> *one;
	CfgRule<AstToken> *rule;
	Ast *a, *b, *right[3];
	
	if(l == nil)
		return prev;

	// Matching grammar.h,
	//  rule[0] = .any
	//  rule[1] = one
	//  rule[2] = one one (or: one sep one)
	assert(listsym->rule[1]->nright == 1);
	assert(listsym->rule[2]->nright >= 2);
	one = listsym->rule[1]->right[0];

	b = l->x;
	if(!b)
		panic("nil ast in list restructure");
	if(b->sym->name != one->name)
		panic("restructuring [Ast.%s] into a [list Ast.%s]", b->sym->name, one->name);

	if(prev == nil){
		right[0] = b;
		a = mkastrule(listsym->rule[1], right);
	}else{
		right[0] = prev;
		rule = listsym->rule[2];
		right[1] = b;
		a = mkastrule(rule, right);
	}
	return _mkastlist(listsym, l->next, a);
}

// Is name a list?  (Does it end in * or + or @)
static int
islistname(char *name)
{
	int l;
	
	l = strlen(name);
	return l > 1 && (name[l-1] == '+' || name[l-1] == '*' || name[l-1] == '@');
}

List<Ast*>*
asttolist(Ast *ast)
{
	List<Ast*> *l;

	if(ast == nil)
		return nil;
	if(!islistname(ast->sym->name))
		panic("asttolist called on %s", ast->sym->name);
	
	// empty?
	if(ast->nright == 0)
		return nil;

	if(ast->listkind != '@'){
		// get list@
		if(ast->nright != 1){
			if(ast->op == AstMerge)
				panic("ambiguous list:\n%#lA", ast);
			else
				panic("bad list:\n%#lA", ast);
		}
		ast = astright(ast, 0);
	}

	// walk list@
	l = ZList;
	while(ast->nright > 1){
		l = listinsert((Arena*)nil, l, astright(ast, ast->nright-1));
		ast = astright(ast, 0);
	}
	if(ast->nright > 0)
		l = listinsert((Arena*)nil, l, astright(ast, 0));
	// XXX could cache the result
	return l;
}

ListNode*
asttolistnode(Ast *ast)
{
	List<Ast*> *l;
	ListNode *zl;

	l = asttolist(ast);
	zl = listtolistnode((List<void*>*)l);
	listfree(l);
	return zl;
}

// ========================================================
// Prev/Next, kids, and parents

static inline int
usefulkid(Ast *kid)
{
	// When looking for a prev/next or last kid, ignore these Asts.
	// Note that 'op == AstString' implied 'isshared'
	return kid && !kid->isshared;
}

void
astprevnext(Ast *ast, Ast **prev, Ast **next)
{
	Ast *p, *a;
	int i, k;

	assert(ast);

	// Need ast parent to determine context.
	if(prev)
		*prev = nil;
	if(next)
		*next = nil;
	p = astparent(ast);
	if(p == nil)
		return;

	// Ugly list special case.
	if(p->listkind == '@'){
		if(!usefulkid(ast))
			return;
		// p->kids is one of:
		//   (ast)
		//   (foo@ ast)
		//   (foo@ sep ast)
		// Go down into 'foo@' to get prev.
		if(prev){
			switch(p->nright){
			case 2:
			case 3:
				a = astright(p, 0);
				assert(a);
				assert(a->symname == p->symname);
				assert(1 <= a->nright && a->nright <= 3);
				*prev = astright(a, a->nright - 1);
				break;
			default:
				assert(p->nright == 1);
			}
		}
		// Go up from 'p' to get next.
		if(next){
			a = astparent(p);
			if(a == nil)
				return;
			if(a->symname == p->symname){
				assert(1 <= a->nright && a->nright <= 3);
				*next = astright(a, a->nright-1);
			}
			else
				assert(astlistkind(a->sym));
		}
		return;
	}

	// Normal tree structure.  Skip over string nodes.
	for(i=0; i < p->nright; i++)
		if(astright(p, i) == ast)
			break;
	if(i == p->nright)
		return;
	if(prev){
		for(k=i-1; k >= 0; k--){
			if(usefulkid(astright(p, k))){
				*prev = astright(p, k);
				break;
			}
		}
	}
	if(next){
		for(k=i+1; k < p->nright; k++){
			if(usefulkid(astright(p, k))){
				*next = astright(p, k);
				break;
			}
		}
	}
}

Ast*
astlastkid(Ast *ast)
{
	Ast *k;
	int i;

	if(!ast)
		return nil;

	// Ugly list special case.
	if(astlistkind(ast->sym)){
		if(ast->nright == 0)
			return nil;
		assert(ast->nright == 1);
		k = astright(ast, 0);
		// k should be a foo@ sym.  The last item of a
		// list is at the top of the foo@ parse tree.
		assert(k->listkind == '@');
		switch(k->nright){
		case 1:
			k = astright(k, 0);
			break;
		case 2:
			k = astright(k, 1);
			break;
		case 3:
			k = astright(k, 2);
			break;
		default:
			panic("unexpected ast in astlastkid:\n%#A\n", ast);
			k = nil;
		}
		if(usefulkid(k))
			return k;
		// All list elements should have the same type, so
		// if this kid isn't useful then none of them are.
		return nil;
	}

	// Normal tree structure.
	for(i=ast->nright-1; i >= 0; i--)
		if(usefulkid(astright(ast, i)))
			return astright(ast, i);
	return nil;
}

Ast*
astrealparent(Ast *ast)
{
	Ast *p;
	Name name;

	if(ast == nil || (p=astparent(ast)) == nil)
		return nil;

	assert(ast->listkind != '@');

	// We want ast.parent to be the foo*/+ list for list elements,
	// but to maintain the integrity of copy-on-ref, ast->_parent
	// must be it's real owner (a foo@ Ast).  This function gets
	// the real foo*/+ parent for list elements.
	if(p->listkind != '@')
		return p;

	name = p->symname;
	for(;;){
		p = astparent(p);
		if(p == nil)
			return nil;
		if(p->symname == name)
			continue;
		if(p->listkind == '+' || p->listkind == '*')
			return p;
		// Shouldn't get here.  A list elements should have
		// a path of foo@ parents ended with nil or foo*/+.
		assert(0);
	}
}

Ast*
asttypedparent(Ast *ast, Type *type)
{
	assert(type);
	assert(type->op == TypePtr);
	type = type->left;
	// Find the first parent of 'ast' that is a <: of 'type'.
	if(!ast)
		return nil;
	while((ast = astparent(ast)) != nil)
		if(issubtypesyntax(type, ast->_specifictype))
			return ast;
	return nil;
}

Ast*
astright(Ast *ast, int index)
{
	Ast *k;
	static Name typerep;
	
	if(typerep == nil)
		typerep = nameof("TypeRep");

	if(ast == nil || index < 0 || index >= ast->nright)
		return nil;
	k = ast->_right[index];
	if(k == nil)
		return nil;
	if(k->_parent == ast || k->isshared)
		return k;
	if(k->symname == typerep)  // ick
		return k;
	// XXX This is not correct:
	// if you substitute a new tree into `expr{\a+\b}
	// in both places, then the first will have NoParentYet
	// changed to the parent, and the second will
	// look okay since it has the same parent.
	// really we need to record the address of the
	// pointer pointing at k.
	if(k->_parent == NoParentYet){
		k->_parent = ast;
		return k;
	}
	k = astcopy(k, 0);
	k->_parent = ast;
	ast->_right[index] = k;
	return k;
}

Ast*
astparent(Ast *ast)
{
	if(ast == nil){
		static int warned;
		if(warned++ == 0)
			fprint(2, "warning: astparent(nil)\n");
		return nil;
	}
	
	if(ast->_parent == NoParentYet){
		// Oops - someone looked!
		ast->_parent = nil;
	}
	return ast->_parent;
}

