#include "grammar.h"

DEFINE_bool(canonicaldebug, false, "Debug canonicalization.");

static Name dontcare;

// README
// There used to be a comment here which said that we must check
// names, not CfgSym/CfgRule pointers, due to modular grammars.
// This is no longer true.  Each grammar family has a single
// unified CFG, so it's correct to say things like a->sym == b->sym
// and a->rule == b->rule.  In fact, sym->name and rule->name are
// now obfuscated due to modular grammars, so it is WRONG to check
// name equality.

// ========================================================
// Structure matching

// Do a and b have the same Ast structure?
int
samestructure(Ast *a, Ast *b)
{
	int i;

	if(a == b)
		return 1;
	if(a == nil || b == nil || a->op != b->op || a->isshared || b->isshared)
		return 0;

	switch(a->op){
	case AstString:
		return 0;	// otherwise a == b

	case AstRule:
		if(a->rule != b->rule)
			return 0;
		goto case_Struct;

	case AstMerge:
		if(a->sym != b->sym)
			return 0;
	case_Struct:
		for(i=0; i<a->nright; i++)
			if(!samestructure(astright(a, i), astright(b, i)))
				return 0;
		return 1;

	case AstSlot:
	case AstListSlot:
		return 1;
	}

	panic("samestructure %d", a->op);
	return 0;
}

// ========================================================
// Destructuring and restructuring

static int _destructure(List<Var*>*, Ast*, Ast*, void***);
static Ast* __restructure(Ast*, void ***, Ast* (*)(Ast*,void*), void*);

// Destructure ast against tmplate.  
// Template is known to have n slots. 
// On failure, return 0.
// On success, return 1 and fill in slots[0..n-1] with
// appropriate pointers to the substructure from Ast.
int
destructure(List<Var*> *canonicals, Ast *tmplate, Ast *ast, int n, void ***slots)
{
	int i;
	void **tslots, **p;

	if(!dontcare)
		dontcare = nameof("_");

	if(FLAGS_canonicaldebug && canonicals){
		fprint(2, "==canonicals:");
		for(List<Var*> *l=canonicals; l; l=l->next)
			fprint(2, " %s(%T)", l->x->name, l->x->host);
		fprint(2, "\n");
	}

	tslots = (void**)emalloc(n*sizeof tslots[0]);
	p = tslots;
	if(!_destructure(canonicals, tmplate, ast, &p)){
		free(tslots);
		return 0;
	}
	// Now that we know it worked, write to the actual slots.
	assert(p == tslots+n);
	for(i=0; i<n; i++)
		*slots[i] = tslots[i];
	free(tslots);
	return 1;
}

static Ast*
canonicalize1(List<Var*> *canonicals, Ast *ast)
{
	for(List<Var*> *l=canonicals; l; l=l->next){
		Var *v = l->x;
		if(issubtypesyntax(v->type->left, ast->_specifictype)){
			Ast *b = *(Ast**)computeattr(ast, v);
			if(b != ast)
				return b;
		}
	}
	return ast;
}

static Ast*
canonicalize(List<Var*> *canonicals, Ast *ast)
{
	for(int i=0; i<20; i++){
		Ast *b = canonicalize1(canonicals, ast);
		if(b == ast)
			return ast;
		ast = b;
	}
	panic("canonicalization loop");
}

// Destructure ast against tmplate.  
// *slotp is the next slot pointer: if a slot is encountered,
// write it to **slotp and increment *slotp.
// If slotp == nil, don't fill it in.
static int
_destructure(List<Var*> *canonicals, Ast *tmplate, Ast *ast, void ***slotp)
{
	int i;

	if(tmplate == ast)
		return 1;
	if(tmplate == nil || ast == nil)
		return 0;
	if(tmplate->isshared)
		return 0;

	Ast *otmplate = tmplate;
	Ast *oast = ast;

	if(canonicals){
		tmplate = canonicalize(canonicals, otmplate);
		ast = canonicalize(canonicals, oast);

		if(FLAGS_canonicaldebug)
			fprint(2, "==canonicalize\nA: %#A\n   %#A\nT: %#A\n   %#A\n",
				oast, ast, otmplate, tmplate);
	}

	switch(tmplate->op){
	default:
		panic("bad destructure tmplate op %d", tmplate->op);
		return 0;

	case AstString:
		return 0;	// otherwise a == b

	case AstRule:
	case AstMerge:
		if(ast->op != tmplate->op)
			return 0;
		if(ast->rule != tmplate->rule)
			return 0;
		for(i=0; i<ast->nright; i++)
			if(!_destructure(canonicals, astright(tmplate, i), astright(ast, i), slotp))
				return 0;
		return 1;

	case AstSlot:
		if(tmplate->sym && tmplate->sym != ast->sym){
			// This should not happen, since the rules matched
			// in the caller, so all the syms should match
			// during the iteration.
			panic("_destructure slot type: %#A vs %#A", tmplate, ast);
			return 0;
		}
		if(slotp && tmplate->text != dontcare)
			*(*slotp)++ = ast;
		return 1;

	case AstListSlot:
		if(tmplate->sym && tmplate->sym != ast->sym){
			// This should not happen, since the rules matched
			// in the caller, so all the syms should match
			// during the iteration.
			panic("_destructure list slot type: %#A vs %#A", tmplate, ast);
			return 0;
		}
		if(slotp && tmplate->text != dontcare)
			*(*slotp)++ = asttolistnode(ast);
		return 1;
	}
}

// Restructure: insert slots[0..n-1] into tmplate
// to create new Ast, and return it.
Ast*
_restructure(Ast *tmplate, int n, void **slots, Ast* (*xform)(Ast*,void*), void *arg)
{
	void **p;
	Ast *ast;
	
	p = slots;
	ast = __restructure(tmplate, &p, xform, arg);
	assert(p == slots+n);
	return ast;
}

// Restructure: copy tmplate, inserting slots, and return result.
// Next slot to use is **slotp.  After use, increment *slotp.
static Ast*
__restructure(Ast *tmplate, void ***slotp, Ast* (*xform)(Ast*,void*), void *arg)
{
	int i;
	Ast *a, **right;
	List<Ast*> *l;

	if(tmplate == nil || tmplate->isshared)
		return tmplate;

	switch(tmplate->op){
	default:
		panic("bad restructure tmplate op %d", tmplate->op);
		return nil;

	case AstString:
		a = tmplate;
		break;

	case AstRule:
		if(xform && !tmplate->hasslots){
			a = xform(tmplate, arg);
			break;
		}

		// Just copy the tmplate.
		right = (Ast**)emalloc(tmplate->nright * sizeof right[0]);
		for(i=0; i<tmplate->nright; i++)
			right[i] = __restructure(astright(tmplate, i), slotp, xform, arg);
		a = mkastrule(tmplate->rule, right);
		a->zline = tmplate->zline;
		free(right);

		// Should add an ugly case for "expr?"

		// Apply the user-specified transformation, if any.
		if(xform)
			a = xform(a, arg);
		break;

	case AstSlot:
		a = (Ast*)*(*slotp)++;
		break;

	case AstListSlot:
		l = (List<Ast*>*)listnodetolist(nil, (ListNode*)*(*slotp)++);

		// Check emptiness constraints.
		switch(astlistkind(tmplate->sym)){
		case '+':
			if(l == ZList)
				panic("%s: %d: restructure %s with empty list",
				      tmplate->zline.file ? (char*)tmplate->zline.file : "???", tmplate->zline.line,
				      tmplate->sym->name);
		case '*':
			break;
		default:
			panic("bad list slot: %s\n", tmplate->sym->name);
		}

		// Convert the list to an Ast.
		a = mkastlist(tmplate, l);
		a->zline = tmplate->zline;
		listfree(l);
		break;

	}
	return a;
}

// FIXME  ??
// Create an Ast representing a slot.  The only time
// it makes sense to construct a slot is in a pattern
// translator (where the translated Ast was itself a
// slot).  We link to that original slot so that we
// can check if the slot ordering has changed, and if
// so account for that.
Ast*
restructslot(Type *type, Ast *link)
{
	Ast *a;

	assert(type);
	assert(type->op == TypeSyntax || type->op == TypeUnion);
	assert(link);

	if(type->name != Name::slot){
		type = lookuptag(type, Name::slot, nil);
		assert(type);
	}
	if(astlistkind(symsupertype(type)->cfgsym))
		a = mkastlistslot(type, link->text);
	else
		a = mkastslot(type, link->text);
	a->sep = link;
	return a;
}

// ========================================================
// Ast join, split, and copy

// Return array of Asts pointed at by this Ast.
// Exposed to zeta programs as astsplit.
Array*
astsplit(Ast *ast)
{
	Array *a;
	int sz, n, i;
	static Array *zarray;
	Ast **kids;
	
	sz = sizeof astright(ast, 0);
	if(zarray == nil)
		zarray = newarray(0, sz);
	
	if(ast == nil || ast->nright == 0)
		return zarray;

	if(astlistkind(ast->sym)){
		// Copy the embedded list for list nodes.
		List<Ast*> *lkids = asttolist(ast);
		n = listlen(lkids);
		a = newarray(n, sz);
		kids = (Ast**)a->v;
		List<Ast*> *l;
		for(l=lkids, i=0; l; l=l->next, i++)
			kids[i] = l->x;
		listfree(lkids);
	}else{
		// Copy right[] for other nodes.
		a = newarray(ast->nright, sz);
		kids = (Ast**)a->v;
		for(i=0; i<ast->nright; i++)
			kids[i] = astright(ast, i);
	}
	return a;
}

// Rebuild an ast node with the given kids.  If this is a
// list, we need to expand to the full list structure.
// FIXME: is parent set correctly for lists?
Ast*
astjoin(Ast *old, int n, Ast **kids)
{
	Ast *ast;
	List<Ast*> *l;
	int i;
	
	if(old->rule == nil){
		assert(n == 0);
		return old;
	}

	if(astlistkind(old->sym)){
		l = ZList;
		if(n){
			for(i=n-1; i>=0; i--)
				l = listinsert(nil, l, kids[i]);
		}
		ast = mkastlist(old, l);
		listfree(l);
	}
	else{
		if(n != old->nright){
			fprint(2, "astjoin: wrong argument count: want %d got %d\n",
				old->nright, n);
		Dead:
			dumpstack(running);
			panic("astjoin");
		}
		for(i=0; i<n; i++){
			if(astright(old, i)->sym != kids[i]->sym){
				fprint(2, "astjoin: sym mismatch: got %s want %s\n",
					astright(old, i)->sym->name, kids[i]->sym->name);
				goto Dead;
			}
		}
		ast = mkastrule(old->rule, kids);
	}

	ast->zline = old->zline;
	ast->copiedfrom = old;

	// FIXME: is this what we want?
	copyxstructs(ast->extensions, old->extensions,
		old->_specifictype, old->_specifictype);
	return ast;
}

Ast*
astcopy(Ast *tmplate, int deep)
{
	Ast *a, **right;
	int i;

	if(tmplate == nil || tmplate->isshared)
		return tmplate;

	switch(tmplate->op){	
	case AstString:
		return tmplate;

	case AstRule:
		right = (Ast**)emalloc(tmplate->nright * sizeof right[0]);
		for(i=0; i<tmplate->nright; i++){
			right[i] = astright(tmplate, i);
			if(deep)
				right[i] = astcopy(right[i], 1);
		}
		a = mkastrule(tmplate->rule, right);
		a->copiedfrom = tmplate;
		a->zline = tmplate->zline;
		free(right);
		return a;

	case AstSlot:
		a = mkastslot(tmplate->_specifictype, tmplate->text);
		a->sep = tmplate->sep;
		a->copiedfrom = tmplate;
		a->zline = tmplate->zline;
		return a;

	case AstListSlot:
		a = mkastlistslot(tmplate->_specifictype, tmplate->text);
		a->sep = tmplate->sep;
		a->copiedfrom = tmplate;
		a->zline = tmplate->zline;
		return a;

	case AstMerge:
		if(deep)
			return mkastmerge(astcopy(astright(tmplate, 0), 1),
			                  astcopy(astright(tmplate, 1), 1));
		else
			return mkastmerge(astright(tmplate, 0), astright(tmplate, 1));
	}

	panic("astcopy op %d", tmplate->op);
	return nil;
}

Ast*
astdeepcopy(Ast *tmplate)
{
	return astcopy(tmplate, 1);
}

void
astremovecopiedfrom(Ast *ast)
{
	int i;

	ast->copiedfrom = nil;
	for(i=0; i<ast->nright; i++)
		astremovecopiedfrom(astright(ast, i));
}


// ========================================================
// Gather all slots in the Ast

void
astcollectslots(Ast *ast, int *n, Ast ***slotp)
{
	int i;

	if(ast == nil)
		return;

	switch(ast->op){
	default:
		panic("bad collectslots op %d", ast->op);
		return;

	case AstString:
		return;

	case AstRule:
	case AstMerge:
		for(i=0; i<ast->nright; i++)
			astcollectslots(astright(ast, i), n, slotp);
		return;

	case AstSlot:
	case AstListSlot:
		if(strcmp(ast->text, "_") == 0)
			return;
		assert(*n > 0);
		(*n)--;
		*(*slotp)++ = ast;
		return;
	}
}

