// Table-based lexer.

#include <stdio.h>
#include <ctype.h>
#include "a.h"
extern "C" {
#include "y.tab.h"
}
#include "nfa.h"

int ingrammar;
int inbacktick;
int intermname;

Name *files;
int nfiles;

static void
addfile(char *file)
{
	int i;
	
	for(i=0; i<nfiles; i++)
		if(strcmp(file, files[i]) == 0)
			return;
	files = (Name*)erealloc(files, (nfiles+1)*sizeof files[0]);
	files[nfiles++] = nameof(file);
}

int
file2num(char *file)
{
	int i;
	
	if(file == nil)
		return -1;

	for(i=0; i<nfiles; i++)
		if(strcmp(file, files[i]) == 0)
			return i;
	addfile(file);
	return nfiles-1;
}

Name
num2file(int n)
{
	if(0 <= n && n < nfiles)
		return files[n];
	return nameof("???");
}

#define isident(c) ((c) >= 0x80 || isalpha(c) || isdigit(c) || (c) == '_')

typedef struct Inbuf Inbuf;
struct Inbuf {
	char *bp;
	char *wp;
	char *ep;
	char *p;
	Inbuf *next;
	FILE *fin;
	Line curpos;
	Line yyline;
	char *dir;
};
Inbuf *inbuf;
Inbuf *paused;

typedef struct Lexsym Lexsym;
struct Lexsym
{
	char	*str;
	int	token;
};

Lexsym symlist[] =
{
#include "y.syms"
};

static int
yygetc(void)
{
	int c, n;
	char *obp;

	assert(inbuf->curpos.byte == inbuf->p - inbuf->bp);
	if(inbuf->p == inbuf->wp){
		if(inbuf->fin == nil)
			return EOF;
		c = getc(inbuf->fin);
		if(c == EOF)
			return EOF;
		if(inbuf->wp == inbuf->ep){
			n = inbuf->ep - inbuf->bp;
			if(n < 1024)
				n = 1024;
			else
				n *= 2;
			obp = inbuf->bp;
			inbuf->bp = (char*)erealloc(inbuf->bp, n);
			inbuf->p += inbuf->bp - obp;
			inbuf->wp += inbuf->bp - obp;
			inbuf->ep = inbuf->bp + n;
		}
		*inbuf->wp++ = c;
	}
	c = *inbuf->p++;
	inbuf->curpos.byte++;
	assert(inbuf->curpos.byte == inbuf->p - inbuf->bp);
	if(c == '\n')
		inbuf->curpos.line++;
	return c;
}

static void
yyungetc(int c)
{
	if(c == EOF)
		return;
	assert(inbuf->p > inbuf->bp);
	inbuf->curpos.byte--;
	inbuf->p--;
	if(c == '\n')
		inbuf->curpos.line--;
}

static Lexsym*
symsearch(char *str)
{
	int i, lo, hi, m;

	lo = 0;
	hi = nelem(symlist);
	while(lo < hi){
		m = (lo+hi)/2;
		i = strcmp(str, symlist[m].str);
		if(i < 0)
			hi = m;
		else if(i > 0)
			lo = m+1;
		else
			return &symlist[m];
	}
	if(lo < nelem(symlist) && strcmp(symlist[lo].str, str) == 0)
		return &symlist[lo];
	return nil;
}

static int
bufappend(char **p, int c, char *buflast)
{
	if(*p >= buflast){
		**p = 0;
		error(inbuf->curpos, "token is too long");
		return 0;
	}
	**p = c;
	(*p)++;
	return 1;
}

static int yylexcomment(void);
static int yylexstring(int c, char*, char*, char*, char**);
static int yylexname(int c, char*, char*, char**);
static int yylextermname(int c, char*, char*, char**);
static int yylexbacktick(int c, Yystype*);

static int
_yylex(Yystype *yylval)
{
	int c, c1;
	char sbuf[300], *p, *sbufend;
	Lexsym* sym, *nsym;
	int sol;
	Type *t;
	long double f;
	uint64 v;
	
	sbufend = sbuf + sizeof sbuf - 1;
	sol = 0;
	if(inbuf->curpos.byte == 0)
		sol = 1;

	while((c = yygetc()) == ' ' || c == '\t' || c == '\n')
		if(c == '\n')
			sol = 1;

	inbuf->yyline = inbuf->curpos;
	inbuf->yyline.byte--;
	
	if(inbacktick && c != EOF && (!sol || c != '#'))
		return yylexbacktick(c, yylval);

	switch(c){
	case EOF:
		return Eof;

	case '#':
		if(!sol)
			goto Default;
		if((c = yygetc()) == '!'){	// skip #! line
			while((c = yygetc()) != '\n' && c != EOF)
				;
			return _yylex(yylval);
		}
		yyungetc(c);
		c = '#';
		goto Default;

	case '.':
		c1 = yygetc();
		yyungetc(c1);
		if(isdigit(c1))
			goto Number;
		goto Default;

	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	Number:
		/*
		 * Parse pp numbers - pattern is [0-9.]([0-9.]|ident|[eEpP][+-])*
		 */
		p = sbuf;
		for(;;){
			if(p >= sbuf + sizeof sbuf - 1){
				*p = 0;
				error(inbuf->curpos, "number too long");
				return Error;
			}
			*p++ = c;
			c1 = c;
			c = yygetc();
			if(c == EOF)
				return Eof;
			if(isdigit(c) || c == '.' || isident(c))
				continue;
			if(p == sbuf+1 && sbuf[0] == '0' && (c == 'x' || c == 'X'))
				continue;
			if((c == '+' || c == '-') && (c1 == 'p' || c1 == 'e' || c1 == 'P' || c1 == 'E'))
				continue;
			yyungetc(c);
			break;
		}
		*p = 0;
		yylval->name = nameof(sbuf);
		yylval->len = strlen(sbuf);
		t = nil;
		f = 0;
		v = 0;
		if(parsefloat(inbuf->yyline, sbuf, &f, &t) >= 0){
			yylval->f = f;
			yylval->type = t;
			return LFloat;
		}else if(parseinteger(inbuf->yyline, sbuf, &v, &t) >= 0){
			yylval->i = v;
			yylval->type = t;
			return LInteger;
		}
		error(inbuf->yyline, "malformed numeric constant: %s", sbuf);
		yylval->i = 1;
		yylval->type = mkTypeInt();
		return LInteger;

	case '\"':
		if(!yylexstring(c, "string", sbuf, sbufend, &p))
			return Error;
		*p = 0;
		yylval->name = nameof(sbuf);
		yylval->len = strlen(sbuf);
		return LString;

	case '%':
		if(ingrammar)
			goto Ident;
		goto Default;

	case '/':
		if(yylexcomment())
			return _yylex(yylval);

		if(!ingrammar)
			goto Default;
		// Parse regular expression /foo/[as]
		if(!yylexstring('/', "regexp", sbuf, sbufend, &p))
			return Error;
		*p = 0;
		yylval->name = nameof(sbuf);
		yylval->len = strlen(sbuf);
		yylval->flags = 0;
		while((c = yygetc()) == 'a' || c == 's'){
			if(c == 'a')
				yylval->flags |= RegexpAmbiguous;
			if(c == 's')
				yylval->flags |= RegexpShortest;
		}
		yyungetc(c);
		return LRegexp;
	
	case '\\':
		if((c = yygetc()) == '\n')
			return _yylex(yylval);
		yyungetc(c);
		goto Default;
	
	default:
	Default:
		// A messy by-hand hack: when this flag is set, do not parse keywords.
		// Instead, parse names which may be used as a grammar term name.
		// This includes special decoration characters for option/list terms.
		if(intermname){
			if(c == '{'){
				yylval->name = nameof("{");
				yylval->len = 1;
				return LTermLBracket;
			}
			if(c == '('){
				yylval->name = nameof("(");
				yylval->len = 1;
				return LTermLParen;
			}
			// hack
			if(c < 0x80 && !isalnum(c) && c != '_')
				goto Lexsym;
			if(!yylextermname(c, sbuf, sbufend, &p))
				return Error;
			*p = 0;
			yylval->name = nameof(sbuf);
			yylval->len = strlen(sbuf);
			return LTermName;
		}

		// Usual identifier lexing
		if(c >= 0x80 || isalnum(c) || c == '_'){
		Ident:
			if(!yylexname(c, sbuf, sbufend, &p))
				return Error;
			yylval->name = nameof(sbuf);
			yylval->len = strlen(sbuf);
			// All grammar keywords are prefixed with '%'
			if(ingrammar && sbuf[0] != '%')
				return LName;
			if((sym = symsearch(yylval->name)) != nil)
				return sym->token;
			return LName;
		}

	Lexsym:
		p = sbuf;
		sym = nil;
		// Usual symbol lexing
		do {
			*p++ = c;
			*p = 0;
			if((nsym = symsearch(sbuf)) != nil)
				sym = nsym;
			else{
				//yyungetc(c);
				break;
			}
		}while((c = yygetc()) != EOF && c < 0x80 && !isalnum(c) && c != '_');
		if(sym == nil)
			return Error;
		yylval->name = nameof(sym->str);
		yylval->len = strlen(sym->str);
		if(c != EOF)
			yyungetc(c);
		return sym->token;
	}
}

int
yylex(Yystype *yylval, Line *line)
{
	int tok;
	
	tok = _yylex(yylval);
	inbuf->yyline.line1 = inbuf->curpos.line;
	inbuf->yyline.byte1 = inbuf->curpos.byte;
	*line = inbuf->yyline;
	return tok;
}

char *cleanname(char*);

int
pushinput(char *sfile, char *dir)
{
	static Hash<Name,Name> *opened;
	Inbuf *ib;
	Name file = nameof(sfile);

	addfile(file);

	if(!opened)
		opened = mkhash<Name,Name>();

	// Relative path?
	if(file && file[0] != '/' && (file[0] != '.' || file[1] != '/')){
		// Use the given directory, if any.
		if(dir && strcmp(dir, "/dev/") != 0){
			sfile = smprint("%s%s", dir, file);
			cleanname(sfile);
			file = nameof(sfile);
		}
		// Otherwise relative to the current file, if any.
		else if(inbuf && inbuf->dir && strcmp(inbuf->dir, "/dev/") != 0){
			sfile = smprint("%s%s", inbuf->dir, file);
			cleanname(sfile);
			file = nameof(sfile);
		}
	}
	// Absolute.
	else if(file)
		file = nameof(file);

	if(hashget(opened, file))
		return 0;
	hashput(opened, file, file);

	ib = (Inbuf*)emalloc(sizeof *ib);
	if(file == nil)
		ib->fin = stdin;
	else if((ib->fin = fopen(file, "r")) == nil)
		panic("open %s: %r", file);

	if (strchr(file, '/'))
		dir = nameofn(file, strrchr(file, '/')-file+1);
	else
		dir = nameof("./");

	ib->curpos.file = nameof(file);
	ib->curpos.line = 1;
	ib->curpos.byte = 0;
	ib->next = inbuf;
	ib->dir = dir;
	inbuf = ib;

	return 1;
}

void
popinput(void)
{
	Inbuf *ib;

	if(inbuf == nil)
		panic("popinput");
	ib = inbuf;
	inbuf = ib->next;
	fclose(ib->fin);
	free(ib->bp);
	free(ib);
}

void
lexinit(char* file)
{
	fmtinstall('L', Lconv);
	pushinput(file, nil);
}

int
isinteractive(void)
{
	return isatty(fileno(inbuf->fin));
}

void
yyerror(char *msg)
{
	if (isinteractive()) {
		error(inbuf->curpos, "yyerror: %s", msg);
	} else {
		panic("%L: yyerror: %s", inbuf->curpos, msg);
	}
}

static int
inputfmt(Fmt *fmt, Line line)
{
	int n;
	char *p;
	
	n = line.byte1 - line.byte;
	p = inbuf->bp + line.byte;
	if(p > inbuf->ep)
		return fmtstrcpy(fmt, "");
	if(p+n > inbuf->ep)
		n = inbuf->ep - p;
	return fmtprint(fmt, "%.*s", utfnlen(p, n), p);
}

int
exprfmt(Fmt *fmt)
{
	Expr *e;

	e = va_arg(fmt->args, Expr*);
	if(e == nil)
		return fmtprint(fmt, "<nil expr>");
	if(fmt->flags&FmtSharp){
		pprintexpr(fmt, e);
		return 0;
	}
	else
		return inputfmt(fmt, e->line_);
}

int
stmtfmt(Fmt *fmt)
{
	Stmt *s;

	s = va_arg(fmt->args, Stmt*);
	if(s == nil)
		return fmtprint(fmt, "<nil stmt>");
	if(fmt->flags&FmtSharp){
		pprintstmt(fmt, s);
		return 0;
	}
	else
		return inputfmt(fmt, s->line_);
}

int
patfmt(Fmt *fmt)
{
	Pattern *p;

	p = va_arg(fmt->args, Pattern*);
	if(p == nil)
		return fmtprint(fmt, "<nil pattern>");
	if(fmt->flags&FmtSharp){
		pprintpattern(fmt, p);
		return 0;
	}
	else
		return inputfmt(fmt, p->line);
}

static int
yylexcomment()
{
	int c;

	// Try C++-style comment.
	if((c = yygetc()) == '/'){
		while((c = yygetc()) != EOF && c != '\n')
			;
		if(c == '\n')
			yyungetc(c);
		return 1;
	}
	// Try C-style comment.
	if(c == '*'){
		for(;;){
			while((c = yygetc()) != EOF && c != '*')
				;
			if(c == '*')
				c = yygetc();
			if(c == EOF || c == '/')
				return 1;
		}
	}
	// Not a comment.
	yyungetc(c);
	return 0;
}

static int
yylexstring(int end, char *type, char *bufstart, char *buflast, char **bufnext)
{
	char *p;
	int c;
	int q, needq;

	q = 1;
	p = bufstart;
	if(end != '/'){
		while((c = yygetc()) == end && ++q < 3)
			;
		if(c != end)
			yyungetc(c);
		if(q == 2){
			// empty string
			goto out;
		}
	}
	needq = q;
	
	q = 0;
	for(p = bufstart;; ){
		c = yygetc();
		if(c == end){
			if(++q == needq)
				break;
		}else
			q = 0;
		if(c == EOF || (c == '\n' && needq != 3)){
			*p = 0;
			error(inbuf->curpos, "end of line in %s", type);
			if(c == '\n')
				yyungetc(c);
			goto error;
		}
		if(p >= buflast){
			error(inbuf->curpos, "%s too long", type);
			goto error;
		}
		*p++ = c;
		if(c == '\\'){
			c = yygetc();
			if(c == '\n' && needq != 3){
				static int warned;
				if(warned++ == 0)
					warn(inbuf->curpos, "escaped newline in %s", type);
				*p++ = c;
				continue;
			}
			if(c == EOF){
				error(inbuf->curpos, "end of file in %s", type);
				goto error;
			}
			if(p >= buflast){
				error(inbuf->curpos, "%s too long", type);
				goto error;
			}
			*p++ = c;
		}
	}
	p -= needq-1;

out:
	*p = 0;
	if(bufnext)
		*bufnext = p;
	return 1;
error:
	*p = 0;
	if(bufnext)
		*bufnext = p;
	return 0;
}


static int
yylexname(int c, char *bufstart, char *buflast, char **bufnext)
{
	char *p;

	p = bufstart;
	do {
		if(!bufappend(&p,c,buflast))
			return 0;
	}while((c = yygetc()) != EOF && (c >= 0x80 || isalnum(c) || c == '_'));
	yyungetc(c);
	*p = 0;
	if(bufnext)
		*bufnext = p;
	return 1;
}

static int
yylextermname(int c, char *bufstart, char *buflast, char **bufnext)
{
	char *p;

	p = bufstart;
	if(c >= 0x80 || isalnum(c) || c == '_'){
		// name(?|*|+) or name[sep](*|+)
		if(!yylexname(c, p, buflast, &p))
			return 0;
		c = yygetc();
		if(c == '?' || c == '*' || c == '+'){
			if(!bufappend(&p, c, buflast))
				return 0;
		}
		else if(c == '['){
			// name[sep](*|+)
			do{
				if(!bufappend(&p, c, buflast))
					return 0;
			}while((c = yygetc()) != EOF && c != ']');
			if(c == ']' && !bufappend(&p, c, buflast))
				return 0;
			if((c = yygetc()) == '+' || c == '*'){
				if(!bufappend(&p, c, buflast))
					return 0;
			}
			else
				return 0;
		}
		else
			yyungetc(c);
	}
	else{
		error(inbuf->curpos, "illegal term name near '%c'", c);
		return 0;
	}
	*p = 0;
	if(bufnext)
		*bufnext = p;
	return 1;
}

// TO DO: Quoted strings?
static int
yylexbacktick(int c, Yystype *yylval)
{
	char sbuf[4096], *p, *sbufend;
	int token;
	static int nbrace = 1;

	sbufend = sbuf + sizeof(sbuf) - 1;
	p = sbuf;

	if(c == '\\'){
		c = yygetc();
		// \name(::termname)
		if(c >= 0x80 || isalnum(c) || c == '_'){
			if(!yylexname(c, sbuf, sbufend, &p))
				return Error;
			if((c = yygetc()) != ':'){
				yyungetc(c);
				token = LSlot;
				goto havetoken;
			}
			if((c = yygetc()) != ':'){
			//	warn(inbuf->curpos, "did you mean ::?");
				yyungetc(c);
				yyungetc(c);
				token = LSlot;
				goto havetoken;
			}
			*p++ = ':';
			*p++ = ':';
			c = yygetc();
			if(!yylextermname(c, p, sbufend, &p))
				return Error;
			token = LTypedSlot;
			goto havetoken;
		}
		else if(c == '('){
			*p++ = c;
			token = LSlotExprStart;
			goto havetoken;
		}
		else{
			*p++ = c;
			warn(inbuf->curpos, "this is probably a bad slot");
			goto dobraces;
		}
	havetoken:
		*p = 0;
		yylval->name = nameof(sbuf);
		yylval->len = p-sbuf;
		return token;
	}

dobraces:
	if(c == '}' && nbrace == 1)
		return '}';
	
	do{
		if(c == '{')
			nbrace++;
		else if(c == '}'){
			if(--nbrace == 0){
				nbrace = 1;
				break;
			}
		}
		if(c == '/'){
			if(yylexcomment())
				continue;
		}
		if(!bufappend(&p,c,sbufend))
			panic("yylexbacktick");
		if(c == '"' || c == '\''){
			if(!yylexstring(c, "string", p, sbufend, &p))
				return 0;
			if(!bufappend(&p,c,sbufend))
				panic("yylexbacktick");
		}
	}while((c = yygetc()) != EOF && c != '\\');
	
	if(c != EOF){
		yyungetc(c);
	}
	*p = 0;
	yylval->name = nameof(sbuf);
	yylval->len = p-sbuf;
	return LToken;
}

// XXX Lifted from Plan 9

/*
 * In place, rewrite name to compress multiple /, eliminate ., and process ..
 */
#define SEP(x)	((x)=='/' || (x) == 0)
char*
cleanname(char *name)
{
	char *p, *q, *dotdot;
	int rooted;

	rooted = name[0] == '/';

	/*
	 * invariants:
	 *	p points at beginning of path element we're considering.
	 *	q points just past the last path element we wrote (no slash).
	 *	dotdot points just past the point where .. cannot backtrack
	 *		any further (no slash).
	 */
	p = q = dotdot = name+rooted;
	while(*p) {
		if(p[0] == '/')	/* null element */
			p++;
		else if(p[0] == '.' && SEP(p[1]))
			p += 1;	/* don't count the separator in case it is nul */
		else if(p[0] == '.' && p[1] == '.' && SEP(p[2])) {
			p += 2;
			if(q > dotdot) {	/* can backtrack */
				while(--q > dotdot && *q != '/')
					;
			} else if(!rooted) {	/* /.. is / but ./../ is .. */
				if(q != name)
					*q++ = '/';
				*q++ = '.';
				*q++ = '.';
				dotdot = q;
			}
		} else {	/* real path element */
			if(q != name+rooted)
				*q++ = '/';
			while((*q = *p) != '/' && *q != 0)
				p++, q++;
		}
	}
	if(q == name)	/* empty string is really ``.'' */
		*q++ = '.';
	*q = '\0';
	return name;
}
