/*
 * CFG computations.
 */
#include "u.h"
#include "utf.h"
#include "fmt.h"
#include "name.h"
#include "arena.h"
#include "emalloc.h"
#include "list2.h"
#include "cfg.h"
#include "panic.h"
#include "bitset.h"

/*
 * Compute whether each symbol can produce an empty string.
 */
static void
empty(CfgGram<void> *g)
{
	int i, j, k, did, flag;
	CfgRule<void> *r;
	CfgSym<void> *s, *ss;
	
	/* Step 1: clear CfgSymEmpty flags */
	for(i=0; i<g->nsym; i++){
		s = g->sym[i];
		s->flags &= ~CfgSymEmpty;
	}

	/* Step 2: loop looking for symbols with RHS that can all empty */
	/* could keep a queue for speed, but not a bottleneck */
	do{
		did = 0;
		for(i=0; i<g->nsym; i++){
			s = g->sym[i];
			if(s->flags&CfgSymEmpty)
				continue;
			for(j=0; j<s->nrule; j++){
				r = s->rule[j];
				flag = CfgSymEmpty;
				for(k=0; k<r->nright; k++){
					ss = r->right[k];
					flag &= ss->flags;
				}
				if(flag){
					s->flags |= flag;
					did = 1;
				}
			}
		}
	}while(did);
}

/*
 * Compute first set for each symbol s:
 * symbols that can start a sentence derived from s.
 * Must include non-terminals too, in case lexer is handing back
 * non-terminals.
 *
 * Caller must have called empty already.
 */
static void
first(CfgGram<void> *g)
{
	int did, i, j, k;
	CfgRule<void> *r;
	CfgSym<void> *s, *ss;

	/* allocate sets, initialize all symbols with themselves */
	for(i=0; i<g->nsym; i++){
		s = g->sym[i];
		freebitset(s->first);
		s->first = newbitset(g->nsym);
		addbit(s->first, s->n);
	}

	/* iterate to find first sets for right hand sides */
	do{
		did = 0;
		for(i=0; i<g->nsym; i++){
			s = g->sym[i];
			for(j=0; j<s->nrule; j++){
				r = s->rule[j];
				for(k=0; k<r->nright; k++){
					ss = r->right[k];
					did += addbitset(s->first, ss->first);
					if(!(ss->flags&CfgSymEmpty))
						break;
				}
			}
		}
	}while(did);
}

/*
 * Compute follow set for each symbol s:
 * symbols that can follow an s in a valid sentence of any type.
 */
static void
follow(CfgGram<void> *g)
{
	int i, did, k, lastnonempty, l;
	CfgRule<void> *r;
	CfgSym<void> **rhs, *s;

	for(i=0; i<g->nsym; i++){
		s = g->sym[i];
		freebitset(s->follow);
		s->follow = newbitset(g->nsym);
	}

	do{
		did = 0;
		for(i=0; i<g->nrule; i++){
			r = g->rule[i];
			s = r->left;
			rhs = r->right;
			lastnonempty = 0;
			for(k=0; k<r->nright; k++){
				/* X => ... Y Z ..., so follow(Y) += first(Z) */
				for(l=lastnonempty; l<k; l++)
					did += addbitset(rhs[l]->follow, rhs[k]->first);
				if(!(rhs[k]->flags&CfgSymEmpty))
					lastnonempty = k;
			}
			/* X => ... Y, so follow(Y) += follow(X) */
			for(l=lastnonempty; l<r->nright; l++)
				did += addbitset(rhs[l]->follow, s->follow);
		}
	}while(did);
}

/*
 * Compute `leads to' set for each symbol.
 * A leads to B if A => ... => B
 */
static void
leadsto(CfgGram<void> *g)
{
	int did, i, j, k, lastnonempty, nnonempty;
	CfgSym<void> *s;
	CfgRule<void> *r;

	for(i=0; i<g->nsym; i++){
		s = g->sym[i];
		freebitset(s->leadsto);
		s->leadsto = newbitset(g->nsym);
		addbit(s->leadsto, s->n);
	}
	
	do{
		did = 0;
		for(i=0; i<g->nsym; i++){
			s = g->sym[i];
			for(j=0; j<s->nrule; j++){
				r = s->rule[j];
				nnonempty = 0;
				lastnonempty = 0;
				for(k=0; k<r->nright; k++){
					if(!(r->right[k]->flags&CfgSymEmpty)){
						nnonempty++;
						lastnonempty = k;
					}
				}
				switch(nnonempty){
				case 0:
					for(k=0; k<r->nright; k++)
						did += addbitset(s->leadsto, r->right[k]->leadsto);
					break;
				case 1:
					did += addbitset(s->leadsto, r->right[lastnonempty]->leadsto);
					break;
				}
			}
		}
	}while(did);	
}

void
_cfgcompile(CfgGram<void> *g)
{
	empty(g);
	first(g);
	follow(g);
	leadsto(g);
}
