struct Bitset;

/* associativity */
typedef enum CfgAssoc
{
	CfgAssocLeft    = 1,
	CfgAssocNone    = 0,
	CfgAssocRight   = -1
} CfgAssoc;

/* precedence resolution kinds */
typedef enum CfgPrecKind
{
	CfgPrecPriority,      /* operator priority style resolution */
	CfgPrecPrefer,        /* rule preference resolution (like SDF's 'prefer') */
} CfgPrecKind;

template<typename T> struct CfgSym;
template<typename T> struct CfgRule;
template<typename T> struct CfgPrec;
#define CfgAction(T) typeof(int(CfgGram<T>*, CfgRule<T>*, T*, T*))

// A CfgGram<T> is a grammar whose value stack has type T.
template<typename T>
struct CfgGram
{
	int     debug;
	int     valsize;
	CfgSym<T>  *wildcard;      /* obsolete? */
	CfgSym<T>  *eofsym;        /* obsolete? */

	/* internal use only */
	int     dirty;          /* needs compilation */
	uint    seqnum;         /* incremented on compilation */
	CfgSym<T>  **sym;          /* all symbols */
	int     nsym;
	CfgRule<T> **rule;         /* all rules */
	int     nrule;
	CfgPrec<T> **prec;         /* all precedence levels */
	int     nprec;
	int     *precorder;     /* precedence relationship (2d matrix) */
	
	uint	mtime;		/* last modification time */

	/* extra for SLR */
	struct SlrState **slrstate;     /* all states */
	int             nslrstate;
	struct SlrState *slrstatehash[128];
	void (*slrfreestate)(struct SlrState*);

	/* extra for GLR */
	struct GlrState **glrstate;
	int             nglrstate;
	struct GlrState *glrstatehash[128];
	void (*glrfreestate)(struct GlrState*);
};

/* terminal or non-terminal in grammar */
enum
{
	CfgSymEmpty = 1<<1      /* reduces to empty string */
};
template<typename T>
struct CfgSym
{
	CfgGram<T> *gram;          /* parent grammar module */
	Name    name;           /* name of symbol */
	int     n;              /* symbol number */
	int     flags;          /* CfgSymReduces, etc. above */
	Name    type;           /* type */
	uint	mtime;	/* last modification time */

	CfgRule<T> **rule;         /* rules with symbol on left */
	int     nrule;
	CfgRule<T> *parseme;

	struct Bitset   *first;         /* first terms of derived strings */
	struct Bitset   *follow;        /* follow set */
	struct Bitset   *leadsto;       /* reduces to (A =>* B) */
	int     code;
	int     slotted;                /* has slot rule been added */
	int     reflags;                /* hacky: for sym copying */

	/* Opaque pointer representing the origin of this symbol. */
	/* In zeta, this is the Type* of a symbol (TypeSyntax). */
	void *_origin;
};

/* grammar production */
template<typename T>
struct CfgRule
{
	int        n;
	CfgPrec<T>    *prec;
	CfgSym<T>     *left;       /* left hand side */
	CfgSym<T>     **right;     /* right hand side */
	int        nright;
	int	nsavedright;
	int        toplevel;
	CfgAction(T)  *fn;         /* reduce fn */
	int        code;
	Name       fmtname;     /* name from cfgrulefmt */
	Name       name;        /* user-specified rule name, defaults to "%R" */

	/* Opaque pointer representing the origin of this rule. */
	/* In zeta, this is the Type* of a rule (TypeSyntax). */
	void *_origin;
};

/* grammar precedence level (partial-order) */
template<typename T>
struct CfgPrec
{
	int         n;
	CfgGram<T>     *g;
	CfgPrecKind kind;
	CfgAssoc   assoc;       /* only for kind == CfgPrecKind */
	Name        name;
	List<CfgPrec<T>*>        *aboveme;    /* immediately above only */
	List<CfgPrec<T>*>        *belowme;    /* immediately below only */
};

void     _cfgcompile(CfgGram<void>*);
void     _cfgfreegram(CfgGram<void>*);

CfgGram<void>* _cfgnewgram(int);
CfgSym<void>*  _cfgnewsym(CfgGram<void>*, char*, char*);
CfgPrec<void>* _cfgnewprec(CfgGram<void>*, char*, CfgPrecKind);

CfgSym<void>*  _cfglooksym(CfgGram<void>*, char*);
CfgRule<void>* _cfglookrule(CfgGram<void>*, char*);
CfgPrec<void>* _cfglookprec(CfgGram<void>*, char*);

int      _cfglinkprec(CfgPrec<void>*, CfgPrec<void>*);
int      _cfgcompareprec(CfgPrec<void>*, CfgPrec<void>*);

int      cfgrulefmt(Fmt*);

CfgRule<void>* _cfgrule(CfgGram<void>*, CfgAction(void)*, CfgPrec<void>*, CfgSym<void>*, CfgSym<void>**, int);
CfgRule<void>* _vcfgrulestr(CfgGram<void>*, CfgAction(void)*, char*, char*, char*, va_list);
CfgRule<void>* _vcfgrulesym(CfgGram<void>*, CfgAction(void)*, CfgPrec<void>*, CfgSym<void>*, CfgSym<void>*, va_list);

template<typename T>
void
cfgcompile(CfgGram<T> *g)
{
	_cfgcompile((CfgGram<void>*)g);
}

template<typename T>
void
cfgfreegram(CfgGram<T> *g)
{
	_cfgfreegram((CfgGram<void>*)g);
}

template<typename T>
CfgGram<T>*
cfgnewgram(int sz)
{
	return (CfgGram<T>*)_cfgnewgram(sz);
}

template<typename T>
CfgSym<T>*
cfgnewsym(CfgGram<T> *g, char *name, char *type)
{
	return (CfgSym<T>*)_cfgnewsym((CfgGram<void>*)g, name, type);
}

template<typename T>
CfgPrec<T>*
cfgnewprec(CfgGram<T> *g, char *name, CfgPrecKind k)
{
	return (CfgPrec<T>*)_cfgnewprec((CfgGram<void>*)g, name, k);
}

template<typename T>
CfgSym<T>*
cfglooksym(CfgGram<T> *g, char *name)
{
	return (CfgSym<T>*)_cfglooksym((CfgGram<void>*)g, name);
}

template<typename T>
CfgRule<T>*
cfglookrule(CfgGram<T> *g, char *name)
{
	return (CfgRule<T>*)_cfglookrule((CfgGram<void>*)g, name);
}

template<typename T>
CfgPrec<T>*
cfglookprec(CfgGram<T> *g, char *name)
{
	return (CfgPrec<T>*)_cfglookprec((CfgGram<void>*)g, name);
}

template<typename T>
int
cfglinkprec(CfgPrec<T> *p1, CfgPrec<T> *p2)
{
	return _cfglinkprec((CfgPrec<void>*)p1, (CfgPrec<void>*)p2);
}

template<typename T>
int
cfgcompareprec(CfgPrec<T> *p1, CfgPrec<T> *p2)
{
	return _cfgcompareprec((CfgPrec<void>*)p1, (CfgPrec<void>*)p2);
}

template<typename T>
CfgRule<T>*
cfgrule(CfgGram<T> *g, CfgAction(T) *action, CfgPrec<T> *prec, CfgSym<T> *left, CfgSym<T> **right, int nright)
{
	return (CfgRule<T>*)_cfgrule((CfgGram<void>*)g, (CfgAction(void)*)action, (CfgPrec<void>*)prec,
		(CfgSym<void>*)left, (CfgSym<void>**)right, nright);
}

template<typename T>
CfgRule<T>*
vcfgrulestr(CfgGram<T> *g, CfgAction(T) *action, char *prec, char *left, char *right0, va_list arg)
{
	return (CfgRule<T>*)_vcfgrulestr((CfgGram<void>*)g, (CfgAction(void)*)action, prec, left, right0, arg);
}

template<typename T>
CfgRule<T>*
cfgrulestr(CfgGram<T> *g, CfgAction(T) *action, char *prec,  char *left, char *right0, ...)
{
	va_list arg;
	
	va_start(arg, right0);
	CfgRule<T> *r = vcfgrulestr(g, action, prec, left, right0, arg);
	va_end(arg);
	return r;
}

template<typename T>
CfgRule<T>*
vcfgrulesym(CfgGram<T> *g, CfgAction(T) *action, CfgPrec<T> *prec, CfgSym<T> *left, CfgSym<T> *right0, va_list arg)
{
	return (CfgRule<T>*)_vcfgrulesym((CfgGram<void>*)g, (CfgAction(void)*)action, (CfgPrec<void>*)prec, (CfgSym<void>*)left, (CfgSym<void>*)right0, arg);
}

template<typename T>
CfgRule<T>*
cfgrulesym(CfgGram<T> *g, CfgAction(T) *action, CfgPrec<T> *prec,  CfgSym<T> *left, CfgSym<T> *right0, ...)
{
	va_list arg;
	
	va_start(arg, right0);
	CfgRule<T> *r = vcfgrulesym(g, action, prec, left, right0, arg);
	va_end(arg);
	return r;
}
