/*
 * GLR parser. 
 *
 * Thanks to Scott McPeak for writing UCB TR CSD-2-1214.
 */
#include "u.h"
#include "utf.h"
#include "fmt.h"
#include "name.h"
#include "emalloc.h"
#include "arena.h"
#include "list2.h"
#include "cfg.h"
#include "slr.h"
#include "glr.h"
#include "bitset.h"

typedef struct GlrPathelem GlrPathelem;
struct GlrPathelem
{
	GlrEdge *edge;
	GlrPathelem *next;
};

int nedge, nstack, npath;
int nmkedge, nfreeedge;
int nmkstack, nfreestack;
int nmkpath, nfreepath;

static GlrEdge*
mkedge(GlrParse<void> *p, CfgGram<void> *g, CfgSym<void> *sym, void *val, GlrStack *down, GlrEdge *sib)
{
	GlrEdge *e;
	
	e = (GlrEdge*)emalloc(sizeof *e+g->valsize);
	e->sib = sib;
	e->down = down;
	down->ref++;
	e->sym = sym;
	memmove(e->val, val, g->valsize);
	if(p->dup)
		p->dup(sym, (void*)e->val, val);
	nmkedge++;
	return e;
}

static void freestack(GlrParse<void>*, GlrStack*);

static void
freeedge(GlrParse<void> *p, GlrEdge *e)
{
	if(e == nil)
		return;
	nfreeedge++;
	freeedge(p, e->sib);
	freestack(p, e->down);
	if(p->free)
		p->free(e->sym, (void*)e->val);
	efree(e);
}

static GlrStack*
mkstack(GlrParse<void> *p, SlrState *state)
{
	GlrStack *s;
	
	s = (GlrStack*)emalloc(sizeof *s);
	s->ipos = p->ipos;
	s->state = state;
	s->ref++;
	nmkstack++;
	return s;
}

static void
freestack(GlrParse<void> *p, GlrStack *s)
{
	if(--s->ref > 0)
		return;
	nfreestack++;
	if(s->ref < 0)
		abort();
	freeedge(p, s->edge);
	efree(s);
}

static GlrPath*
mkpath(GlrParse<void> *p, CfgRule<void> *r, GlrStack *sp, GlrPathelem *pe)
{
	int i;
	GlrPath *path;
	
	nmkpath++;
	path = (GlrPath*)emalloc(sizeof *path+r->nright*sizeof path->edge[0]);
	path->sp = sp;
	path->rule = r;
	for(i=0; i<r->nright; i++){
		path->edge[i] = pe->edge;
		pe = pe->next;
	}
	if(r->nright>0)
		path->leftpos = path->edge[0]->down->ipos;
	else
		path->leftpos = p->ipos;
	npath++;
	return path;
}

static void
freepath(GlrPath *path)
{
	nfreepath++;
	efree(path);
}

GlrParse<void>*
_glrnewparse(CfgGram<void> *g)
{
	GlrParse<void> *p;
	
	p = (GlrParse<void>*)emalloc(sizeof(GlrParse<void>));
	p->g = g;
	return p;
}

static GlrParse<void>*
glrparseinit(CfgGram<void> *g, GlrParse<void> *p, CfgSym<void> *start)
{
	char *val;
	SlrState* init;
	
	slrcompile(g);
	init = slrinitstate(g, start);
	if(!init)
		return nil;

	if(p == nil)
		p = glrnewparse(g);

	p->mtop = 32;
	p->top = (GlrStack**)emalloc(p->mtop*sizeof p->top[0]);
	p->top[0] = mkstack(p, init);
	p->ntop = 1;
	val = (char*)emalloc(20*g->valsize);
	p->rtokval = val;
	p->tmpval = val+g->valsize;
	p->toks = nil;
	p->seqnum = g->seqnum;
	return p;
}

void
glrtokenlistfree(GlrParse<void> *p, List<GlrToken<void>*> *l)
{
	List<GlrToken<void>*> *l1;
	
	for(l1=l; l1; l1=l1->next)
		glrfreetoken(p, l1->x);
	listfree(l);
}

void
_glrfreeparse(GlrParse<void> *p)
{
	List<GlrToken<void>*> *l;

	// XXX free paths inside p?
	glrtokenlistfree(p, p->toks);
	for(l=p->freetoks; l; l=l->next)
		free(l->x);
	listfree(p->freetoks);
	free(p);
}

static void
printinput(GlrStack *sp)
{
	if(sp->edge == nil)
		return;
	printinput(sp->edge->down);
	print(" %s", sp->edge->sym->name);
}

/*
 * Parsing work queue (heap).
 * Sorted by:
 *	Rule 1: reductions which span fewer tokens com first.
 *	Rule 2: If two reductions A -> alpha and B -> beta span
 *		the same tokens, then A -> alpha comes first if B ->+ A.
 * See section 4.2 of McPeak's report.
 */
static int
leadsto(CfgSym<void> *a, CfgSym<void> *b)
{
	return inbitset(a->leadsto, b->n);
}
static int
ypathcmp(GlrPath *a, GlrPath *b)
{
	if(a->leftpos != b->leftpos)
		return b->leftpos - a->leftpos;
	if(a->rule->left == b->rule->left)
		return 0;
	if(leadsto(b->rule->left, a->rule->left))
		return -1;
	if(leadsto(a->rule->left, b->rule->left))
		return 1;
	return 0;
}

static void
addpath(GlrParse<void> *p, GlrPath *path)
{
	int i, up;
	
	if(p->nheap >= p->mheap){
		assert(p->nheap == p->mheap);
		p->mheap += 100;
		p->heap = (GlrPath**)erealloc(p->heap, p->mheap*sizeof p->heap[0]);
	}
	p->heap[i=p->nheap++] = path;
	for(; i>0; i=up){
		up = (i-1)/2;
		if(ypathcmp(path, p->heap[up]) >= 0)
			break;
		p->heap[i] = p->heap[up];
		p->heap[up] = path;
	}
}

static GlrPath*
nextpath(GlrParse<void> *p)
{
	int i, down;
	GlrPath *ret;
	GlrPath *path;
	
	if(p->nheap == 0)
		return nil;
	ret = p->heap[0];
	--p->nheap;
	if(p->nheap > 0){
		path = p->heap[p->nheap];
		p->heap[0] = path;
		for(i=0; i<p->nheap; i=down){
			down = i*2+1;
			if(down >= p->nheap)	/* no kids */
				break;
			/* determine smaller kid */
			if(down+1 < p->nheap && ypathcmp(p->heap[down], p->heap[down+1]) > 0)
				down++;
			if(ypathcmp(path, p->heap[down]) < 0)
				break;
			p->heap[i] = p->heap[down];
			p->heap[down] = path;
		}
	}
	return ret;
}

/*
 * Parse
 */
static void walkpaths(GlrPathelem*, GlrParse<void>*, GlrStack*, uint, CfgRule<void>*);
static void reducepath(CfgGram<void>*, GlrParse<void>*, GlrPath*);

static void
doreduce(CfgGram<void> *g, GlrParse<void> *p)
{
	int i, j;
	CfgRule<void> *r;
	GlrStack *sp;
	GlrPath *path;

	for(i=0; i<p->ntop; i++){
		sp = p->top[i];
		if(!sp->state->compiled)
			slrcompilestate(g, sp->state);
		for(j=0; j<sp->state->nreduce; j++){
			r = sp->state->reduce[j];
			walkpaths(nil, p, sp, r->nright, r);
		}
	}

	while((path = nextpath(p)) != nil){
		reducepath(g, p, path);
		freepath(path);
	}
}

static void
qreduce1(CfgGram<void> *g, GlrParse<void> *p, GlrStack *sp, GlrEdge *e)
{
	int i;
	CfgRule<void> *r;
	GlrPathelem pe;

	/*
	 * Add all possible reductions that use the (newly added) edge e
	 * to the work queue.
	 */
	if(!sp->state->compiled)
		slrcompilestate(g, sp->state);
	for(i=0; i<sp->state->nreduce; i++){
		r = sp->state->reduce[i];
		if(r->nright == 0){
			addpath(p, mkpath(p, r, sp, nil));
			continue;
		}
		pe.edge = e;
		pe.next = nil;
		walkpaths(&pe, p, e->down, r->nright-1, r);
	}
}

/*
 * Explore paths and queue them.
 */
static void
walkpaths(GlrPathelem *pep, GlrParse<void> *p, GlrStack *sp, uint n, CfgRule<void> *r)
{
	GlrEdge *e;
	GlrPathelem pe;
	
	if(n == 0){
		addpath(p, mkpath(p, r, sp, pep));
		return;
	}
	
	for(e=sp->edge; e; e=e->sib){
		pe.edge = e;
		pe.next = pep;
		walkpaths(&pe, p, e->down, n-1, r);
	}
}

/*
 * Find the stack top ending in state.
 */
static GlrStack*
findtop(GlrParse<void> *p, SlrState *state)
{
	int i;
	GlrStack *sp;
	
	for(i=0; i<p->ntop; i++){
		sp = p->top[i];
		if(sp->state == state)
			return sp;
	}
	return nil;
}

/*
 * Create a new stack top for state.
 */
static GlrStack*
newtop(GlrParse<void> *p, SlrState *state)
{
	GlrStack *sp;

	sp = mkstack(p, state);
	if(p->ntop >= p->mtop){
		p->mtop += 40;
		p->top = (GlrStack**)erealloc(p->top, p->mtop*sizeof p->top[0]);
	}
	p->top[p->ntop++] = sp;
	return sp;
}

/*
 * Find the edge coming out of sp that leads to down.
 */
static GlrEdge*
findedge(GlrStack *sp, GlrStack *down)
{
	GlrEdge *e;
	
	for(e=sp->edge; e; e=e->sib)
		if(e->down == down)
			return e;
	return nil;
}

/*
 * Apply a reduction along a given path.
 */
static void
reducepath(CfgGram<void> *g, GlrParse<void> *p, GlrPath *path)
{
	int i;
	GlrStack *sp;
	GlrEdge *e;
	CfgRule<void> *rule;
	SlrState *nextstate;

if(g->debug) print("reduce %R\n", path->rule);
	rule = path->rule;
	// XXX check for overflow in p->tmpval? (unlikely)
	for(i=0; i<rule->nright; i++){
		memmove((char*)p->tmpval+(1+i)*g->valsize, path->edge[i]->val, g->valsize);
		/* no dup - the action rule must dup if it wants it */
		/*
		if(p->dup)
			p->dup(p->tmpval+(1+i)*g->valsize, path->edge[i]->val);
		*/
	}
//	memset(p->tmpval, 0, g->valsize);
	memmove(p->tmpval, (char*)p->tmpval+g->valsize, g->valsize);
	if(rule->fn){
		if(rule->fn(g, rule, p->tmpval, (char*)p->tmpval+g->valsize) < 0){
			return;
		}
	//	reshift(g, start, p);
	}
	// *(CfgSym<void>**)p->tmpval = rule->left;
	if(!slrcanshift(g, path->sp->state, rule->left, &nextstate))
		return;

	if((sp = findtop(p, nextstate)) == nil)
		sp = newtop(p, nextstate);

	if((e = findedge(sp, path->sp)) == nil){
		sp->edge = mkedge(p, g, rule->left, p->tmpval, path->sp, sp->edge);
		/* McPeak skips qreduce if we called newtop ? */
		qreduce1(g, p, sp, sp->edge);
	}else{
		if(p->merge)
			p->merge(rule->left, e->val, p->tmpval);
	}
	if(p->free)
		p->free(rule->left, p->tmpval);
}

/*
 * Apply all possible shifts.
 */
static void doshift1(CfgGram<void>*, GlrParse<void>*, GlrStack**, int, CfgSym<void>*, void*);
static void
doshift(CfgGram<void> *g, GlrParse<void> *p, List<GlrToken<void>*> *toks)
{
	int i, ntop;
	List<GlrToken<void>*> *l;
	GlrStack **top;
	GlrToken<void> *tok;

	/*
	 * Copy old stack tops - we use them many times.
	 */
	ntop = p->ntop;
	top = p->top;
	p->top = nil;
	p->mtop = 0;
	p->ntop = 0;
	
	/*
	 * Try to shift each possible symbol.
	 */
	for(l=toks; l; l=l->next){
		tok = l->x;
		doshift1(g, p, top, ntop, tok->sym, tok->val);
	}

	/*
	 * Drop refs to old stack tops.
	 */
	for(i=0; i<ntop; i++)
		freestack(p, top[i]);
	efree(top);
}

static void
doshift1(CfgGram<void> *g, GlrParse<void> *p, GlrStack **top, int ntop, CfgSym<void> *sym, void *val)
{
	int i;
	GlrStack *sp, *nsp;
	SlrState *nstate;

	for(i=0; i<ntop; i++){
		sp = top[i];
		slrcanshift(g, sp->state, sym, &nstate);
		if(nstate == nil)
			continue;
		if((nsp = findtop(p, nstate)) == nil)
			nsp = newtop(p, nstate);
		nsp->edge = mkedge(p, g, sym, val, sp, nsp->edge);
	}
}

void
_glrprintstate(GlrParse<void> *p)
{
	int i;
	
	for(i=0; i<p->ntop; i++)
		print("%d.%p. %l@\n", i, p->top[i]->state, p->top[i]->state);
}

#define H ((void*)-1)

void
_glrfreetoken(GlrParse<void> *p, GlrToken<void> *t)
{
	if(t == nil)
		return;
assert(t->sym != H);
	t->sym = (CfgSym<void>*)H;
	// XXX free val
	memset(t->val, 0, p->g->valsize);
	p->freetoks = listinsert(nil, p->freetoks, t);
}

GlrToken<void>*
_glrnewtoken(GlrParse<void> *p, CfgSym<void> *sym)
{
	List<GlrToken<void>*> *l;
	GlrToken<void> *t;

	l = p->freetoks;
	if(l != nil){
		p->freetoks = l->next;
		t = l->x;
		assert(t->sym == H);
		free(l);
	}else
		t = (GlrToken<void>*)emalloc(offsetof(GlrToken<void>, val) + p->g->valsize);
	t->sym = sym;
	return t;
}

void*
_glrparse(CfgGram<void> *g, CfgSym<void> *start, GlrParse<void> *p0, GlrLexer *yylex, void *arg)
{
	int i;
	void *rv;
	GlrParse<void> *p;
	CfgSym<void> *eofsym;
	GlrToken<void> *t;
	List<GlrToken<void>*> *l;

	p = glrparseinit(g, p0, start);
	if(p == nil)
		return nil;

	rv = nil;
	eofsym = cfglooksym(g, "EOF");
	if(g->debug>1)
		glrprintstate(p);
	for(;;){
		glrtokenlistfree(p, p->toks);
		if((p->toks = yylex(p, g, arg)) == nil){
			fprint(2, "weird: got unexpected nil from yylex\n");
			break;
		}
		p->ipos++;
		if(g->debug){
			print("yylex");
			for(l=p->toks; l; l=l->next){
				t = l->x;
				print(" %s", t->sym->name);
			}
			print("\n");
			if(g->debug>1)
				print("reduce\n");
		}
		doreduce(g, p);
		if(g->debug>1)
			glrprintstate(p);
		if(g->debug>1)
			print("shift\n");
		doshift(g, p, p->toks);
		if(p->ntop == 0){
			if(g->debug){
				fprint(2, "lost parse near %s\n", ((CfgSym<void>*)p->toks->x)->name);
				glrprintstate(p);
			}
			break;
		}
		if(g->debug>1)
			glrprintstate(p);
		t = (GlrToken<void>*)p->toks->x;
		if(t->sym == eofsym){
			if(g->debug)
				fprint(2, "success\n");
			if(p->ntop > 1){
				fprint(2, "ambiguous parse -- too many stack tops\n");
				break;
			}
			memmove(p->rtokval, p->top[0]->edge->down->edge->val, g->valsize);
			if(p->dup)
				p->dup((CfgSym<void>*)p->toks->x, p->rtokval, p->top[0]->edge->down->edge->val);
			rv = p->rtokval;
			break;
		}
	}

	for(i=0; i<p->ntop; i++)
		freestack(p, p->top[i]);
	free(p->top);
	p->top = nil;
	p->ntop = 0;
	p->mtop = 0;
	glrtokenlistfree(p, p->toks);
	p->toks = nil;
//	if(p != p0)
//		yyfreeparse(p);
	return rv;
}
