/*
 * Hash tables.
 *
 * Keys and values are arbitrary pointer-sized pieces of data.
 * 
 * Hash tables are resized automatically as necessary, so that
 * an empty hash table doesn't take much memory but a large
 * hash table is still efficient.
 */

#include "u.h"
#include <assert.h>
#include "arena.h"
#include "emalloc.h"
#include "list2.h"
#include "hash1.h"

/*
 * Hash table sizes: powers of three.
 * Shouldn't be correlated with pointer values
 * (except on ternary machines).
 */
static int hashsize[] = {
 	9,
 	27,
 	81,
 	243,
 	729,
 	2187,
 	6561,
 	19683,
 	59049,
 	177147,
 	531441,
 	1594323,
 	4782969,
 	14348907,
 	43046721,
 	129140163,
 	387420489,
 	1162261467	/* (getting a little ridiculous...) */
};

typedef struct Xel Xel;
struct Xel
{
	Xel *next;
	void *key;
	void *value;
};

template<typename K, typename V>
struct Hash
{
	int nel;      /* number of elements */
	Xel **tab;		/* hash chain heads */
	int size;     /* hashsize[size] heads in tab */
	Arena *arena; /* where to allocate tab (use emalloc if nil) */
};

#define H(h, p)	((uintptr)p%hashsize[h->size])		/* hash value */

/* Allocate a hash table. */
Hash<void*,void*>*
_mkhash(void)
{
	return _mkhasha(nil);
}

Hash<void*,void*>*
_mkhasha(Arena *a)
{
	Hash<void*,void*> *h;
	
	h = (Hash<void*,void*>*)amalloc(a, sizeof *h);
	h->tab = (Xel**)amalloc(a, hashsize[h->size]*sizeof h->tab[0]);
	h->arena = a;
	return h;
}

/* Allocate a hash table element. */
static Xel*
mkxel(Arena *a, void *key, void *value)
{
	Xel *x;

	x = (Xel*)amalloc(a, sizeof *x);
	x->key = key;
	x->value = value;
	x->next = nil;
	return x;
}

/* Rehash the table to have size hashsize[size]. */
static void
rehash(Hash<void*,void*> *h, int size)
{
	int i;
	Xel *all, *next, *x;
	
	/* Make list of all elements. */
	all = nil;
	for(i=0; i<hashsize[h->size]; i++){
		for(x=h->tab[i]; x; x=next){
			next = x->next;
			x->next = all;
			all = x;
		}
	}
	
	/* Resize. */
	h->size = size;
	if(!h->arena)
		free(h->tab);
	h->tab = (Xel**)amalloc(h->arena, hashsize[h->size]*sizeof h->tab[0]);
	
	/* Put the elements back. */
	for(x=all; x; x=next){
		next = x->next;
		i = H(h, x->key);
		x->next = h->tab[i];
		h->tab[i] = x;
	}
}

/* Decide whether to rehash. */
static void
autorehash(Hash<void*,void*> *h)
{
	int s;
	
	s = h->size;
	
	/* Grow if we've reached the next level. */
	while(s < nelem(hashsize) && h->nel > hashsize[s+1])
		s++;
	
	/* 
	 * Shrink if we've gotten way too small.
	 * (Can't happen because there is no way to 
	 * remove elements from the hash table yet.)
	 *
	while(s >= 2 && h->nel < hashsize[s-2])
		s--;
	 */
	
	if(s != h->size)
		rehash(h, s);
}

/*
 * Add key=value to the hash table. 
 * If value==nil, could remove the element.
 */
void*
_hashput(Hash<void*,void*> *h, void *key, void *value)
{
	uint i;
	Xel *e;
	void *old;
	
	assert(h != nil);

	i = H(h, key);
	for(e=h->tab[i]; e; e=e->next){
		if(e->key == key){
			old = e->value;
			e->value = value;
			return old;
		}
	}
	e = mkxel(h->arena, key, value);
	e->next = h->tab[i];
	h->tab[i] = e;
	h->nel++;
	autorehash(h);
	return nil;
}

/* Retrieve value for key from the hash table. */
void*
_hashget(Hash<void*,void*> *h, void *key)
{
	uint i;
	Xel *e;

	/* Convenience: nil pointer is an empty hash table. */
	if(h == nil)
		return nil;

	i = H(h, key);
	for(e=h->tab[i]; e; e=e->next)
		if(e->key == key)
			return e->value;
	return nil;
}

List<void*>*
_hashkeys(Hash<void*,void*> *h)
{
	int i;
	List<void*> *l;
	Xel *x;
	
	l = ZList;
	for(i=0; i<hashsize[h->size]; i++)
		for(x=h->tab[i]; x; x=x->next)
			l = listappend(nil, l, x->key);
	return l;
}

List<void*>*
_hashvalues(Hash<void*,void*> *h)
{
	int i;
	List<void*> *l;
	Xel *x;
	
	l = ZList;
	for(i=0; i<hashsize[h->size]; i++)
		for(x=h->tab[i]; x; x=x->next)
			l = listappend(nil, l, x->value);
	return l;
}

/* Iterate through the hash table. */
int
_hashiterstart(Hashiter<void*,void*> *hi, Hash<void*,void*> *h)
{
	if(h == nil)
		return 0;

	hi->i = 0;
	hi->j = 0;
	hi->h = h;
	return 1;
}

static Xel*
_hashiternext(Hashiter<void*,void*> *hi)
{
	int j;
	Xel *x;
	Hash<void*,void*> *h;
	
	if(hi == nil || hi->h == nil)
		return nil;
	h = hi->h;
	for(;;){
		if(hi->i >= hashsize[hi->h->size])
			return nil;
		for(j=0, x=h->tab[hi->i]; j<hi->j && x!=nil; j++, x=x->next)
			;
		if(x){
			hi->j++;
			return x;
		}
		hi->j = 0;
		hi->i++;
	}
}

int
_hashiternextkey(Hashiter<void*,void*> *hi, void **v)
{
	Xel *x;
	
	x = _hashiternext(hi);
	if(x == nil)
		return 0;
	*v = x->key;
	return 1;
}

int
_hashiternextkv(Hashiter<void*,void*> *hi, Hashkv<void*,void*> *kv)
{
	Xel *x;
	
	x = _hashiternext(hi);
	if(x == nil)
		return 0;
	kv->key = x->key;
	kv->value = x->value;
	return 1;
}

void
_hashiterend(Hashiter<void*,void*> *hi)
{
}
