#include "u.h"
#include "fmt.h"
#include "name.h"
#include "emalloc.h"
#include "panic.h"

enum
{
	Stringtabs = 1024,
	Stringchunk = 1048576,
};

/*
 *	Custom allocators to avoid malloc overheads on small objects.
 * 	We never free these.  (See below.)
 */
typedef struct Stringtab	Stringtab;
struct Stringtab {
	Stringtab *link;
	char *str;
};
static Stringtab*
taballoc(void)
{
	static Stringtab *t;
	static uint nt;

	if(nt == 0){
		t = (Stringtab*)emalloc(Stringtabs*sizeof(Stringtab));
		nt = Stringtabs;
	}
	nt--;
	return t++;
}

static char*
xstrndup(const char *s, int len)
{
	char *r;
	static char *t;
	static int nt;

	len++;
	if(len >= Stringchunk)
		panic("xstrndup big string; %d", len);

	if(nt < len){
		t = (char*)emalloc(Stringchunk);
		nt = Stringchunk;
	}
	r = t;
	t += len;
	nt -= len;
	strncpy(r, s, len-1);
	return r;
}

static char*
xstrdup(const char *s)
{
	return xstrndup(s, strlen(s));
}

/*
 *	Return a uniquely allocated copy of a string.
 *	Don't free these -- they stay in the table for the 
 *	next caller who wants that particular string.
 *	String comparison can be done with pointer comparison 
 *	if you know both strings are atoms.
 */
static Stringtab *stab[1024];

static uint
hash(const char *s)
{
	uint h;
	uchar *p;

	h = 0;
	for(p=(uchar*)s; *p; p++)
		h = h*37 + *p;
	return h;
}

static uint
hashn(const char *s, int len)
{
	uint h;
	uchar *p;

	h = 0;
	for(p=(uchar*)s; len; p++, len--)
		h = h*37 + *p;
	return h;
}

static Name empty = { "" };
static char one[256][2];

Name
nameof(const char *str)
{
	uint h;
	Stringtab *tab;
	
	if(str == nil)
		return Name::null;

	/*
	 * Handle empty and one-char strings quickly.
	 */
	if(str[0] == 0)
		return empty;
	if(str[1] == 0){
		h = str[0];
		one[h][0] = h;
		return namecast(one[h]);
	}
	
	h = hash(str) % nelem(stab);
	for(tab=stab[h]; tab; tab=tab->link)
		if(strcmp(str, tab->str) == 0)
			return namecast(tab->str);

	tab = taballoc();
	tab->str = xstrdup(str);
	tab->link = stab[h];
	stab[h] = tab;
	return namecast(tab->str);
}

Name
nameofn(const char *str, int len)
{
	uint h;
	Stringtab *tab;
	
	if(str == nil)
		return Name::null;

	/*
	 * Handle empty and one-char strings quickly.
	 */
	if(str[0] == 0 || len == 0)
		return empty;
	if(str[1] == 0 || len == 1){
		h = str[0];
		one[h][0] = h;
		return namecast(one[h]);
	}
	
	h = hashn(str,len) % nelem(stab);
	for(tab=stab[h]; tab; tab=tab->link)
		if(strncmp(str, tab->str, len) == 0 && tab->str[len] == '\0')
			return namecast(tab->str);

	tab = taballoc();
	tab->str = xstrndup(str, len);
	tab->link = stab[h];
	stab[h] = tab;
	return namecast(tab->str);
}

Name
nameprint(char *fmt, ...)
{
	va_list arg;
	char *s;
	Name x;

	va_start(arg, fmt);
	s = vsmprint(fmt, arg);
	va_end(arg);
	x = nameof(s);
	free(s);
	return x;
}

Name Name::null;
Name Name::slot;
Name Name::merge;
Name Name::prev;
Name Name::next;

void
Name::init()
{
	slot = nameof("slot");
	merge = nameof("merge");
	prev = nameof("prev");
	next = nameof("next");
}
