#include "grammar.h"
#include "code.h"

typedef struct SlotXForm SlotXForm;
struct SlotXForm
{
	Type *in;
	Type *out;
	Var *fn;
};

static List<SlotXForm*> *restructures;

static Name findrestructure(Type *in, Type **out);

// ========================================================
// Subtyping.
// These functions infer the type of a dotted-name list
// (eg, Gram.expr.rule) rooted at some type context.
// They try to be lenient where it makes sense.

static Type*
lookuptaglist(Typecheck *tc, Type *ctx, List<Name> *names)
{
	Type *t;
	List<Name> *l;

	t = ctx;
	for(l=names; l; l=l->next){
		if((t = lookuptag(t, l->x, tc->scope_)) == nil)
			return nil;
	}
	return t;
}

Type*
infersubtype(Typecheck *tc, Type *ctx, List<Name> *names)
{
	Type *t, *s;
	int atename = 0;

	// Given a subtyping list (eg, Foo.Bar.Baz), determine the
	// type it represents assuming that it is a subtype of 'ctx'.
	// Empty context: get 'ctx' from the first name.
	if(!ctx){
		if(!names)
			return nil;
		if((ctx = lookuptype(tc->scope_, namecast((char*)names->x), 0)) == nil)
			return nil;
		if(ctx->op != TypeStruct && ctx->op != TypeSyntax && ctx->op != TypeUnion)
			return nil;
		names = names->next;
		atename = 1;
	}
	// We try a few different ways of extracting a subtype.
	assert(ctx);
	assert(ctx->op == TypeStruct || ctx->op == TypeSyntax || ctx->op == TypeUnion);
	if(names == ZList)
		return ctx;
	// If this is a syntax object, resolve (G+?) types.
	if(ctx->op == TypeSyntax || ctx->op == TypeUnion){
		ctx = resolveparsetype(tc, ctx);
		assert(ctx);
	}
	// First try: direct tag subtyping from 'ctx', eg
	//    struct T {
	//       tag Foo {};
	//    };
	//    T ~ Foo.xxx
	// ctx   = T
	// names = Foo, xxx
	if((t = lookuptaglist(tc, ctx, names)) != nil)
		return t;
	// Second try: the first name matches 'ctx', eg
	//    T ~ T.Foo.xxx
	// ctx   = T
	// names = T, Foo, xxx
	if((names->x == ctx->name) && (t = lookuptaglist(tc, ctx, names->next)) != nil)
		return t;
	// Third try: the first name is a tag in the tag owner of 'ctx'.
	// This will make the following work:
	//    struct Sym {
	//       tag Var {};
	//       tag Fndef : Var {};
	//    };
	//    v: ptr Sym.Var;
	//    switch(v){
	//    case Var ...
	//    case Fndef ...   // <--- Fndef is a tag of Sym, not Var
	//    }
	// ctx   = Sym.Var
	// names = Fndef
	if(ctx->op == TypeStruct){
		s = lookuptagowner(ctx);
		if(s && (t = lookuptaglist(tc, s, names)) != nil)
			return t;
	}
	// Fourth try: retry recursively try with a nil context.
	// This will make the following work:
	//    a: ptr A.expr;
	//    switch(a){
	//    case ~B.expr ...   // ok iff B <: A
	//    }
	if(!atename && (ctx->op == TypeSyntax || ctx->op == TypeUnion)){
		t = infersubtype(tc, nil, names);
		if(t && issubtypesyntax(ctx, t))
			return t;
	}
	// Failure!
	return nil;
}

Type*
inferparsetype(Typecheck *tc, Type *ctx, List<Name> *names)
{
	Type *t;

	// Infer a parsing type (eg, `Gram.sym.rule{}).  Just a
	// wrapper around infersubtype, but we input and return
	// a (ptr Ast) type.
	if(ctx && !isptrast(ctx))
		return nil;
	// Unfold the ptr
	if(ctx){
		ctx = ctx->left;
		assert(ctx);
	}
	// Make sure to return a ptr
	if((t = infersubtype(tc, ctx, names)) != nil)
		return typeptr(t);
	return nil;
}

Type*
resolveparsetype(Typecheck *tc, Type *raw)
{
	Type *tgram, *tsym, *trule, *full;

	assert(raw);
	// Resolve (G+?) types to a default grammar.
	tgram = gramsupertype(raw);
	tsym  = symsupertype(raw);
	trule = rulesupertype(raw);
	assert(tgram);
	full  = resolvedefaultxgram(tc, tgram);
	assert(full);
	// No change?
	if(full == tgram)
		return raw;
	// Changed, so resolve subtypes.
	if(tsym){
		full = lookuptag(full, tsym->name, tc->scope_);
		assert(full);
	}
	if(trule){
		assert(tsym);
		full = lookuptag(full, trule->name, tc->scope_);
		assert(full);
		// Sanity check in case we ever have G.sym.rule.slot types.
		if(raw->name != full->name){
			assert(raw->super == trule);
			full = lookuptag(full, raw->name, tc->scope_);
			assert(full);
		}
	}
	return full;
}

// ========================================================
// Case patterns.

// Bind a pattern variable, checking for disjunct joins.
static int
bindpatternvar(Typecheck *tc, Line line, Name name, Type *t, Var **var)
{
	PatternCompileState *pcs;
	Type *j;

	pcs = tc->pcs_;
	assert(pcs);
	assert(var);

	// Check for duplicate binding
	for(List<Var*> *l=pcs->boundVars; l; l=l->next) {
		Var *v = l->x;
		if (v->name == name) {
			error(line,
						"duplicate binding of '%s' in pattern",
						name);
			*var = v;
			return 0;
		}
	}
	if(!*var && pcs->joining) {
		int found = 0;
		// This variable must already have been
		// declared on the left side of a
		// disjunction.  Look it up and mark it.
		List<Var*> **prev = &pcs->joinVars;
		for(List<Var*> *l=pcs->joinVars; l; prev=&l->next, l=l->next) {
			Var *v = l->x;
			if(v->name == name) {
				*var = v;
				found = 1;
				// Remove it from joinVars
				*prev = l->next;
				if (pcs->joinVars &&
						pcs->joinVars->end == l)
					pcs->joinVars->end = *prev;
				if (l->arena == nil)
					free(l);
				break;
			}
		}
		if(!found) {
			error(pcs->joining->line,
						"'%s' bound only on right side of %#P",
						name, pcs->joining);
			return 0;
		} else if ((*var)->type != t) {
			// Try joining the types
			j = subtypejoin((*var)->type, t);
			if(j == nil){
				error(line,
							"'%s' used as [%T] on left and [%T] on right of %#P",
							name, (*var)->type, t, pcs->joining);
				return 0;
			}
			(*var)->type = j;
		}
	}
	if(!*var) {
		// We're not joining, or joining gave an error
		*var = declare(tc, line, name, t);
		if (!*var)
			return 0;
	}
	assert(*var);
	pcs->boundVars = listinsert((Arena*)nil, pcs->boundVars, *var);
	return 1;
}

static void
typecheckpattern1(Typecheck *tc, Pattern *pat, Type *ctx)
{
	// Call typecheckpattern1 from this function, not
	// typecheckpattern
#	define typecheckpattern dont_call_typecheckpattern_from_typecheckpattern1
	// Also don't call declare directly.  Instead set isLiteral to
	// 1 and t to the type to declare.  pat->name must be the name
	// of the declared variable.
#	define declare dont_call_declare_from_typecheckpattern1
#	define declareunique dont_call_declareunique_from_typecheckpattern1

	Type *t;
	int isLiteral = 0;
	PatternCompileState *pcs;

	pat->type = ctx;
	t = nil;
	pcs = tc->pcs_;
	assert(pcs);

	switch(pat->op){
	case PatternTag:
		// Should be a subtype of 'ctx', which should be a struct pointer.
		if(ctx->op != TypePtr){
		Badtag:
			error(pat->line, "pattern %P expected discriminated struct pointer or "
			                 "syntax pointer, got [%T]", pat, ctx);
			goto error;
		}
		if(!isptrast(ctx) && ctx->left->op != TypeStruct)
			goto Badtag;
		if((pat->tagtype = infersubtype(tc, ctx->left, pat->tags)) == nil){
			error(pat->line, "bad type for pattern %#P, expected [%T]", pat, ctx);
			goto error;
		}
		typecheckpattern1(tc, pat->subpat, typeptr(pat->tagtype));
		break;
	case PatternDtor: {
		t = ctx;
		// Should be a tuple, a struct, or a ptr to a struct
		if(ctx->op == TypePtr){
			if(ctx->left->op != TypeStruct)
				goto BadDtorPattern;
			t = ctx->left;
		}
		if(t->op != TypeStruct && t->op != TypeTuple)
			goto BadDtorPattern;
		// Check all sub patterns
		List<Pattern*> *l;
		List<Type*> *lt;
		for(l=pat->subpats, lt=t->ctorargs; l && lt;
		    l=l->next, lt=lt->next){
			Pattern *subpat = l->x;
			Type *argtype = lt->x;
			typecheckpattern1(tc, subpat, argtype);
		}
		if(l)
			error(pat->line, "too many arguments to pattern %P", pat);
		else if(lt)
			error(pat->line, "too few arguments to pattern %P", pat);
		break; }
	BadDtorPattern:
		error(pat->line, "pattern %P expected tuple or struct, got [%T]", pat, ctx);
		goto error;
	case PatternEmptyList:
		if(ctx->op != TypeList){
			error(pat->line, "pattern %P expected list, got [%T]", pat, ctx);
			goto error;
		}
		break;
	case PatternList:
		if(ctx->op != TypeList){
			error(pat->line, "pattern %P expected list, got [%T]", pat, ctx);
			goto error;
		}
		typecheckpattern1(tc, pat->subpat, ctx->left);
		typecheckpattern1(tc, pat->subpat2, ctx);
		break;
	case PatternParse:
		pat->valuep->infertype_ = ctx;
		typecheckexpr(tc, pat->valuep, BindValue);
		if(!pat->valuep->type_)
			goto error;
		t = pat->valuep->type_;
		if(pat->name)
			isLiteral = 1;
		break;
	case PatternConst:
		typecheckexpr(tc, pat->value, Value);
		t = pat->value->type_;
		if(t != ctx && (t != typeptr(nil) || ctx->op != TypePtr)){
			error(pat->line, "pattern %P expected [%T], got [%T]", pat, ctx, t);
			goto error;
		}
		break;
	case PatternLit:
		if(pat->isptr)
			t = typeptr(ctx);
		else
			t = ctx;
		assert(strcmp(pat->name, "_") != 0);
		isLiteral = 1;
		break;
	case PatternAny:
		// Binds nothing
		break;
	case PatternDisjunct:
		{
			List<Var*> *alreadyBoundVars = pcs->boundVars;
			// Simply pass through to the left branch
			typecheckpattern1(tc, pat->subpat, ctx);
			// Keep track of which variables were bound by
			// the left branch
			List<Var*> *leftBoundVars = pcs->boundVars;
			// Record the set of variables that need to be
			// joined and remove them from the bound
			// variables so they can be "re-bound" by the
			// right side
			pcs->joinVars = listsplit(leftBoundVars, alreadyBoundVars);
			pcs->boundVars = alreadyBoundVars;
			// Join the right side
			Pattern *oldJoining = pcs->joining;
			pcs->joining = pat;
			typecheckpattern1(tc, pat->subpat2, ctx);
			pcs->joining = oldJoining;
			// Check that all variables were joined
			for(List<Var*> *l=pcs->joinVars; l; l=l->next) {
				Var *v = l->x;
				error(pat->line,
				      "'%s' bound only on left side of %#P",
				      v->name, pat);
			}
			listfree(pcs->joinVars);
			pcs->joinVars = ZList;
		}
		break;
	}

#	undef declare
#	undef declareunique

	if (isLiteral && !bindpatternvar(tc, pat->line, pat->name, t, &pat->var))
		goto error;

	return;

 error:
	pat->type = typevoid();
	if(pat->name && !pat->var)
		pat->var = declare(tc, pat->line, pat->name, pat->type);

#	undef typecheckpattern
}

static void
clearpatternvars(Pattern *pat)
{
	List<Pattern*> *l;

	pat->var = nil;

	switch(pat->op){
	case PatternTag:
		clearpatternvars(pat->subpat);
		break;
	case PatternDtor:
		for(l=pat->subpats; l; l=l->next)
			clearpatternvars(l->x);
		break;
	case PatternList:
		clearpatternvars(pat->subpat);
		clearpatternvars(pat->subpat2);
		break;
	case PatternEmptyList:
	case PatternParse:
	case PatternConst:
	case PatternLit:
	case PatternAny:
		break;
	case PatternDisjunct:
		clearpatternvars(pat->subpat);
		clearpatternvars(pat->subpat2);
		break;
	}
}

// `ctx' is the type of the value being pattern-matched against
void
typecheckpattern(Typecheck *tc, Pattern *pat, Type *ctx)
{
	PatternCompileState pcs, *oldpcs;

	memset(&pcs, 0, sizeof pcs);
	// The typerules may typecheck the same statement multiple times
	// (as it is being rewritten).  The dijsoint pattern checker
	// assumes that pat->var is always nil, so always make an inital
	// pass over the pattern to clear pat->var.
	oldpcs = tc->pcs_;
	tc->pcs_ = &pcs;
	clearpatternvars(pat);
	typecheckpattern1(tc, pat, ctx);
	listfree(pcs.boundVars);
	tc->pcs_ = oldpcs;
}

void
StmtCase::typecheck(Typecheck *tc)
{
	Type *t;
	List<Clause*> *l;

	pushflow(tc);
	typecheckexpr(tc, cond_, Value);
	if((t = cond_->type_) == nil)
		return;
	for(l=clauses_; l; l=l->next){
		Clause *c = l->x;
		pushscope(tc);
		typecheckpattern(tc, c->pat, cond_->type_);
		pushflow(tc);
		typecheckstmt(tc, c->body);
		popflow(tc);
		// XXX This isn't exactly right.  An exhaustive series of
		// XXX patterns, where each clause ends with a 'return',
		// XXX will not propagate reachability outside the case.
		// XXX For now, we assume that the stmt after a case is
		// XXX always reacable.
		tc->reachable_ = 1;
		popscope(tc);
	}
	popflow(tc);
}

// ========================================================
// Parse patterns.

static int
typecheckslottoken(Typecheck *tc, Type *tgram, PToken *pt, int kind, int isview)
{
	Type *t, *tt, *tsym, *tout;
	Expr *e;
	ExprName *en;
	Name name;  // element name for lists, or termname for non-lists
	Name fn;
	const char *str;

	assert(pt->op == PTokenExpr);
	e = pt->expr;
	name = pt->symname;

	en = e->exprname();
	if(en && strcmp(en->name_, "_") == 0){
		if(kind&Value){
			error(e->line_, "'don't care' in restructure");
			return 0;
		}
		e->type_ = typedontcare();
		return 1;
	}

	if(kind&BindValue){
		if(!e->exprname()){
			error(e->line_, "unbindable expression %#E", e);
			return 0;
		}
		return 1;
	}

	// Type check this slot expression and fill in the
	// 'symname', 'elemname', and 'islist' fields of pt
	// for the benefit of convertptokens().
	typecheckexpr(tc, e, kind);
	t = e->type_;
	if(t == nil){
		error(e->line_, "bad escape %#E [%T]", e, t);
		return 0;
	}

	// Extract the element type of lists.
	if(t->op == TypeList){
		// parse the ::elemname(*|+|[]) annotation 
		if(pt->symname){
			if(!isastlistsymname(pt->symname, &name)){
				error(e->line_, "bad type %s for list slot %#E", pt->symname, e);
				return 0;
			}
			pt->elemname = name;
		}
		pt->islist = 1;
		t = t->left;
	}

	// TODO: support destructure?
	//   (this is harder because we'd really want 'destructure' conversions to
	//    act as a predicate, and possibly fail if the conversion is impossible)

	// Restructure conversions
	if(kind == Value && !isview && !pt->symname && (fn = findrestructure(t, &tout)) != nil){
		vector<Expr*> v;
		ExprCall *ec;
		if(pt->islist){
			Expr *n;
			n = new ExprName(fn);
			n->line_ = e->line_;
			v.push_back(e);
			v.push_back(n);
			pt->expr = ec = new ExprCall(new ExprName(nameof("listmap")), v);
		}
		else{
			v.push_back(e);
			pt->expr = ec = new ExprCall(new ExprName(fn), v);
		}
		ec->line_ = e->line_;
		ec->left_->line_ = e->line_;
		e = pt->expr;
		typecheckexpr(tc, e, kind);
		assert(e->type_);
		if(pt->islist){
			assert(e->type_->op == TypeList);
			t = e->type_->left;
		}
		else
			t = e->type_;
		assert(t == tout);
	}

	// Must be ptr Ast
	if(!isptrast(t)){
		error(e->line_, "bad escape %#E [%T]", e, t);
		return 0;
	}
	t = t->left;

	// Restructure patterns: slot type <: grammar type.
	if(kind == Value && !issubtypesyntax(tgram, t)){
		error(e->line_, "bad type %T for slot %#E, expected <: [%T]", t, e, tgram);
		return 0;
	}
	// Destructure patterns: slot type :> grammar.sym type, for some 'sym'.
	if(kind == LValue){
		// The type must be a sym or grammar type.
		// XXX in theory we could support rule types also, eg,
		//       x: ptr C.expr.Plus;
		//       e ~ {if(\x) \s}
		//     in fact it might "just work" ...
		tt = t;
		if(tt->super){
			tt = tt->super;
			if(tt->super)
				goto BadLValue;
		}
		if(!issubtypesyntax(tt, tgram)){
		BadLValue:
			str = pt->symname ? (char*)pt->symname : "";
			error(e->line_, "bad type [%T] for slot %#E, expected :> [%T].%s", t, e, tgram, str);
			return 0;
		}
	}

	// If a ::symname annotation was given, check it.
	// 'name' is the expected symname or elemname (for lists),
	// and 't' is the actual symtype or elemtype (for lists).
	tsym = symsupertype(t);
	if(name && tsym && name != tsym->name
	&& (!tsym->mirrors || name != tsym->mirrors->name)){
		error(e->line_, "expected type [%s] for slot %#E did not match actual type [%T]",
		               pt->symname, e, t);
		return 0;
	}
	// If no annotation was given, provide one from the computed type.
	if(pt->symname == nil && (t = symsupertype(t)) != nil){
		if(pt->islist)
			pt->elemname = t->name;
		else
			pt->symname = t->name;
	}
	return 1;
}

void
ExprParse::typecheck(Typecheck *tc, int kind)
{
	PToken *pt;
	List<PToken*> *l;
	Expr *e;
	Type *t, *tgram, *tstart;

	// kind is set to what the subexpressions need to be- Value, LValue, or BindValue.
	type_ = nil;
	canonicals_ = tc->scope_->canonicals;

	// Infer the parsing type for this pattern.
	t = inferparsetype(tc, infertype_, names_);

	// If that failed, try again with a nil context.
	// This makes the following work:
	//    T: ptr Type.Ptr;
	//    T = `Type{\b *};
	// This fails the initial inferparsetype() call because we
	// can only infer the pattern type "Type"; the more specific
	// type "Type.Ptr" will be inferred after we parse the pattern
	// and discover its starting rule.
	if(!t)
		t = inferparsetype(tc, nil, names_);

	// Could we find a starting type?
	if(!t){
		error(line_, "bad type for syntax pattern %#E in context [%T]",
		                  this, infertype_);
		return;
	}
	assert(isptrast(t));

	// Need to make sure we've resolved all wildcard types.
	tstart = resolveparsetype(tc, t->left);
	if(!tstart || !tstart->grammar){
		error(line_, "type [%T] is not parsable", t);
		return;
	}
	tgram = gramsupertype(tstart);
	assert(tgram);
	type_ = typeptr(tstart);

	// Typecheck all pattern tokens and convert to an expression list.
	// TODO: check for duplicate names in a destructure (and restructure?) pattern
	exprs_.clear();
	for(l=ptokens_; l; l=l->next){
		pt = l->x;
		switch(pt->op){
		case PTokenString:
			break;
		case PTokenExpr:
			if(!typecheckslottoken(tc, tgram, pt, kind, tstart->mirrors != nil))
				break;
			if(pt->expr->type_ != typedontcare())
				exprs_.push_back(pt->expr);
			break;
		}
	}

	// Parse the syntax pattern.
	if(!parsepattern(this, kind&BindValue)){
		type_ = nil;
		return;
	}

	// Now that we know their types, declare all slot variables.
	if(kind&BindValue){
		for(l=ptokens_; l; l=l->next){
			pt = l->x;
			e = pt->expr;
			if(pt->op != PTokenExpr || e->type_ == typedontcare())
				continue;
			ExprName *en = e->exprname();
			if(en == nil)
				continue;
			// grammarparse.c set e->type when it deduced the type.
			en->var_ = nil;
			if(!bindpatternvar(tc, e->line_, pt->slot, e->type_, &en->var_))
				type_ = nil;
		}
	}
}

// ========================================================
// Structuring expressions.

void
ExprDestructure::typecheck(Typecheck *tc, int kind)
{
	// a ~ sym{...}
	noassign(tc, this, kind);
	Expr *l = left_;
	typecheckexpr(tc, l, Value);
	if(!isptrast(l->type_)){
		error(line_, "bad type for %#E [%T]", l, l->type_);
		return;
	}
	Expr *r = right_;
	r->infertype_ = l->type_;
	typecheckexpr(tc, r, LValue);
	if(r->type_ != nil){
		type_ = typebool();
		addtyperefinement(tc, l, r->type_);
	}
}

void
ExprRestructure::typecheck(Typecheck *tc, int kind)
{
	ExprParse *l;
	Type *t;

	// `sym{...}
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	l = left_;
	t = nil;
	if(isptrast(infertype_))
		l->infertype_ = infertype_;
	typecheckexpr(tc, l, Value);
	if(l->type_ == nil)
		return;
	if(right_){
		typecheckexpr(tc, right_, Value);
		if(!isptrast(right_->type_)){
			error(line_, "need ast in @%#E", right_);
		}
	}
	type_ = l->type_;
}

void
ExprDestructType::typecheck(Typecheck *tc, int kind)
{
	Type *t, *tl;

	// a ~ type
	// This checks dynamically if 'type' <: typeof(a), and
	// works for any struct type (not just Asts).
	noassign(tc, this, kind);
	typecheckexpr(tc, left_, Value);
	tl = left_->type_;
	if(tl == nil || tl->op != TypePtr){
		error(line_, "bad type for %#E [%T]", left_, tl);
		return;
	}
	tl = tl->left;
	t = infersubtype(tc, tl, names_);
	if(!t){
		error(line_, "bad type for %#E", this);
		return;
	}
	xtype_ = t;
	type_ = typebool();
	addtyperefinement(tc, left_, typeptr(t));
}

void
ExprParseString::typecheck(Typecheck *tc, int kind)
{
	Expr *e;
	Type *t;

	assert(exprs_.size());
	// `sym(text)
	// `sym(text, file)
	// `slot(slotvar)
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	t = nil;
	if(isptrast(infertype_))
		t = infertype_;
	type_ = inferparsetype(tc, t, names_);
	if(type_ == nil || type_->left == nil){
		error(line_, "couldn't determine symtype for %#E", this);
		return;
	}
	t = type_->left;
	// Don't allow views.
	if(t->mirrors){
		error(line_, "can't parse view [%T] (in %#E)", t, this);
		goto Bad;
	}
	// Check the args
	if(strcmp(t->name, "slot") == 0){
		// `slot(slotvar)
		if(exprs_.size() != 1){
		Badslot:
			error(line_, "bad slot constructor %#E", this);
			goto Bad;
		}
		e = exprs_[0];
		typecheckexpr(tc, e, Value);
		if(!isptrast(e->type_))
			goto Badslot;
		if(strcmp(e->type_->left->name, "slot") != 0)
			goto Badslot;
	}
	else{
		// `sym(text)
		// `sym(text, file)
		if(symsupertype(t) == nil){
			error(line_, "couldn't determine start symbol for %#E", this);
			goto Bad;
		}
		if(exprs_.size() > 2){
			error(line_, "too many arguments for parser %#E (expected 1 or 2)", this);
			goto Bad;
		}
		for(int i=0; i<exprs_.size(); i++){
			e = exprs_[i];
			typecheckexpr(tc, e, Value);
			if(e->type_ != typestring()){
				error(line_, "bad type [%T], expected string", e->type_);
				goto Bad;
			}
		}
	}
	return;

Bad:
	type_ = nil;
	return;
}

void
ExprAstParent::typecheck(Typecheck *tc, int kind)
{
	assert(left_);
	assert(names_);
	// astparent(ast, G.sym.rule)
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	typecheckexpr(tc, left_, Value);
	if(left_->type_ && !isptrast(left_->type_)){
		error(line_, "bad type for %#E [%T]", left_, left_->type_);
		return;
	}
	type_ = inferparsetype(tc, nil, names_);
	if(type_ == nil || type_->left == nil){
		error(line_, "couldn't determine parent type for %#E", this);
		return;
	}
}

// ========================================================
// Canonicalizers.

void
StmtCanonicalize::typecheck(Typecheck *tc)
{
	// Must have the form G.sym.attribute.
	// XXX Could relax this to G.attribute and G.sym.rule.attribute.
	if(listlen(names_) != 3){
	BadCanon:
		error(line_, "bad canonicalizer: expected G.sym.attribute");
		return;
	}
	// Pick off the attr name.
	// Why oh why does this fail?
	// Name name = listnth(names_, 2);
	Name name = names_->next->next->x;
	names_->next->next = nil;
	// Infer the symbol type.
	Type *host = inferparsetype(tc, nil, names_);
	if(!host){
		error(line_, "bad canonicalizer: could not infer a symbol type");
		return;
	}
	assert(host->op == TypePtr);
	host = host->left;
	// Grammar must be a subtype of (G+?), since it can
	// be given any arbitrary syntax fragment in that family.
	Gram *g = host->grammar;
	if(!g || !host->super)
		goto BadCanon;
	if(!issubtypesyntax(g->wildroottype, host)){
		error(line_, "bad canonicalizer: must be a subtype of [%T]", g->wildroottype);
		return;
	}
	// Check the attribute.
	Var *v = lookupattr(host, name, tc->scope_, 0);
	if(!v){
		error(line_, "cannot find attribute %s in [%T]", (char*)name, host);
		return;
	}
	assert(v->type);
	assert(v->type->op == TypePtr);
	if(v->type->left != host){
		error(line_, "bad return type for canonicalizer, expected [%T]", host);
		return;
	}
	tc->scope_->canonicals = listinsert(nil, tc->scope_->canonicals, v);
}

// ========================================================
// Slot transformations.
//
// TODO: destructuring
//
// This is a general mechanism for converting
// arbitrary types into Asts for restructuring,
// and for converting Asts into arbitrary types
// for destructuring.  For example,
//
//    int -> ptr C.number  (restructuring)
//    ptr C.number -> int  (destructuring)
//
// and then you could do:
//
//    n := 10;
//    `C.expr{\n}
//
// and 
//
//    n: int;
//    if(e ~ C.expr{\n}){
//       ...
//    }
//
// (In the latter case, the \n is implicitly \n::number.)

void
addrestructure(Line line, Name name)
{
	Var *v;
	SlotXForm *xf;
	Type *in;
	Name old;

	if((v = lookup(globalscope, name, 0)) == nil){
		error(line, "'%s' doesn't exist", name);
		return;
	}
	if(v->type == nil || v->type->op != TypeFn){
		error(line, "%s is a [%T], expected a function", name, v->type);
		return;
	}
	if(listlen(v->type->types) != 1 || !isptrast(v->type->left)){
		error(line, "%s is a [%T], expected (T) -> ptr Ast", name, v->type);
		return;
	}
	in = listpeek(v->type->types);
	if((old = findrestructure(in, nil)) != nil)
		warn(line, "restructure '%s' already defined for [%T] (overriding)", old, in);
	xf = (SlotXForm*)emalloc(sizeof *xf);
	xf->fn = v;
	xf->in = in;
	xf->out = v->type->left;
	restructures = listinsert(nil, restructures, xf);
}

static Name
findrestructure(Type *in, Type **out)
{
	List<SlotXForm*> *l;
	SlotXForm *xf;

	if(out)
		*out = nil;
	for(l=restructures; l; l=l->next){
		xf = l->x;
		if(issubtype(xf->in, in)){
			if(out)
				*out = xf->out;
			return xf->fn->name;
		}
	}
	return Name::null;
}

// Emit a triple (ast, nslots, nil) as part of a call
// to xdestructure or xrestructure.  Also remember if
// this pattern is view-transformed, so we can apply
// the actual transformations later.
void
emitpatternast(Compile *cs, ExprParse *expr)
{
	PatternXForm *xf;

	assert(expr->ast_);

	if(!expr->type_->left->view){
		cs->emitp(expr->ast_);
		cs->emiti(expr->exprs_.size());
		cs->emitp(nil);          // no view
		cs->emitp(nil);          // placeholder, used by patterns xforms
		return;
	}
	xf = (PatternXForm*)emalloc(sizeof *xf);
	xf->codebuf = cs->codebuf_;
	xf->off  = cs->codebuf_->p - cs->codebuf_->bp;
	xf->view = expr->type_->left->view;
	xf->line = expr->line_;
	queuedpatternxforms = listappend(nil, queuedpatternxforms, xf);
	cs->emitp(expr->ast_);
	cs->emiti(expr->exprs_.size());
	cs->emitp(xf->view);
	cs->emitp(nil);           // placeholder
}

// Emit code for xcancastast, if necessary.
// This is a special version if xissubtype for Asts.
// It checks dynamically if specifictype(var) <: tsuper.
// Returns non-zero if code was emitted.
int
emitcancastast(Compile *cs, Line line, Loc *dst, Loc *val, Type *tsuper)
{
	Type *tval;

	assert(val->type->op == TypePtr);
	tval = val->type->left;

	if(tval->op == TypeStruct || tsuper->op == TypeStruct)
		return 0;

	assert(isptrast(val->type));
	assert(isptrast(typeptr(tsuper)));
	cs->emitCLPL(line, xcancastast, dst, tsuper, val);
	return 1;
}

// Emit code for xissubtype
// This checks dynamically if typeof(val) <: tsuper.
// It is known statically that tsuper <: tval, where tval is
// the static type of val.
void
emitissubtype(Compile *cs, Line line, Loc *dst, Loc *val, Type *tsuper, int checknil)
{
	Loc *lsuper, *lvaltype, *tmp;

	// Try an Ast check first.
	if(emitcancastast(cs, line, dst, val, tsuper))
		return;

	// If 'checknil' is set, we need to check 'val' for nil first
	if(checknil){
		tmp = cs->mktmp(typeint());          // XXX typeptr()
		cs->emitloadptr(line, tmp, nil);
		cs->emitCLLL(line, xnei, dst, val, tmp);
	}
	// Otherwise, load '1' into dst to mark that 'val' is known to be non-nil
	else{
		assert(sizeoftype(dst->type) == sizeoftype(typeint()));
		cs->emitloadint(line, dst, 1);
	}
	lvaltype = cs->offset(cs->deref(line, val), tagtypeoffset(val->type->left));
	lvaltype->type = typeptr(nil);
	lsuper = cs->mktmp(typeint());          // XXX typeptr()
	cs->emitloadptr(line, lsuper, tsuper);
	cs->emitCLLL(line, xissubtype, dst, lsuper, lvaltype);
}

Loc*
ExprDestructure::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	// Evaluate exprs into successive tmp slots.
	static Loc locx;
	memset(&locx, 0, sizeof locx);
	Loc *loc = &locx;
	Loc *src = nil;
	if(dstatomic == nil)
		dstatomic = cs->mktmp(typebool());
	ExprParse *east = right_;
	src = cs->evaltmp(left_);

	// Construct arguments.  Relies on the fact that mktmp
	// will allocate temporaries in order, so that successive
	// calls to mktmp will allocate an array of temporaries.
	// Loc ends up pointing at the first one.
	for(uint j=0; j<east->exprs_.size(); j++){
		Expr *e = east->exprs_[j];
		Loc *tmp = cs->mktmp(e->type_);
		if(loc == &locx)
			loc = tmp;
		cs->pushtmp();
		cs->leaexpr(e, tmp);
		cs->poptmp();
	}
	cs->emitCLL(line_, xdestructure, dstatomic, src);
	// Is this a horizontal downcast?
	Type *tl = left_->type_;
	Type *tr = right_->type_;
	if(!isverticalsubtype(tl->left, tr->left))
		cs->emitp(tr->left);  // yes, we need the dynamic check
	else
		cs->emitp(nil);
	emitpatternast(cs, east);
	cs->emitloc(loc);
	cs->emitp(right_->canonicals_);
	return dstatomic;
}

Loc*
ExprRestructure::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	// Evaluate exprs into successive tmp slots.
	Loc locx;
	memset(&locx, 0, sizeof locx);
	Loc *loc = &locx;
	if(dstatomic == nil)
		dstatomic = cs->mktmp(typebool());
	ExprParse *east = left_;

	// Construct arguments.  Relies on the fact that mktmp
	// will allocate temporaries in order, so that successive
	// calls to mktmp will allocate an array of temporaries.
	// Loc ends up pointing at the first one.
	for(uint j=0; j<east->exprs_.size(); j++){
		Expr *e = east->exprs_[j];
		Loc *tmp = cs->mktmp(e->type_);
		if(loc == &locx)
			loc = tmp;
		cs->pushtmp();
		cs->evalinto(e, tmp);
		cs->poptmp();
	}
	Loc *tmp;
	if(right_)
		tmp = cs->evaltmp(right_);
	else
		tmp = nil;
	cs->emitCL(line_, xrestructure, dstatomic);
	cs->emitloc(tmp);
	emitpatternast(cs, east);
	cs->emitloc(loc);
	return dstatomic;
}

Loc*
ExprParseString::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	if(dstatomic == nil)
		return nil;
	Type *t = type_->left;
	if(strcmp(t->name, "slot") == 0){
		Loc *tmp = cs->evaltmp(exprs_[0]);
		cs->emitCLPL(line_, xrestructslot, dstatomic, type_->left, tmp);
	}
	else{
		Loc *tmp[2];
		int i;
		for(i=0; i<exprs_.size() && i<2; i++)
			tmp[i] = cs->evaltmp(exprs_[i]);
		for(; i<2; i++)
			tmp[i] = nil;
		cs->emitCLPLL(line_, xparse, dstatomic, type_->left,
		          tmp[0], tmp[1]);
	}
	return dstatomic;
}

Loc*
ExprDestructType::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	if(dstatomic == nil)
		return nil;
	emitissubtype(cs, line_, dstatomic, cs->evaltmp(left_), xtype_, 1);
	return dstatomic;
}

Loc*
ExprSameStructure::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	if(dstatomic == nil){
		compileexpr(cs, left_);
		compileexpr(cs, right_);
		return nil;
	}
	Loc *tmp = cs->evaltmp(left_);
	Loc *tmp1 = cs->evaltmp(right_);
	cs->emitCLLL(line_, xsamestructure, dstatomic, tmp, tmp1);
	return dstatomic;
}

Loc*
ExprAstParent::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	if(dstatomic == nil)
		return nil;
	cs->emitCLLP(line_, xastparent, dstatomic, cs->evaltmp(left_), type_);
	return dstatomic;
}
