// A quick and dirty pretty printer for zeta.

#include "a.h"

static void pprintexprs(Fmt *f, const vector<Expr*>& exprs);

// Fmt.width is used as the current indentation level.
// This print preserved the .width parameter.
int
pprintf(Fmt *f, char *fmt, ...)
{
	va_list arg;
	int n, w;

	w = f->width;
	va_start(arg, fmt);
	n = fmtvprint(f, fmt, arg);
	va_end(arg);
	f->width = w;
	return n;
}

static void
pprintindent(Fmt *f)
{
	int i;
	for(i=0; i < f->width; i++)
		pprintf(f, "  ");
}

static void
pprintdotnames(Fmt *f, List<Name> *names)
{
	char *sep = "";

	for(; names; names=names->next){
		pprintf(f, "%s%s", sep, (char*)names->x);
		sep = ".";
	}
}

// Print a blocked stmt (branch of an if or loop)
static void
pprintblock(Fmt *f, Stmt *stmt)
{
	if(stmt == nil){
		pprintf(f, "{\n");
		pprintindent(f);
		pprintf(f, "}\n");
		return;
	}
	
	stmt->printblock(f);
}

void
Stmt::printblock(Fmt *f)
{
	pprintf(f, "\n");
	f->width++;
	pprintindent(f);
	pprintstmt(f, this);
	f->width--;
	pprintf(f, "\n");
}

void
StmtList::printblock(Fmt *f)
{
	int n = listlen(stmts_);
	if(n == 0){
		pprintf(f, "{\n");
		pprintindent(f);
		pprintf(f, "}\n");
	}
	else if(n == 1){
		pprintf(f, "\n");
		f->width++;
		pprintindent(f);
		pprintstmt(f, listpeek(stmts_));
		f->width--;
		pprintf(f, "\n");
	}
	else{
		pprintf(f, "{\n");
		f->width++;
		for(List<Stmt*> *l=stmts_; l; l=l->next){
			pprintindent(f);
			pprintstmt(f, l->x);
			pprintf(f, "\n");
		}
		f->width--;
		pprintindent(f);
		pprintf(f, "}\n");
	}
}

static void
pprintfn(Fmt *f, Fn *fn, List<Name> *typevars)
{
	Fnarg *a;

	pprintindent(f);
	pprintf(f, "fn %s", fn->name);
	if(typevars){
		pprintf(f, "<");
		for(List<Name> *l=typevars; l; l=l->next)
			pprintf(f, "%s%s", l->x, l->next ? ", " : "");
		pprintf(f, ">");
	}
	pprintf(f, "(");
	for(List<Fnarg*> *l=fn->fnargs; l; l=l->next){
		a = l->x;
		pprintf(f, "%s: %T%s", a->name, a->type, l->next ? ", " : "");
	}
	pprintf(f, ") : %T\n", fn->rettype);
	if(f->width == 0)
		pprintindent(f);
	pprintf(f, "{\n");
	f->width++;
	for(List<Stmt*> *l=fn->stmts; l; l=l->next){
		pprintindent(f);
		pprintstmt(f, l->x);
		pprintf(f, "\n");
	}
	f->width--;
	pprintindent(f);
	pprintf(f, "}\n");
}

// Print a stmt.  Assume indentation is already printed.
void
pprintstmt(Fmt *f, Stmt *stmt)
{
	if(stmt == nil){
		pprintf(f, "<nil-stmt>");
		return;
	}
	stmt->pprint(f);
}

void
StmtExpr::pprint(Fmt *f)
{
	if(expr_){
		if(flags_ & DeclareConst)
			pprintf(f, "const ");
		pprintf(f, "%#E;", expr_);
	}
	else
		pprintf(f, ";");
}

void
StmtFor::pprint(Fmt *f)
{
	pprintf(f, "for(%#E; %#E; %#E)", pre_, cond_, post_);
	pprintblock(f, body_);
}

void
StmtForList::pprint(Fmt *f)
{
	pprintf(f, "for(%#E in %#E)", expr_, cond_);
	pprintblock(f, body_);
}

void
StmtWhile::pprint(Fmt *f)
{
	pprintf(f, "while(%#E)", cond_);
	pprintblock(f, body_);
}

void
StmtDoWhile::pprint(Fmt *f)
{
	pprintf(f, "do");
	pprintblock(f, body_);
	pprintf(f, "while(%#E);", cond_);
}

void
StmtIf::pprint(Fmt *f)
{
	pprintf(f, "if(%#E)", cond_);
	pprintblock(f, left_);
	if(right_){
		pprintindent(f);
		pprintf(f, "else");
		pprintblock(f, right_);
	}
}

void
StmtBreak::pprint(Fmt *f)
{
	pprintf(f, "break;");
}

void
StmtContinue::pprint(Fmt *f)
{
	pprintf(f, "continue;");
}

void
StmtList::pprint(Fmt *f)
{
	pprintblock(f, this);
}

void
StmtPrint::pprint(Fmt *f)
{
	pprintf(f, "print %#E;", expr_);
}

void
StmtReturn::pprint(Fmt *f)
{
	if(expr_)
		pprintf(f, "return %#E;", expr_);
	else
		pprintf(f, "return;");
}

void
StmtGoto::pprint(Fmt *f)
{
	pprintf(f, "goto %s;", name_);
}

void
StmtLabel::pprint(Fmt *f)
{
	pprintf(f, "label %s;", name_);
}

void
StmtDeclare::pprint(Fmt *f)
{
	pprintf(f, "%s : %T;", name_, type_);
}

void
StmtDeclareAttribute::pprint(Fmt *f)
{
	pprintf(f, "attribute ");
	pprintdotnames(f, names_);
	pprintf(f, " :: %s : %T;", name_, type_);
}

void
StmtDeclareType::pprint(Fmt *f)
{
	pprintf(f, "typedef %s : %T;", name_, type_);
}

void
StmtFndef::pprint(Fmt *f)
{
	if(flags_ & FnExtend)
		pprintf(f, "extend ");
	else if(flags_ & FnExtensible)
		pprintf(f, "extensible ");
	if(flags_ & FnAttribute)
		pprintf(f, "attribute ");
	pprintfn(f, fn_, typevars_);
}

void
StmtCase::pprint(Fmt *f)
{
	pprintf(f, "switch(%#E){\n", cond_);
	for(List<Clause*> *l=clauses_; l; l=l->next){
		Clause *c = l->x;
		pprintindent(f);
		pprintf(f, "case %#P:", c->pat);
		pprintblock(f, c->body);
	}
	pprintindent(f);
	pprintf(f, "}");
}

// Not implemented
void
StmtStruct::pprint(Fmt *f)
{
	pprintf(f, "<StmtStruct>");
}


static void
pprintexprs(Fmt *f, const vector<Expr*>& exprs)
{
	for(int i=0; i<exprs.size(); i++)
		pprintf(f, "%s%#E", i ? ", " : "", exprs[i]);
}

// Print an expression.
void
ExprCode::pprint(Fmt *f)
{
	if(right_)
		pprintf(f, "%#E %s %#E", left_, oper(), right_);
	else
		pprintf(f, "%s %#E", oper(), left_);
}

void
ExprName::pprint(Fmt *f)
{
	pprintf(f, "%s", name_);
}

void
ExprArray::pprint(Fmt *f)
{
	pprintf(f, "%#E[%#E]", left_, right_);
}
void
ExprSlice::pprint(Fmt *f)
{
	pprintf(f, "%#E[%#E,%#E]", left_, middle_, right_);
}

void
ExprTuple::pprint(Fmt *f)
{
	pprintf(f, "(");
	pprintexprs(f, exprs_);
	pprintf(f, ")");
}

void
ExprCall::pprint(Fmt *f)
{
	pprintf(f, "%#E(", left_);
	pprintexprs(f, exprs_);
	pprintf(f, ")");
}

void
ExprNil::pprint(Fmt *f)
{
	pprintf(f, "nil");
}
void
ExprBool::pprint(Fmt *f)
{
	pprintf(f, "%s", b_ ? "true" : "false");
}
void
ExprInteger::pprint(Fmt *f)
{
	pprintf(f, "%d", i_);
}
void
ExprFloat::pprint(Fmt *f)
{
	pprintf(f, "%f", f_);
}

void
ExprString::pprint(Fmt *f)
{
	pprintf(f, "\"%s\"", str_);
}

void
ExprDot::pprint(Fmt *f)
{
	pprintf(f, "%#E.%s", left_, name_);
}

void
ExprConvert::pprint(Fmt *f)
{
	pprintf(f, "(%T)(%#E)", type_, left_);
}
void
ExprLength::pprint(Fmt *f)
{
	pprintf(f, "length(%#E)", left_);
}
void
ExprReverse::pprint(Fmt *f)
{
	pprintf(f, "reverse(%#E)", left_);
}
void
ExprAstParent::pprint(Fmt *f)
{
	pprintf(f, "astparent(");
	pprintdotnames(f, names_);
	pprintf(f, ")");
}
void
ExprTag::pprint(Fmt *f)
{
	pprintf(f, "tag(%#E)", left_);
}
void
ExprFn::pprint(Fmt *f)
{
	pprintfn(f, fn_, typevars_);
}

void
ExprDestructType::pprint(Fmt *f)
{
	pprintf(f, "%#E ~ ", left_);
	pprintdotnames(f, names_);
}

void
ExprParseString::pprint(Fmt *f)
{
	pprintf(f, "`");
	pprintdotnames(f, names_);
	pprintf(f, "(");
	pprintexprs(f, exprs_);
	pprintf(f, ")");
}

void
ExprParse::pprint(Fmt *f)
{
	pprintdotnames(f, names_);
	pprintf(f, "{");
	for(List<PToken*> *l=ptokens_; l; l=l->next){
		PToken *gt = l->x;
		if(gt->op == PTokenString)
			pprintf(f, "%s", gt->str);
		else if(gt->expr != nil){
			ExprName *en = gt->expr->exprname();
			if(en)
				pprintf(f, "\\%s", en->name_);
			else
				pprintf(f, "\\(%#E)", gt->expr);
			if(gt->symname)
				pprintf(f, "::%s", gt->symname);
		}
		else
			pprintf(f, "!NIL!");
		if(l->next != nil)
			pprintf(f, " ");
	}
	pprintf(f, "}");
}

void
ExprInit::pprint(Fmt *f)
{
	pprintf(f, "%#E(", left_);
	pprintexprs(f, *(vector<Expr*>*)&exprs_);
	pprintf(f, ")");
}
void
ExprInitField::pprint(Fmt *f)
{
	pprintf(f, ".%s = %#E", name_, left_);
}
void
ExprMkListOrArray::pprint(Fmt *f)
{
	pprintf(f, "[");
	pprintexprs(f, exprs_);
	pprintf(f, "]");
}
void
ExprMkList::pprint(Fmt *f)
{
	pprintf(f, "list [");
	pprintexprs(f, exprs_);
	pprintf(f, "]");
}
void
ExprMkArray::pprint(Fmt *f)
{
	pprintf(f, "array [");
	pprintexprs(f, exprs_);
	pprintf(f, "]");
}
void
ExprMkMap::pprint(Fmt *f)
{
	pprintf(f, "{");
	pprintexprs(f, *(vector<Expr*>*)&exprs_);
	pprintf(f, "}");
}
void
ExprMap::pprint(Fmt *f)
{
	pprintf(f, "%#E => %#E", left_, right_);
}

void
ExprRestructure::pprint(Fmt *f)
{
	pprintf(f, "`%#E", left_);
}

void
ExprDestructure::pprint(Fmt *f)
{
	pprintf(f, "%#E ~ %#E", left_, right_);
}

void
ExprEq::pprint(Fmt *f)
{
	pprintf(f, "%#E = %#E", left_, right_);
}

void
ExprEqEq::pprint(Fmt *f)
{
	pprintf(f, "%#E == %#E", left_, right_);
}

void
ExprNotEq::pprint(Fmt *f)
{
	pprintf(f, "%#E != %#E", left_, right_);
}

void
ExprSameStructure::pprint(Fmt *f)
{
	pprintf(f, "%#E ~~ %#E", left_, right_);
}

void
ExprListCons::pprint(Fmt *f)
{
	pprintf(f, "%#E # %#E", left_, right_);
}

void
ExprNew::pprint(Fmt *f)
{
	pprintf(f, "new %#E", left_);
}

void
ExprDeclEq::pprint(Fmt *f)
{
	pprintf(f, "%#E := %#E", left_, right_);
}

void
ExprOrOr::pprint(Fmt *f)
{
	pprintf(f, "%#E || %#E", left_, right_);
}

void
ExprAndAnd::pprint(Fmt *f)
{
	pprintf(f, "%#E && %#E", left_, right_);
}

void
ExprAddr::pprint(Fmt *f)
{
	pprintf(f, "&%#E", left_);
}

void
ExprStar::pprint(Fmt *f)
{
	pprintf(f, "*%#E", left_);
}

void
pprintexpr(Fmt *f, Expr *expr)
{
	if(expr == nil){
		pprintf(f, "<nil-expr>");
		return;
	}
	expr->pprint(f);
}

// Print a pattern.
void
pprintpattern(Fmt *f, Pattern *pat)
{
	List<Pattern*> *l;

	switch(pat->op){
	case PatternTag:
		pprintdotnames(f, pat->tags);
		pprintf(f, " ");
		pprintpattern(f, pat->subpat);
		break;
	case PatternDtor:
		pprintf(f, "(");
		for(l=pat->subpats; l; l=l->next){
			pprintpattern(f, l->x);
			if(l->next)
				pprintf(f, ", ");
		}
		pprintf(f, ")");
		break;
	case PatternEmptyList:
		pprintf(f, "[]");
		break;
	case PatternList:
		pprintpattern(f, pat->subpat);
		pprintf(f, " # ");
		pprintpattern(f, pat->subpat2);
		break;
	case PatternParse:
		pprintf(f, "~%#E", pat->value);
		if(pat->name)
			pprintf(f, " %s", pat->name);
		break;
	case PatternConst:
		pprintf(f, "%#E", pat->value);
		break;
	case PatternLit:
		if(pat->isptr)
			pprintf(f, "* ");
		pprintf(f, "%s", pat->name);
		break;
	case PatternAny:
		pprintf(f, "_");
		break;
	case PatternDisjunct:
		pprintpattern(f, pat->subpat);
		pprintf(f, " || ");
		pprintpattern(f, pat->subpat2);
		break;
	}
}

void
ExprIncDec::pprint(Fmt *f)
{
	const char *text;
	
	if(inc_ < 0)
		text = "--";
	else
		text = "++";
	if(post_)
		pprintf(f, "%#E%s", left_, text);
	else
		pprintf(f, "%s%#E", text, left_);
}

