// Runtime execution of virtual machine.
// See also opcodes in code.c.

#include "a.h"
#include <sys/signal.h>
#include <sys/time.h>
#include <time.h>

DEFINE_bool(steptrace, false, "Print every bytecode as it executes.");
DEFINE_bool(catchsegv, true, "Catch SIGSEGV and print zeta state.");

enum {
	FnIndex = FrameOffsetFN / sizeof(void*),
	FpIndex = FrameOffsetFP / sizeof(void*),
	
	ProfileHz = 10,
};

int calls;
int runprofile;   // also the resolution (in cycles) of the profiler
Thread *running;
void dumpstack(Thread*);
void segv(int);
void alrm(int);
void tracealrm(int);
int frametot;

// Allocate a function frame.  Return a pointer into the
// frame offset by the function argsize.
void*
mkfnframe(Fn *fn)
{
	char *fp;

	fp = (char*)emallocnz(fn->frame->size + fn->frame->argsize);
	fp += fn->frame->argsize;
frametot += fn->frame->size + fn->frame->argsize;
	((void**)fp)[FnIndex] = fn;
	return fp;
}

void
freefnframe(void *v)
{
	char *fp;
	Fn *fn;
	
	fp = (char*)v;
	fn = ((Fn**)fp)[FnIndex];
	fp -= fn->frame->argsize;
	free(fp);
}

// Make an upcall into a Zeta function.  `args' should point to
// data containing the arguments followed immediately by the
// return value, and should have room for fn->frame->argsize bytes.
void
callfn(Thread *t, Fn *fn, void *args)
{
	void **fp;
	char *base;
	int upcall;
	jmp_buf stop;
	
	profilecall(nil, fn);

	if(fn->builtin){
		fn->builtin(args);
		return;
	}

	if(--t->depth_ < 0)
		panic("call depth - %s (upcall)", fn->name);

	fp = (void**)mkfnframe(fn);
	base = (char*)fp - fn->frame->argsize;
	memmove(base, args, fn->frame->argsize);
	memmove(stop, t->stop_, sizeof stop);
	upcall = t->upcall_;
	t->upcall_ = 1;
	t->upcalls_ = listinsert(nil, t->upcalls_, t->fp_);
	fp[0] = t->pc_.c;
	fp[1] = t->fp_;
	t->pc_.c = (Code**)fn->code->bp;
	t->fp_ = fp;
	run(t);
	t->upcalls_ = listpop(t->upcalls_);
	t->upcall_ = upcall;
	memmove(t->stop_, stop, sizeof stop);
	running = t;
}

// Call a function `name' defined with global scope.
void
callglobalfn(Thread *t, Name name, void *args, int argsize)
{
	extern char* globaldata(void);
	Var *v;
	Fn *fn;

	if(t == nil)
		t = running;
	if((v = lookup(globalscope, name, 0)) == nil)
		panic("no global variable named '%s'", name);
	if(v->type->op != TypeFn)
		panic("'%s' is not a function", name);
	assert(v->f == globalframe);
	fn = *(Fn**)(globaldata() + v->off);
	if(fn == nil)
		panic("'%s' is an undefined function", name);
	// Sanity check (ignore when <0; xforcelazy doesn't know the argsize statically)
	if(argsize != -1 && argsize != fn->frame->argsize)
		panic("bad argsize for fn call %d want %d", argsize, fn->frame->argsize);
	callfn(t, fn, args);
}

// Run a given thread until it reaches a stop instruction.
void
run(Thread *t)
{
	volatile int depth;
	static int did;

	if(!did){
		struct itimerval it;
		struct sigaction sa;

		did = 1;
		if(FLAGS_catchsegv)
			signal(SIGSEGV, segv);
#if 0
		if(runprofile){
			// Can use ITIMER_VIRUAL and SIGVTALRM here,
			// but I like real time.
			memset(&sa, 0, sizeof sa);
			sa.sa_handler = alrm;
			sa.sa_flags = SA_RESTART;
			if(sigaction(SIGALRM, &sa, nil) < 0)
				fprint(2, "sigaction: %r\n");
			it.it_interval.tv_sec = 0;
			it.it_interval.tv_usec = 1000000 / ProfileHz;
			it.it_value = it.it_interval;
			if(setitimer(ITIMER_REAL, &it, nil) < 0)
				fprint(2, "setitimer: %r\n");
		}
		fnticks = mkhash<Fn*,uintptr>();
		fncalls = mkhash<Fn*,uintptr>();
#endif
	}

	t->rundepth_++;
	depth = t->depth_;

	// Recursive calls are upcalls, and callfn() below decrements
	// t->depth_ to counter the increment that will happen in xret.
	if(t->rundepth_ > 1)
		depth++;

	if(setjmp(t->stop_)){
		t->rundepth_--;
		if(depth != t->depth_){
			fprint(2, "bad depth: was %d, now %d\n", depth, t->depth_);
			dumpstack(t);
		}
		running = nil;
		return;
	}
	running = t;
	if(FLAGS_steptrace){
		// ``Slow'' execution path
		for(;;){
			if(FLAGS_steptrace)
				fprint(2, "x:%p[%d]...", t->pc_.c, t->depth_);
			t->pc_.i++;
			(*t->pc_.c++)(t);
		}
	}else{
		// Fast path: tight loop.
		for(;;){
			t->pc_.i++;
			(*t->pc_.c++)(t);
		}
	}
}

void
dumpstack(Thread *t)
{
	void **fp;
	void *pc;
	Fn *fn;
	List<void*> *l;
	Line line;
	int i;

	fp = (void**)t->fp_;
	pc = t->pc_.c;
	l = t->upcalls_;
	line = ZL;
	while(fp){
		fn = (Fn*)fp[FnIndex];
		i = *(int*)pc;
		line.file = num2file(i>>16);
		line.line = i&0xffff;
		fprint(2, "fp=%p| pc=%p fn=%p %s %L\n",
			fp, pc, fn, fn->name, line);
		pc = fp[0];
		fp = (void**)fp[1];
	}
}

void
abort(void)
{
	if(running)
		dumpstack(running);
	for(;;)
		*(int*)0 = 0;
}

void
segv(int s)
{
	void **fp;
	Fn *fn;

	signal(SIGSEGV, SIG_DFL);
	if(running){
		fprint(2, "segv at (before) %p\n", running->pc_.c);
		dumpstack(running);
		fp = (void**)running->fp_;
		if(fp){
			fn = (Fn*)fp[FnIndex];
			while(fn){
				fprint(2, "-- %s\n", fn->name);
				disasm(fn->code->bp, fn->code->p);
				fprint(2, "--\n");
				if(fn == fn->link)
					break;
				fn = fn->link;
			}
		}
	}
}

List<Fn*> *allfns;

Fn*
mkfn(Line line, Name name, List<Fnarg*> *fnargs, Type *rettype, List<Stmt*> *stmts)
{
	Fn *fn;
	
	fn = (Fn*)emalloc(sizeof *fn);
	fn->line = line;
	fn->name = name;
	fn->fnargs = fnargs;
	fn->rettype = rettype;
	fn->stmts = stmts;
	fn->callouts = mkhash<Fn*,uintptr>();
	fn->link = fn;

	allfns = listappend((Arena*)nil, allfns, fn);

	return fn;
}
