struct Compile;
class Expr;
struct Fmt;

typedef struct Expr Expr;

class Stmt
{
public:
	Line line_;
	Stmt *expanded_;
	
	Stmt() : 
		expanded_(nil) {
		line_.file = nil;
		line_.line = 0;
	}

	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
	virtual void printblock(Fmt*);
	virtual bool reachable();
	virtual ~Stmt();
};

struct StmtExpr : public Stmt
{
	StmtExpr(Expr *expr) :
		Stmt(), expr_(expr),
		flags_(0) {}

	Expr *expr_;
	
	int flags_;  // DeclareConst

	virtual ~StmtExpr() {}
	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
};

struct StmtIf : public Stmt
{
	StmtIf(Expr *cond, Stmt *left, Stmt *right) :
		Stmt(), cond_(cond), left_(left), right_(right) {}
	
	Expr *cond_;
	Stmt *left_;
	Stmt *right_;

	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
};

struct StmtLoop : public Stmt
{
	StmtLoop(Expr *pre, Expr *cond, Expr *post, Stmt *body, int dowhile) :
		Stmt(), pre_(pre), cond_(cond), post_(post), body_(body), dowhile_(dowhile),
		lbreak_(nil), lcontinue_(nil) {}

	Expr *pre_;
	Expr *cond_;
	Expr *post_;
	Stmt *body_;
	int dowhile_;
	
	Label *lbreak_;
	Label *lcontinue_;

	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
};

struct StmtFor : public StmtLoop
{
	StmtFor(Expr *pre, Expr *cond, Expr *post, Stmt *body) :
		StmtLoop(pre, cond, post, body, 0) {}
	
	virtual void pprint(Fmt*);
};

struct StmtWhile : public StmtLoop
{
	StmtWhile(Expr *cond, Stmt *body) :
		StmtLoop(nil, cond, nil, body, 0) {}

	virtual void pprint(Fmt*);
};

struct StmtDoWhile : public StmtLoop
{
	StmtDoWhile(Stmt *body, Expr *cond) :
		StmtLoop(nil, cond, nil, body, 1) {}

	virtual void pprint(Fmt*);
};

struct StmtReturn : public Stmt
{
	StmtReturn(Expr *expr) :
		Stmt(), expr_(expr) {}
	
	Expr *expr_;

	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
};

struct StmtBreak : public Stmt
{
	StmtBreak() :
		Stmt(), 
		label_(nil) {}

	Label *label_;

	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
};

struct StmtContinue : public Stmt
{
	StmtContinue() :
		Stmt(),
		label_(nil) {}

	Label *label_;

	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
};

struct StmtList : public Stmt
{
	StmtList(List<Stmt*> *stmts) :
		Stmt(), stmts_(stmts) {}
	
	List<Stmt*> *stmts_;

	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
	virtual void printblock(Fmt*);
};

struct StmtFndef : public Stmt
{
	StmtFndef(Name name, Fn *fn, List<Name> *typevars) :
		Stmt(), name_(name), fn_(fn), typevars_(typevars),
		flags_(0), var_(nil), attr_(nil) {}
	
	Name name_;
	Fn *fn_;
	List<Name> *typevars_;
	
	int flags_;
	Var *var_;
	Var *attr_;

	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
};

struct StmtPrint : public Stmt
{
	StmtPrint(Expr *expr) :
		Stmt(), expr_(expr) {}
	
	Expr *expr_;

	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
};

struct StmtLabel : public Stmt
{
	StmtLabel(Name name) :
		Stmt(), name_(name),
		label_(nil) {}
	
	Name name_;
	Label *label_;

	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
	virtual bool reachable();
};

struct StmtGoto : public Stmt
{
	StmtGoto(Name name) :
		Stmt(), name_(name),
		label_(nil) {}
	
	Name name_;
	
	Label *label_;

	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
};

struct StmtStruct : public Stmt
{
	StmtStruct(Name name, List<Structelem*> *elems, List<Name> *typevars) :
		Stmt(), name_(name), elems_(elems), typevars_(typevars),
		flags_(0), discriminated_(0) {}
	
	Name name_;
	List<Structelem*> *elems_;
	List<Name> *typevars_;
	
	int flags_;
	int discriminated_;
	
	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
};

struct StmtCase : public Stmt
{
	StmtCase(Expr *cond, List<Clause*> *clauses) :
		Stmt(), cond_(cond), clauses_(clauses) {}
	
	Expr *cond_;
	List<Clause*> *clauses_;
	
	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
};

struct StmtDeclare : public Stmt
{
	StmtDeclare(Name name, Type *type, List<Name> *typevars) :
		Stmt(), name_(name), type_(type), typevars_(typevars),
		var_(nil) {}
	
	Name name_;
	Type *type_;
	List<Name> *typevars_;

	Var *var_;

	virtual void typecheck(Typecheck*);
	virtual void compile(Compile*);
	virtual void pprint(Fmt*);
};

struct StmtDeclareAttribute : public Stmt
{
	StmtDeclareAttribute(List<Name> *names, Name name, Type *type) :
		Stmt(), names_(names), name_(name), type_(type),
		var_(nil), attr_(nil) {}
	
	List<Name> *names_;
	Name name_;
	Type *type_;
	
	Var *var_;
	Var *attr_;
	
	virtual void typecheck(Typecheck*);
	virtual void compile(Compile*);
	virtual void pprint(Fmt*);
};

struct StmtDeclareType : public Stmt
{
	StmtDeclareType(Name name, Type *type) :
		Stmt(), name_(name), type_(type) {}
	
	Name name_;
	Type *type_;

	virtual void typecheck(Typecheck*);
	virtual void compile(Compile*);
	virtual void pprint(Fmt*);
};

struct StmtForList : public Stmt
{
	StmtForList(Expr *expr, Expr *cond, Stmt *body) :
		Stmt(), expr_(expr), cond_(cond), body_(body) {}
	
	Expr *expr_;
	Expr *cond_;
	Stmt *body_;

	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
};


