// Type checking. 
// Records Type* of each expression for later use by compile.
// Records other operation-specific information too.

#include "a.h"

// Check that the two halves of an expression have the same type.
static void typecheckjoin(ExprBinary*);

// Check that a type can be converted to bool.
static int canconverttobool(Type*);

// Mark all expr variables as const.
static void markvarsconst(Expr*);

// Check that an attribute fn definition is ok.
// If so, return the Var corresponding to that attribute.

static Var *declareattribute(Typecheck*, Line line, Name name, Type *tfn);
static Type *attrfntype(Typecheck*, StmtFndef*);
static Type *attrfntype(Typecheck*, StmtDeclareAttribute*);
static Var* lookupattribute(Typecheck *tc, Line line, Name name, Type *tfn);

// Add a struct field or attribute
static Var* mknewstructvar(Typecheck*, Line, Name, Type*, List<Type*>*);

// debugging
// static void _printframe(Frame*, int);

void
StmtExpr::typecheck(Typecheck *tc)
{
	if(flags_ & DeclareConst){
		ExprDeclEq *e = expr_->exprdecleq();
		if(!e){
			error(line_, "poorly formed const declaration %#E", expr_);
			return;
		}
		e->flags_ |= DeclareConst;
	}
	typecheckexpr(tc, expr_, 0);
}

void
StmtLoop::typecheck(Typecheck *tc)
{
	pushscope(tc);
	lbreak_ = mklabel(tc);
	lcontinue_ = mklabel(tc);
	pushbreak(tc, lbreak_);
	pushcontinue(tc, lcontinue_);
	pushflow(tc);
	typecheckexpr(tc, pre_, Value);
	popflow(tc);
	pushflow(tc);
	if(cond_){
		cond_ = new ExprConvert(cond_, typebool());
		tc->incondition_ = 1;
		typecheckexpr(tc, cond_, Value);
		tc->incondition_ = 0;
	}
	typecheckstmt(tc, body_);
	popflow(tc);
	if(!tc->reachable_){
		if(!lcontinue_->used && post_)
			error(post_->line_, "post-condition not reachable");
		tc->reachable_ = 1;
	}
	pushflow(tc);
	typecheckexpr(tc, post_, Value);
	popflow(tc);
	popcontinue(tc, lcontinue_);
	popbreak(tc, lbreak_);
	popscope(tc);
	tc->reachable_ = (cond_ != nil || lbreak_->used);
}

void
StmtIf::typecheck(Typecheck *tc)
{
	cond_ = new ExprConvert(cond_, typebool());
	pushflow(tc);
	tc->incondition_ = 1;
	typecheckexpr(tc, cond_, Value);
	tc->incondition_ = 0;
	pushscope(tc);
	typecheckstmt(tc, left_);
	popflow(tc);
	int reachable = tc->reachable_;
	tc->reachable_ = 1;
	popscope(tc);
	pushscope(tc);
	pushflow(tc);
	typecheckstmt(tc, right_);
	popflow(tc);
	popscope(tc);
	tc->reachable_ |= reachable;
}

void
StmtBreak::typecheck(Typecheck *tc)
{
	if(tc->breakstack_.empty()){
		error(line_, "break outside loop");
		return;
	}
	label_ = tc->breakstack_.back();
	label_->used = this;
	tc->reachable_ = 0;
}

void
StmtContinue::typecheck(Typecheck *tc)
{
	if(tc->continuestack_.empty()){
		error(line_, "continue outside loop");
		return;
	}
	label_ = tc->continuestack_.back();
	label_->used = this;
	tc->reachable_ = 0;
}

void
StmtList::typecheck(Typecheck *tc)
{
	pushscope(tc);
	for(List<Stmt*> *l=stmts_; l; l=l->next)
		typecheckstmt(tc, l->x);
	popscope(tc);
}

void
StmtPrint::typecheck(Typecheck *tc)
{
	typecheckexpr(tc, expr_, Value);
}

void
StmtReturn::typecheck(Typecheck *tc)
{
	if(tc->fn_ == nil){
		error(line_, "return outside function");
		return;
	}
	if(expr_){
		expr_->infertype_ = tc->fn_->rettype;
		typecheckexpr(tc, expr_, Value);
		if(tc->fn_->rettype == expr_->type_ || issubtype(tc->fn_->rettype, expr_->type_)){
			// okay
		}else if(shouldcompletetype(tc->fn_->rettype, expr_->type_)){
			expr_->type_ = tc->fn_->rettype;
		}else{
			error(line_, "return [%T] (%#E) from [%T] function",
				expr_->type_, expr_, 
				tc->fn_->rettype);
			return;
		}
	}else{
		if(tc->fn_->rettype != typevoid()){
			error(line_, "return value expected from %T function",
				tc->fn_->rettype);
			return;
		}
	}
	tc->reachable_ = 0;
}

void
StmtGoto::typecheck(Typecheck *tc)
{
	label_ = uselabel(tc, name_, this);
	tc->reachable_ = 0;
}

void
StmtLabel::typecheck(Typecheck *tc)
{
	label_ = setlabel(tc, name_, this);
}

void
StmtDeclare::typecheck(Typecheck *tc)
{
	List<Type*> *typevars;

	typevars = names2typevars(typevars_);
	tc->decltypevars_ = typevars;
	type_ = resolvenamedtype(tc, line_, type_);
	tc->decltypevars_ = nil;
	if(typevars){
		if(type_->op != TypeFn){
			error(line_, "polymorphic global variables must have function types");
			return;
		}
		type_ = typeforall(typevars, type_);
	}
	if(name_ && strcmp(name_, "_") == 0){
		warn(line_, "declaring a don't care, ignoring");
		return;
	}
	var_ = declare(tc, line_, name_, type_);
	if(!var_)
		type_ = nil;
}

void
StmtDeclareAttribute::typecheck(Typecheck *tc)
{
	tc->decltypevars_ = nil;
	type_ = resolvenamedtype(tc, line_, type_);
	if(name_ && strcmp(name_, "_") == 0){
		warn(line_, "declaring a don't care, ignoring");
		return;
	}
	Type *tfn = attrfntype(tc, this);
	if((attr_ = declareattribute(tc, line_, name_, tfn)) == nil)
		return;
	var_ = declare(tc, line_, attr_->mangled, tfn);
	attr_->attrfn = var_;
	type_ = tfn;
}

void
StmtDeclareType::typecheck(Typecheck *tc)
{
	type_ = resolvenamedtype(tc, line_, type_);
	Type *t = type_;
	declaretype(tc, line_, name_, &t);
}

// TODO: Factor better.
void
StmtFndef::typecheck(Typecheck *tc)
{
	int old;
	Type *t;

	if((flags_ & FnAttribute) && typevars_){
		error(line_, "attribute functions cannot be polymorphic");
		return;
	}
	switch(flags_){
	// New function
	case FnAttribute:
	case FnAttribute|FnExtensible:
		if((attr_ = declareattribute(tc, line_, name_, attrfntype(tc, this))) == nil)
			break;
		name_ = attr_->mangled;
	case FnExtensible:
		//	extensible f(a: int): int { } 
		//		- just ignore "extensible"
	default:
		var_ = lookup(tc->scope_, name_, 0);
		if(var_){
			old = 1;
			if(var_->isdefined)
				warn(line_, "overriding definition of %s\n", name_);
		}
		else{
			var_ = declare(tc, line_, name_, typefn(nil, nil));
			if(attr_)
				attr_->attrfn = var_;
			old = 0;
		}
		if(var_)
			var_->isdefined = 1;
		typecheckfn(tc, line_, fn_, var_, typevars_, old);
		break;
	// Extending an old function
	case FnAttribute|FnExtend:
		if((attr_ = lookupattribute(tc, line_, name_, attrfntype(tc, this))) == nil)
			break;
		name_ = attr_->mangled;
		// Fall through.
	case FnExtend:
		//	extend f(a: int): int { foo; }
		//		becomes
		//		f = fn(default: int->int): int->int {
		//			return fn(a: int): int { foo; }
		//		}(f);
		var_ = lookup(tc->scope_, name_, 0);
		if(var_ == nil){
			error(line_, "no function %s", name_);
			break;
		}
		t = var_->type;
		if(t && t->op == TypeForall)
			t = t->left;
		if(t == nil || t->op != TypeFn){
			error(line_, "%s is not a function", name_);
			break;
		}
		vector<Expr*> exprs;
		exprs.push_back(new ExprName(name_));
		expanded_ = new StmtList(list1(asyntax, (Stmt*)new StmtExpr(new ExprEq(
			new ExprName(name_),
			new ExprCall(
				new ExprFn(mkfn(line_, Name::null,
					list1(asyntax, mkFnarg(nameof("default"), var_->type)),
					var_->type,
					list1(asyntax, (Stmt*)new StmtReturn(new ExprFn(fn_, typevars_)))),
					typevars_),
				exprs)))));
		typecheckstmt(tc, expanded_);
		break;
	}
}

// TODO: Factor better.
void
StmtStruct::typecheck(Typecheck *tc)
{
	Type *t, *tforall;
	Name n;
	List<Type*> *typevars, *lt;

	switch(flags_){
	case FnExtensible:
	default:
		typevars = names2typevars(typevars_);
		tc->decltypevars_ = typevars;
		t = typecheckstruct(tc, line_, name_, typevars, elems_, nil, nil,
		                    flags_, discriminated_);
		tc->decltypevars_ = nil;
		if(t && t->discriminated && !t->extensible && !t->tags)
			warn(line_, "discriminated struct %s has no tags", t->name);
		if(t)
			declaretype(tc, line_, name_, &t);
		break;
	case FnExtend:
		if((t = lookuptype(tc->scope_, name_, 0)) == nil){
			error(line_, "no type %s", name_);
			break;
		}
		tforall = nil;
		if(t->op == TypeForall){
			tforall = t;
			t = t->left;
		}
		if(t->op == TypeSyntax || t->op == TypeUnion)
			t = t->structrep;
		assert(t);
		if(t->op != TypeStruct){
			error(line_, "type %s is not a struct: %T", name_, t);
			break;
		}
		if(!t->extensible){
			error(line_, "type %s is not extensible", name_);
			break;
		}
		n = nameprint("%s.X%d", name_, listlen(t->extensions));
		// All typevar names must match the original names
		if(tforall){
			List<Name> *l;
			for(l=typevars_, lt=tforall->typevars; l && lt; l=l->next, lt=lt->next){
				if(l->x != lt->x->name){
					error(line_, "typevar mismatch for type %s", name_);
					break;
				}
			}
			if(l || lt){
				error(line_, "typevar mismatch for type %s", name_);
				break;
			}
			tc->decltypevars_ = tforall->typevars;
		}
		else if(typevars_){
			error(line_, "type %s is not polymorphic", name_);
			break;
		}
		typecheckstruct(tc, line_, n, ZList, elems_, nil, t,
		                t->extensible, t->discriminated);
		tc->decltypevars_ = nil;
		break;
	}
}

// Type-check the given statement.
// Performs reachability analysis during the code traversal.
void
typecheckstmt(Typecheck *tc, Stmt *stmt)
{
	if(stmt == nil)
		return;

	// Warn if can't get here from anywhere.
	if(tc->reachable_ == 0 && !stmt->reachable())
		warn(stmt->line_, "unreachable code %$", stmt);
	tc->reachable_ = 1;

	stmt->typecheck(tc);
}

// To do: replace with table?
void
typecheckexpr(Typecheck *tc, Expr *expr, int kind)
{
	Expr *e1;

	if(expr == nil){
		if(expr)
			fprint(2, "expr %#E has no type\n", expr);
		return;
	}

	// Sanity check: this only makes sense for parse exprs
	if(kind&BindValue && !expr->exprparse())
		panic("bind value on bad expr %E", expr);
	
	for(e1=expr; e1->expanded_; e1=e1->expanded_)
		;
	e1->typecheck(tc, kind);

	if(e1->type_ && checkunresolved(expr->line_, e1->type_))
		e1->type_ = typevoid();
	expr->type_ = e1->type_;
}

void
Expr::typecheck(Typecheck *tc, int kind)
{
	panic("expr::typecheck");
}

void
ExprName::typecheck(Typecheck *tc, int kind)
{
	Var *v;
	if((v = lookup(tc->scope_, name_, 0)) == nil){
		if(strcmp(name_, "_") == 0 && kind != Value)
			type_ = typedontcare();
		else{
			error(line_, "unknown variable %s", name_);
			//_printframe(tc->frame_, 1);
		}
		return;
	}
	var_ = v;
	type_ = v->type;
	if(v->isconst)
		noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
}

void
ExprArray::typecheck(Typecheck *tc, int kind)
{
	nosideeffect(tc, this, kind);
	typecheckexpr(tc, left_, Value);
	typecheckexpr(tc, right_, Value);
	Type *tl = left_->type_;
	Type *tr = right_->type_;
	if(tl == nil || tr == nil)
		return;
	if(tl->op == TypeString && tr == typeint()){
		noassign(tc, this, kind);
		type_ = typestring();
		return;
	}
	if(tl->op == TypeArray && tr == typeint()){
		type_ = tl->left;
		return;
	}
	if(tl->op == TypeMap && issubtype(tl->left, tr)){
		type_ = tl->right;
		return;
	}
	// Allow a list of pairs to be treated as an ordered map
	// when the first element is also a valid map key.
	if(tl->op == TypeList){
		Type *t, *tret;
		if(isassoclist(tl, &t, &tret)){
			if(issubtype(t, tr))
				type_ = tret;
			else
				error(line_, "bad associative list index in %E", this);
		}
		// Allow list[n] to index a list
		else if(tr == typeint())
			type_ = tl->left;
		else
			error(line_, "cannot index non-associative list at %E", this);
		return;
	}
	// scope[name] --> scopelook(): ptr Sym
	// scope[int] - -> scopenth(): ptr Sym
	if(tl == typeptr(typescope()) &&
	   (tr == typestring() || isptrast(tr))){
		noassign(tc, this, kind);
		vector<Expr*> args;
		args.push_back(left_);
		args.push_back(new ExprConvert(right_, typestring()));
		expanded_ = new ExprCall(new ExprName(nameof("scopelook")), args);
		typecheckexpr(tc, this, Value);
		return;
	}
	if(tl == typeptr(typescope()) && tr == typeint()){
		noassign(tc, this, kind);
		vector<Expr*> args;
		args.push_back(left_);
		args.push_back(right_);
		expanded_ = new ExprCall(new ExprName(nameof("scopenth")), args);
		typecheckexpr(tc, this, Value);
		return;
	}
	// frame[name] --> framelook(): ptr Sym
	if(tl == typeptr(typeframe()) &&
	   (tr == typestring() || isptrast(tr))){
		noassign(tc, this, kind);
		vector<Expr*> args;
		args.push_back(left_);
		args.push_back(new ExprConvert(right_, typestring()));
		expanded_ = new ExprCall(new ExprName(nameof("framelook")), args);
		typecheckexpr(tc, this, Value);
		return;
	}
	error(line_, "bad array index in %#E - %T[%T]", this, tl, tr);
}

void
ExprSlice::typecheck(Typecheck *tc, int kind)
{
	nosideeffect(tc, this, kind);
	noassign(tc, this, kind);	// maybe later
	typecheckexpr(tc, left_, Value);
	typecheckexpr(tc, middle_, Value);
	typecheckexpr(tc, right_, Value);
	if(left_->type_ == nil)
		return;
	Type *tl = left_->type_;
	if(middle_){
		if(middle_->type_ != typeint()){
			error(line_, "bad type %T for left slice %#E", middle_->type_, middle_);
			return;
		}
	}
	if(right_){
		if(right_->type_ != typeint()){
			error(line_, "bad type %T for right slice %#E", right_->type_, right_);
			return;
		}
	}
	if(tl->op == TypeArray || tl->op == TypeString)
		type_ = tl;
	else
		error(line_, "cannot slice %#E [%T]", left_, left_->type_);
}

void
ExprStar::typecheck(Typecheck *tc, int kind)
{
	nosideeffect(tc, this, kind);
	typecheckexpr(tc, left_, Value);
	Type *t = left_->type_;
	if(t == nil || t->op != TypePtr){
		error(line_, "non-pointer %#E", left_);
		return;
	}
	if(isabstracttype(t->left)){
		error(line_, "cannot dereference abstract type %T", t->left);
		return;
	}
	type_ = t->left;
}

void
ExprAddr::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	typecheckexpr(tc, left_, LValue);
	Type *t = left_->type_;
	type_ = typeptr(t);
}

void
ExprCode::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	typecheckexpr(tc, left_, assign_ ? LValue : Value);
	if(right_)
		typecheckexpr(tc, right_, Value);
	Type *tl = left_->type_;
	if(tl == nil)
		return;
	Type *tr = nil;
	if(right_){
		tr = right_->type_;
		if(tr == nil)
			return;
	}
	const vector<ExprSig>& sigv = sigs();
	for(int i=0; i<sigv.size(); i++){
		const ExprSig& sig = sigv[i];
		if(sig.left == tl && sig.right == tr){
			type_ = sig.out;
			code_ = sig.code;
			return;
		}
	}
	
	if(right_)
		error(line_, "bad types %T and %T in %#E", tl, tr, this);
	else
		error(line_, "bad type %T in %#E", tl, this);
	if(!assign_)
		nosideeffect(tc, this, kind);
}

void
ExprAddEq::typecheck(Typecheck *tc, int kind)
{
	typecheckexpr(tc, left_, LValue);
	if(left_->type_->op == TypeList){
		typecheckexpr(tc, right_, Value);
		noassign(tc, this, kind);
		Type *t = subtypejoin(left_->type_, right_->type_);
		if(t){
			type_ = t;
			return;
		}
	}
	ExprCode::typecheck(tc, kind);
}

void
ExprIncDec::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	typecheckexpr(tc, left_, Value);
	Type *t = left_->type_;
	if(!post_ && inc_ == 1 && t->op == TypeMapIter){
		type_ = typebool();
		return;
	}
	if(t != typeint() && t != typefloat())
		error(line_, "bad type %T in %#E", t, this);
	type_ = t;
}

void
ExprAndAnd::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	typecheckexpr(tc, left_, Value);
	typecheckexpr(tc, right_, kind&Value);
	typecheckjoin(this);
	// must be convertible to a boolean
	if(canconverttobool(left_->type_))
		type_ = left_->type_;
}

void
ExprOrOr::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	// each OR branch is a seperate flow branch
	pushflow(tc);
	typecheckexpr(tc, left_, Value);
	popflow(tc);
	pushflow(tc);
	typecheckexpr(tc, right_, kind&Value);
	popflow(tc);
	typecheckjoin(this);
	// must be convertible to a boolean
	if(canconverttobool(left_->type_))
		type_ = left_->type_;
}

void
ExprNot::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	pushflow(tc);
	left_ = new ExprConvert(left_, typebool());
	typecheckexpr(tc, left_, Value);
	type_ = typebool();
	ExprCode::typecheck(tc, kind);
	popflow(tc);
}

void
ExprDeclEq::typecheck(Typecheck *tc, int kind)
{
	if(type_ != nil){
		type_ = resolvenamedtype(tc, line_, type_);
		right_->infertype_ = type_;
	}
	typecheckexpr(tc, right_, Value);
	if(type_ == nil)
		type_ = right_->type_;
	if(!declareexpr(tc, left_, type_)){
		type_ = nil;
		return;
	}
	typecheckexpr(tc, left_, LValue);
	if(flags_ & DeclareConst)
		markvarsconst(left_);

	noassign(tc, this, kind);
	Type *tl = left_->type_;
	Type *tr = right_->type_;
	if(tl == nil || tr == nil)
		return;
	if(issubtype(tl, tr)){
		type_ = left_->type_;
		return;
	}else if(shouldcompletetype(tl, tr)){
		right_->type_ = left_->type_;
		type_ = left_->type_;
		return;
	}else if(tl->op == TypeTuple && tr->op == TypeTuple){
		List<Type*> *ll, *lr;
		for(ll=tl->types, lr=tr->types; ll && lr; ll=ll->next, lr=lr->next)
			if(ll->x == nil || lr->x == nil || (ll->x != lr->x && ll->x != typedontcare()))
				goto Bad;
		if(ll || lr)
			goto Bad;
		type_ = right_->type_;
		return;
	}
Bad:
	error(line_, "bad types %T and %T in %#E", tl, tr, this);
}

void
ExprTuple::typecheck(Typecheck *tc, int kind)
{
	nosideeffect(tc, this, kind);
	for(int i=0; i<exprs_.size(); i++)
		typecheckexpr(tc, exprs_[i], (kind&(Value|LValue)));
	type_ = typetupleexprs(exprs_);
}

void
ExprBool::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	type_ = typebool();
}

void
ExprInteger::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	type_ = typeint();
}

void
ExprFloat::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	type_ = typefloat();
}

void
ExprString::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	type_ = typestring();
}

void
ExprFn::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	typecheckfn(tc, line_, fn_, nil, typevars_, 1);
	type_ = fn_->type;
}

void
ExprNew::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	// Special case: discriminated struct copying - new *(ptr Struct)
	ExprStar *es;
	if(left_ && (es = left_->exprstar()) != nil && es->left_){
		typecheckexpr(tc, es->left_, Value);
		Type *t = es->left_->type_;
		if(t && t->op == TypePtr && t->left->op == TypeStruct){
			type_ = t;
			left_->type_ = nil;
			return;
		}
	}
	if(infertype_ && infertype_->op == TypePtr)
		left_->infertype_ = infertype_->left;
	typecheckexpr(tc, left_, Value);
	if(isabstracttype(left_->type_))
		error(line_, "cannot construct abstract type %T", left_->type_);
	else
		type_ = typeptr(left_->type_);
}

void
ExprDot::typecheck(Typecheck *tc, int kind)
{
	nosideeffect(tc, this, kind);   // XXX technically not true for lazy attributes
	int wasarray = (left_ && left_->exprarray());
	typecheckexpr(tc, left_, Value);  // sometimes LValue...
	if(left_->type_ == nil)
		return;
	Type *t = left_->type_;
	// Special case for List.hd and List.tl
	if(t->op == TypeList){
		static Var *varhd, *vartl;
		if(varhd == nil){
			vartl = mkVar(nameof("tl"), typelist(nil));
			vartl->off = offsetof(ListNode, next);
			varhd = mkVar(nameof("hd"), nil);
			varhd->off = offsetof(ListNode, v);
		}
		if(name_ == nameof("hd")){
			var_ = varhd;
			type_ = t->left;
			return;
		}else if(name_ == nameof("tl")){
			var_ = vartl;
			type_ = t;
			return;
		}
	}
	if(t->op == TypePtr)
		t = t->left;
	if(checkunresolved(line_, t))
		return;
	// Like unfoldstruct() with better error checking
	Type *bt = nil;
	Type *subst = nil;
	if(t && t->op == TypeSubst){
		subst = t;
		t = t->left;
	}
	if(t && t->op == TypeForall)
		t = t->left;
	if(t && (t->op == TypeSyntax || t->op == TypeUnion)){
		bt = t;
		t = t->structrep;
	}
	if(t == nil || t->op != TypeStruct){
		error(line_, "non-struct %#E [%T]", left_, left_->type_);
		return;
	}
	Var *v;
	Type *xt;
	if((v = lookupstruct(t, name_, tc->scope_, &xt)) == nil){
		error(line_, "%T has no member %s", t, name_);
		return;
	}
	// TODO: propogate const from left_?
	if(v->isconst)
		noassign(tc, this, kind);
	// scope[name].type --> scopelooktype(scope, name)
	// scope[int].type  --> scopenthtype(scope, name)
	if(wasarray && t == typesym()){
		// scope[] was rewritten to scopelook() or scopenth()
		ExprCall *ec = left_->exprcall();
		if(ec){
			noassign(tc, this, kind);
			ExprName *en = ec->left_->exprname();
			if(en == nil)
				panic("%L: bad scope blah blah blah", line_);
			if(strcmp(en->name_, "scopelook") == 0){
				expanded_ = new ExprCall(new ExprName(nameof("scopelooktype")), ec->exprs_);
				typecheckexpr(tc, this, Value);
				return;
			}
			if(strcmp(en->name_, "scopenth") == 0){
				expanded_ = new ExprCall(new ExprName(nameof("scopenthtype")), ec->exprs_);
				typecheckexpr(tc, this, Value);
				return;
			}
			panic("%L: unknown scope look fn %s", line_, en->name_);
			return;
		}
	}
	// frame[name].type --> framelooktype(frame, name)
	if(wasarray && t == typesym()){
		// frame[] was rewritten to framelook()
		ExprCall *ec = left_->exprcall();
		if(ec){
			noassign(tc, this, kind);
			ExprName *en = ec->left_->exprname();
			if(en == nil)
				panic("%L: bad frame blah blah blah", line_);
			if(strcmp(en->name_, "framelook") == 0){
				expanded_ = new ExprCall(new ExprName(nameof("framelooktype")), ec->exprs_);
				typecheckexpr(tc, this, Value);
				return;
			}
			panic("unknown frame look fn %s", en->name_);
			return;
		}
	}
	var_ = v;
	type_ = v->type;
	xtype_ = xt;
	if(subst)
		type_ = applytypesubst(type_, subst->typevars, subst->types);
	if(bt && strcmp(name_, "copiedfrom") == 0){
		type_ = typeptr(bt);
		noassign(tc, this, kind);
		return;
	}
}

void
ExprConvert::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	typecheckexpr(tc, left_, Value);
	int errored = 0;
	Type *t = left_->type_;
	if(t == nil)
		return;
	if(type_ == typebool()){
		errored = !canconverttobool(t);
	}
	else if(type_ == typeint()){
		switch(t->op){
		case TypeBool:
		case TypeInt:
		case TypeFloat:
		case TypeString:
			break;
		default:
			errored = 1;
			break;
		}	
	}
	else if(type_ == typefloat()){
		switch(t->op){
		case TypeBool:
		case TypeInt:
		case TypeFloat:
		case TypeString:
			break;
		default:
			errored = 1;
			break;
		}	
	}
	else if(type_ == typestring()){
		switch(t->op){
		case TypeBool:
		case TypeInt:
		case TypeFloat:
		case TypeString:
			break;
		default:
			Type *xt = typeptr(typeast());
			if(issubtype(xt, left_->type_)){
				// fine
			}else if(shouldcompletetype(xt, left_->type_))
				left_->type_ = xt;
			else
				errored = 1;
			break;
		}	
	}
	else if(type_ == typelist(nil)){
		Type *tret;
		if(isastlisttype(t, &tret))
			type_ = tret;
		else if(t->op == TypeArray)
			type_ = typelist(t->left);
		else
			errored = 1;
	}
	else if(type_ == typearray(nil)){
		//TODO: an ast(*+[]) conversion would be OK here as well
		if(t->op == TypeInt)
			type_ = typearray(nil);
		else if(t->op == TypeList)
			type_ = typearray(t->left);
		else
			errored = 1;
	}
	else if(type_ == typeptr(nil)){
		Type *tret;
		if(isastoptiontype(t, &tret))
			type_ = tret;
		else
			errored = 1;
	}
	else if(type_->op == TypeTuple){
		if(t->op != TypeMapIter || type_ != typetuple(list2(asyntax, t->left, t->right)))
			errored = 1;
	}
	else if(type_->op == TypeMapIter){
		if(t->op != TypeMap)
			errored = 1;
	}
	if(errored){
		error(line_, "illegal conversion %#E of type %T to type %T %d %d", this, t, type_, canconverttobool(t), t->op);
		type_ = nil;
	}
}

void
ExprNil::typecheck(Typecheck *tc, int kind)
{
	type_ = typeptr(nil);
}

static Type*
typechecklist(Typecheck *tc, const vector<Expr*>& exprs)
{
	Type *t = nil;
	for(int i=0; i<exprs.size(); i++){
		Expr *e = exprs[i];
		typecheckexpr(tc, e, Value);
		if(t == nil){
			t = e->type_;
		}
		else if(t != e->type_){
			Type *xt = subtypejoin(t, e->type_);
		 	if(xt == nil){
				error(e->line_, "cannot mix types %T and %T", t, e->type_);
				t = nil;
				break;
			}
			t = xt;
		}
	}
	return t;
}

void
ExprMkListOrArray::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	Type *t = typechecklist(tc, exprs_);
	if(infertype_ && infertype_->op == TypeList){
		type_ = typelist(t);
		expanded_ = new ExprMkList(exprs_);
		return;
	}
	if(infertype_ && infertype_->op == TypeArray){
		type_ = typearray(t);
		expanded_ = new ExprMkArray(exprs_);
		return;
	}
	error(line_, "could not disambiguate array/list constructor");
}

void
ExprMkList::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	type_ = typelist(typechecklist(tc, exprs_));
}

void
ExprMkArray::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	type_ = typearray(typechecklist(tc, exprs_));
}

void
ExprMkMap::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	Type *tl = nil;
	Type *tr = nil;
	for(int i=0; i<exprs_.size(); i++){
		ExprMap *e = exprs_[i];
		typecheckexpr(tc, e->left_, Value);
		typecheckexpr(tc, e->right_, Value);
		if(tl == nil){
			tl = e->left_->type_;
			tr = e->right_->type_;
			if(tl == nil || tr == nil)
				break;
		}
		else{
			Type *xt = subtypejoin(tl, e->left_->type_);
			Type *bt = subtypejoin(tr, e->right_->type_);
			if(xt == nil || bt == nil){
				error(line_, "cannot mix types %T => %T and %T => %T",
				                  tl, tr, e->left_->type_, e->right_->type_);
				tl = nil;
				tr = nil;
				break;
			}
			tl = xt;
			tr = bt;
		}
	}
	type_ = typemap(tl, tr);
}

void
ExprListCons::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
//ExprListCons_again:
	typecheckexpr(tc, right_, Value);
	Type *tr = right_->type_;
/*
	if(tr && tr->op != TypeList && right_->op != ExprConvert){
		right_ = new ExprConvert(right_, typelist(nil));
		right_->line = right_->left->line;
		goto ExprListCons_again;
	}
*/
	if(tr && tr->op == TypeList)
		left_->infertype_ = tr->left;
	typecheckexpr(tc, left_, Value);
	Type *tl = left_->type_;
	if(tr->op != TypeList)
		error(line_, "bad type for %#E, expected a list", tl, tr, this);
	Type *xt = subtypejoin(tl, tr->left);
	if(xt == nil)
		error(line_, "mixed types in list cons: %T and %T", tl, tr);
	else
		type_ = typelist(xt);
}

void
ExprLength::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
//ExprLength_again:
	typecheckexpr(tc, left_, Value);
	Type *tl = left_->type_;
	if(tl == nil)
		return;
	if(tl->op == TypeArray || tl->op == TypeList || tl->op == TypeString){
		type_ = typeint();
		return;
	}
/*
	// Try again with a list() conversion.
	if(left_->op != ExprConvert){
		e = left_;
		left_ = new ExprConvert(e, typelist(nil));
		left_->line = e->line;
		goto ExprLength_again;
	}
*/
	error(line_, "bad type for %#E, cannot compute length()", left_);
}

void
ExprReverse::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
//ExprReverse_again:
	typecheckexpr(tc, left_, Value);
	Type *tl = left_->type_;
	if(tl == nil)
		return;
	if(tl->op == TypeList){
		type_ = left_->type_;
		return;
	}
/*
	// Try again with a list() conversion.
	if(left_->op != ExprConvert){
		e = left_;
		left_ = new ExprConvert(e, typelist(nil));
		left_->line = e->line;
		goto ExprReverse_again;
	}
*/
	error(line_, "cannot reverse %#E [%T]", left_, left_->type_);
}

void
ExprTag::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	typecheckexpr(tc, left_, Value);
	Type *tl = left_->type_;
	if(tl == nil)
		return;
	if(tl->op == TypePtr && tl->left->op == TypeStruct)
		if(rootstruct(tl->left)->discriminated){
			type_ = typestring();
			return;
		}
	error(line_, "expected discriminated struct ptr for tag()");
}

void
noassign(Typecheck *tc, Expr *expr, int kind)
{
	if(kind&LValue)
		error(expr->line_, "cannot assign to %#E", expr);
	if(kind&BindValue)
		error(expr->line_, "cannot bind to %#E", expr);
}

void
nosideeffect(Typecheck *tc, Expr *expr, int kind)
{
	if(!(kind&(Value|LValue)))
		error(expr->line_, "%#E evaluated but not used", expr);
}

void
ExprAdd::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	typecheckexpr(tc, left_, assign_ ? LValue : Value);
	typecheckexpr(tc, right_, Value);
	Type *tl = left_->type_;
	Type *tr = right_->type_;
	if(tl == nil || tr == nil)
		return;

	if(tl->op == TypeList){
		Type *t = subtypejoin(tl, tr);
		if(t != nil){
			type_ = t;
			return;
		}
	}
	
	ExprCode::typecheck(tc, kind);
}

void
ExprSameStructure::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	typecheckexpr(tc, left_, Value);
	typecheckexpr(tc, right_, Value);
	Type *tl = left_->type_;
	Type *tr = right_->type_;
	if(tl == nil || tr == nil)
		return;
	if(!isptrast(tl) || !isptrast(tr))
		goto Bad;
	if(!issubtype(tl, tr) && !issubtype(tr, tl))
		goto Bad;
	type_ = typebool();
	return;
Bad:
	error(line_, "bad types %T and %T in %#E", tl, tr, this);
}

void
typecheckcmp(Typecheck *tc, ExprBinary *expr, int kind)
{
	noassign(tc, expr, kind);
	typecheckexpr(tc, expr->left_, Value);
	typecheckexpr(tc, expr->right_, Value);
	Type *tl = expr->left_->type_;
	Type *tr = expr->right_->type_;
	if(tl == nil || tr == nil)
		return;
	if(tl->op == TypeStruct || tr->op == TypeStruct)
		goto Bad;
	if(issubtype(tl, tr) || issubtype(tr, tl)){
		expr->type_ = typebool();
		return;
	}
	if(shouldcompletetype(tl, tr)){
		expr->right_->type_ = expr->left_->type_;
		expr->type_ = typebool();
		return;
	}
Bad:
	error(expr->line_, "bad types %T and %T in %#E", tl, tr, expr);
}

void
ExprEqEq::typecheck(Typecheck *tc, int kind)
{
	typecheckcmp(tc, this, kind);
}

void
ExprNotEq::typecheck(Typecheck *tc, int kind)
{
	typecheckcmp(tc, this, kind);
}

void
ExprEq::typecheck(Typecheck *tc, int kind)
{
	noassign(tc, this, kind);
	typecheckexpr(tc, left_, LValue);
	right_->infertype_ = left_->type_;
	typecheckexpr(tc, right_, Value);
	Type *tl = left_->type_;
	Type *tr = right_->type_;
	if(tl == nil || tr == nil)
		return;
	if(issubtype(tl, tr)){
		type_ = left_->type_;
		return;
	}else if(shouldcompletetype(tl, tr)){
		right_->type_ = left_->type_;
		type_ = left_->type_;
		return;
	}else if(tl->op == TypeTuple && tr->op == TypeTuple){
		List<Type*> *ll, *lr;
		for(ll=tl->types, lr=tr->types; ll && lr; ll=ll->next, lr=lr->next)
			if(ll->x == nil || lr->x == nil || (ll->x != lr->x && ll->x != typedontcare()))
				goto Bad;
		if(ll || lr)
			goto Bad;
		type_ = right_->type_;
		return;
	}
Bad:
	error(line_, "bad types %T and %T in %#E", tl, tr, this);
}

static void
typecheckjoin(ExprBinary *expr)
{
	if(subtypejoin(expr->left_->type_, expr->right_->type_) != nil)
		return;
	error(expr->line_, "cannot join types %T and %T in %#E",
		expr->left_->type_, expr->right_->type_, expr);
}

static int
canconverttobool(Type *t)
{
	if(t == nil)
		return 0;
	switch(t->op){
	case TypeBool:
	case TypeInt:
	case TypeFloat:
	case TypeString:
	case TypePtr:
	case TypeList:
	case TypeFn:
	case TypeMap:
		return 1;
	default:
		return 0;
	}	
}

static int
determinecalltype(Typecheck *tc, Expr *left, int kind,
                  Type **tforall, Type **host, List<Type*> **targs, CallOp *callop)
{
	Type *t, *xt;

	// XXX The first two cases seem rather silly in the
	// land of higher-order functions.  The first case
	// should go in case ExprName and should look like a
	// regular function at this point.  Likewise, the
	// second case should go in ExprDot and look like a
	// regular function at this point.  The only catch is
	// that the code generation for each is totally
	// different, which is flagged at this point.
	ExprName *en;
	ExprDot *ed;
	if((en = left->exprname()) && (t = lookuptype(tc->scope_, en->name_, 0))){
		// Constructor: Foo()
		if(checkunresolved(left->line_, t))
			return 0;
		if(t->op == TypeForall){
			*tforall = t;
			t = t->left;
		}
		if(t->op != TypeStruct){
			error(left->line_, "%s has non-struct type %T", en->name_, t);
			return 0;
		}
		if(t->discriminated){
			error(left->line_, "no default constructor for discriminated struct %s", t->name);
			return 0;
		}
		*targs = t->ctorargs;
		*host = t;
		*callop = CallCtor;
	}
	else if((ed = left->exprdot()) &&
		(en = ed->left_->exprname()) &&
	        (t = lookuptype(tc->scope_, en->name_, 0)))
	{
		// Constructor: Foo.Bar()
		if(checkunresolved(left->line_, t))
			return 0;
		if(t->op == TypeForall){
			if(tforall)
				*tforall = t;
			t = t->left;
		}
		if(t->op != TypeStruct){
			error(left->line_, "%s has non-struct type %T", en->name_, t);
			return 0;
		}
		if(!t->discriminated){
			error(left->line_, "non-discriminated struct %T", t);
			return 0;
		}
		if((xt = lookuptag(t, ed->name_, tc->scope_)) == nil){
			error(left->line_, "no tag %s in struct %T", ed->name_, t);
			return 0;
		}
		*targs = xt->ctorargs;
		*host = xt;
		*callop = CallCtor;
	}
	else{
		// Fncall: foo()
		typecheckexpr(tc, left, Value);
		t = left->type_;
		if(t == nil)
			return 0;
		if(t->op == TypeForall){
			if(tforall)
				*tforall = t;
			t = t->left;
		}
		if(t->op != TypeFn){
			error(left->line_, "call non-function %#E type [%T]", left, left->type_);
			return 0;
		}
		*targs = t->types;
		*host = t->left;
		*callop = CallRegular;
	}
	return 1;
}

static int
typecheckfnarg(Typecheck *tc, Expr **arg, Type *expected, Type *tforall, List<Type*> *subst)
{
	Expr *e;
	Type *t;

	e = *arg;
	t = expected;
	e->infertype_ = t;
	typecheckexpr(tc, e, Value);

	if(e->type_ == t || issubtype(t, e->type_))
		return 1;
	if(shouldcompletetype(t, e->type_)){
		e->type_ = t;
		return 1;
	}
/*
	// Try again with a list() conversion.
	if(t->op == TypeList && e->type_ && e->type_->op != TypeList && e->op != ExprConvert){
		e = new ExprConvert(e, typelist(nil));
		e->line = e->left->line;
		*arg = e;
		return typecheckfnarg(tc, arg, expected, tforall, subst);
	}
*/
	// Try unification, if we're polymorphic.
	if(subst){
		if(!canunifytypes(e->line_, e->type_, t, tforall->typevars, subst)){
			error(e->line_, "unification failed for %#E: have [%T], want [%T]", e, e->type_, t);
			return 0;
		}
		return 1;
	}

	error(e->line_, "%#E is [%T], want [%T]", e, e->type_, t);
	return 0;
}

static int
checktypesubstsizes(Line line, List<Type*> *types)
{
	static int ptrsz;
	List<Type*> *l;
	int sz;

	// All types substituted for a TypeVar must have ptr size.
	if(!ptrsz)
		ptrsz = sizeoftype(typeptr(nil));
	for(l=types; l; l=l->next){
		if(!l->x)
			continue;
		sz = sizeoftype(l->x);
		if(sz != ptrsz){
			error(line, "cannot instantiate [%T], size is %d (wanted %d)\n", l->x, sz, ptrsz);
			return 0;
		}
	}
	return 1;
}

void
ExprCall::typecheck(Typecheck *tc, int kind)
{
	List<Type*> *lt, *targs, *subst;
	Type *t, *host, *tret, *tforall;
	Expr *e, *left;

	noassign(tc, this, kind);
	left = left_;
	host = nil;
	tforall = nil;
	targs = ZList;
	subst = ZList;
	if(!determinecalltype(tc, left_, kind, &tforall, &host, &targs, &callop_))
		return;
	if(tforall)
		tret = tforall;
	else
		tret = host;

	// If this fn is polymorphic, create an empty substitution set.
	if(tforall){
		int i;
		for(i=0; i < listlen(tforall->typevars); i++)
			subst = listinsert((Arena*)nil, subst, (Type*)nil);
	}

	// Allow empty struct constructors (e.g., Foo())
	if(callop_ != CallRegular && !exprs_.size()){
		if(infertype_ && tforall){
			tret = typesubst(tforall->typevars, tforall->typevars, tforall);
			canunifytypes(line_, infertype_, tret, tforall->typevars, subst);
			tret = typesubst(subst, tforall->typevars, tforall);
		}
		type_ = tret;
		return;
	}

	// Check args
	int i;
	for(i=0, lt=targs; i<exprs_.size() && lt; i++, lt=lt->next)
		typecheckfnarg(tc, &exprs_[i], lt->x, tforall, subst);

	if(i<exprs_.size()){
		error(line_, "too many call arguments (at %#E), expected %d",
		                  exprs_[i], listlen(targs));
		return;
	}
	if(lt){
		error(line_, "not enough call arguments (want %T)", lt->x);
		return;
	}

	ExprName *en = left->exprname();
	Name name = en ? en->name_ : Name::null;

	if(subst){
		if(callop_ == CallCtor){
			tret = typesubst(subst, tforall->typevars, tforall);
		}
		else{
			tret = applytypesubst(host, tforall->typevars, subst);
			if(!name || strcmp(name, "listmap") != 0)
				checktypesubstsizes(line_, subst);
		}
	}

	// Super-ugly special cases (would go away with bounded quantification)
	if(name && exprs_.size() &&
	        (strcmp(name, "astcopy") == 0
	      || strcmp(name, "astdeepcopy") == 0
	      || strcmp(name, "astjoin") == 0))
	{
		e = exprs_[0];   // return type same as first arg type
		type_ = e->type_;
	}
	else
		type_ = tret;

	// Mark listmap's in/out types (needed by compile)
	if(name && strcmp(name, "listmap") == 0 && exprs_.size()){
		e = exprs_[0];
		t = e->type_;
		if(t && t->op == TypeList && tret && tret->op == TypeList){
			if(subst)
				t = applytypesubst(t, tforall->typevars, subst);
			xtype_ = typetuple(list2(asyntax, t->left, tret->left));
		}
	}
}

void
ExprInit::typecheck(Typecheck *tc, int kind)
{
	List<Type*> *largs, *subst;
	Type *xt, *host, *tforall;
	Expr *left;
	Var *v;

	noassign(tc, this, kind);
	nosideeffect(tc, this, kind);
	left = left_;
	host = nil;
	tforall = nil;
	subst = ZList;
	if(!determinecalltype(tc, left_, kind, &tforall, &host, &largs, &callop_))
		return;
	if(callop_ != CallCtor || !host || host->op != TypeStruct){
		error(line_, "non-struct type [%T]", host);
		return;
	}

	// If this fn is polymorphic, create an empty substitution set.
	if(tforall){
		int i;
		for(i=0; i < listlen(tforall->typevars); i++)
			subst = listinsert((Arena*)nil, subst, (Type*)nil);
	}

	// Check the field initializers
	for(int i=0; i<exprs_.size(); i++){
		ExprInitField *e = exprs_[i];
		if((v = lookupstruct(host, e->name_, tc->scope_, &xt)) == nil){
			error(e->line_, "%T has no member %s", host, e->name_);
			continue;
		}
		if(!typecheckfnarg(tc, &e->left_, v->type, tforall, subst))
			continue;
		e->var_ = v;
		e->type_ = v->type;
		e->xtype_ = xt;
	}

	if(tforall){
		checktypesubstsizes(line_, subst);
		type_ = typesubst(subst, tforall->typevars, tforall);
	}
	else
		type_ = host;
}

int
checkunresolved(Line l, Type *t)
{
	Type *u;
	if (isunresolved(t, &u)) {
		error(l, "unresolved type%s%s in %T", u ? " " : "", u ? (char*)u->name : "", t); 
		return 1;
	}
	return 0;
}

void
pushscope(Typecheck *tc)
{
	tc->scope_->start = tc->frame_->vars;
	tc->scope_ = mkScope(tc->frame_, tc->scope_, tc->frame_->vars);
	if(tc->scope_->outer)
		tc->scope_->canonicals = tc->scope_->outer->canonicals;
}

void
popscope(Typecheck *tc)
{	
	assert(tc->scope_);
	tc->scope_->start = tc->frame_->vars;   // done with this scope
	tc->frame_->vars = tc->scope_->end;     // restore
	tc->scope_ = tc->scope_->outer;
	tc->scope_->start = nil;               // adding to this scope again
}

void
pushbreak(Typecheck *tc, Label *l)
{
	tc->breakstack_.push_back(l);
}

void
popbreak(Typecheck *tc, Label *l)
{
	if(l != tc->breakstack_.back())
		panic("popbreak");
	tc->breakstack_.pop_back();
}

void
pushcontinue(Typecheck *tc, Label *l)
{
	tc->continuestack_.push_back(l);
}

void
popcontinue(Typecheck *tc, Label *l)
{
	if(l != tc->continuestack_.back())
		panic("popcontinue");
	tc->continuestack_.pop_back();
}

Label*
mklabel(Typecheck *tc)
{
	Label *l;
	
	l = (Label*)emalloc(sizeof *l);
	l->next = (Label*)&tc->labels_;
	l->prev = l->next->prev;
	l->next->prev = l;
	l->prev->next = l;
	return l;
}

Label*
getlabel(Typecheck *tc, Name name)
{
	Label *l;
	
	for(l=tc->labels_.next; l != (Label*)&tc->labels_; l=l->next)
		if(l->name == name)
			return l;
	l = mklabel(tc);
	l->name = name;
	return l;
}

Label*
uselabel(Typecheck *tc, Name name, Stmt *stmt)
{
	Label *l;
	
	l = getlabel(tc, name);
	if(l->used == nil)
		l->used = stmt;
	return l;
}

Label*
setlabel(Typecheck *tc, Name name, Stmt *stmt)
{
	Label *l;
	
	l = getlabel(tc, name);
	if(l->stmt)
		error(stmt->line_, "already have label %s at %L",
			name, l->stmt->line_);
	else
		l->stmt = stmt;
	return l;
}

void
checklabels(Typecheck *tc)
{
	Label *l;

	for(l=tc->labels_.next; l != (Label*)&tc->labels_; l=l->next)
		if(l->stmt == nil && l->name)
			error(l->used ? l->used->line_ : ZL, "label %s not defined [%p]", l->name, l);
}

typedef struct TypeRefinement TypeRefinement;
struct TypeRefinement
{
	Var *var;
	Type *oldtype;
};

void
pushflow(Typecheck *tc)
{
	// Mark the beginning of a control flow branch
	// (very rough approximation of control flow, just enough for if/while)
	tc->typerefinestack_ = listinsert(&tc->arena_, tc->typerefinestack_, tc->typerefines_);
}

void
popflow(Typecheck *tc)
{	
	TypeRefinement *r;
	List<TypeRefinement*> *l, *end;

	// Mark the end of a control flow branch
	end = listpeek(tc->typerefinestack_);
	for(l=tc->typerefines_; l != end; l=l->next){
		r = l->x;
		r->var->type = r->oldtype;
	}
	tc->typerefines_ = end;
	tc->typerefinestack_ = listpop(tc->typerefinestack_);
}

void
addtyperefinement(Typecheck *tc, Expr *e, Type *t)
{
	TypeRefinement *r;

	if(e == nil)
		return;
	ExprName *en = e->exprname();
	if(en == nil || en->var_ == nil)
		return;
	assert(issubtype(en->var_->type, t));
	if(!tc->incondition_)
		return;

	// Mark this variable with the refined type.
	// This will be undone later by poptyperefines().
	r = (TypeRefinement*)amalloc(&tc->arena_, sizeof *r);
	r->var = en->var_;
	r->oldtype = en->var_->type;
	tc->typerefines_ = listinsert(&tc->arena_, tc->typerefines_, r);
	en->var_->type = t;
}

void
printframe(Frame *frame, int n)
{
	Frame *f;
	List<Var*> *l;
	Var *v;
	char *sep;
	
	fprint(2, "{");
	sep = "";
	for(f=frame; f && n; f=f->outer, n--){
		for(l=f->vars; l; l=l->next){
			v = l->x;
			fprint(2, "%s%s:%T", sep, v->name, v->type);
			sep = ", ";
		}
	}
	fprint(2, "}\n");
}

static void
printlocalscope(Scope *scope, int typesonly)
{
	List<Var*> *l;
	Var *v;
	char *sep;
	
	l = scope->start ? scope->start : scope->frame->vars;
	sep = "";
	for(; l && l != scope->end; l=l->next){
		v = l->x;
		if(typesonly && !v->istype)
			continue;
		fprint(2, "%s%s:%T", sep, v->name, v->type);
		sep = ",\n  ";
	}
}

void
printscope(Scope *scope, int localonly, int typesonly)
{
	Scope *s;
	List<Scope*> *l;
	
	fprint(2, "{\n  ");
	for(s=scope; s; s=s->outer){
		printlocalscope(s, typesonly);
		if(localonly)
			break;
		for(l=s->imported; l; l=l->next)
			printscope(l->x, 0, typesonly);
	}
	fprint(2, "}\n");
}

static Var*
_lookuplocal(Frame *frame, Scope *scope, Name name, int looktype)
{
	List<Var*> *l;
	Var *v;
	Frame *f;

	assert(scope);
	// Lookup locally in 'scope' (don't check outer scopes).
	l = scope->start ? scope->start : scope->frame->vars;
	for(; l && l != scope->end; l=l->next){
		v = l->x;
		if(v->name != name)
			continue;
		if(looktype && v->istype)
			return v;
		if(looktype)
			continue;
		if(v->f != frame && v->f != globalframe){
			for(f=frame; f != v->f; f=f->outer)
				escapef(f);
			v = escapev(v);
		}
		return v;
	}
	return nil;
}

static Var*
_lookup(Scope *scope, Name name, int looklocal, int looktype)
{
	Scope *s;
	List<Scope*> *l;
	Var *v;

	assert(scope);

	if(looklocal)
		return _lookuplocal(scope->frame, scope, name, looktype);

	// Lookup in this scope, all imported scopes, and all outer scopes.
	for(s=scope; s; s=s->outer){
		if((v = _lookuplocal(scope->frame, s, name, looktype)) != nil)
			return v;
		for(l=s->imported; l; l=l->next)
			if((v = _lookup(l->x, name, 0, looktype)) != nil)
				return v;
	}
	return nil;
}

Type*
lookuptype(Scope *scope, Name name, int looklocal)
{
	Var *v;

	if((v = _lookup(scope, name, looklocal, 1)) != nil)
		return v->type;
	return nil;
}

Var*
lookup(Scope *scope, Name name, int looklocal)
{
	return _lookup(scope, name, looklocal, 0);
}

typedef struct ResolveState ResolveState;
struct ResolveState
{
	Typecheck *tc;
	Line line;
};

static Type*
resolvenamedtype1(ResolveState *rs, Type *t)
{
	Typecheck *tc = rs->tc;
	Type *x, *out;
	Var *v;
	List<Type*> *l;

	if(t == nil)
		return nil;
	out = maptypechildren(rs, t, resolvenamedtype1);
	if(out)
		t = out;

	if(t->op == TypeName){
		// Resolve the base name.  Check our actual scope,
		// and then any typevars in scope (kludge for decls).
		v = _lookup(tc->scope_, t->name, 0, 1);
		if(v != nil)
			out = v->type;
		else{
			out = nil;
			for(l=tc->decltypevars_; l; l=l->next){
				x = l->x;
				if(x->name == t->name){
					out = x;
					break;
				}
			}
			if(!out){
				out = typeunresolved(tc->scope_, t->name);
				declaretype(tc, ZL, t->name, &out);
			}
		}
		assert(out);
		// Convert this to a TypeSubst if it was name<types>
		if(out->op != TypeUnresolved && t->types){
			if(out->op != TypeForall){
				error(rs->line, "cannot instantiate non-polymorphic type [%T]", out);
				return out;
			}
			int have = listlen(t->types);
			int want = listlen(out->typevars);
			if(have != want){
				error(rs->line, "wrong number of types in instantiation [%T]; have %d, want %d",
				                out, have, want);
				return out;
			}
			if(checktypesubstsizes(rs->line, t->types))
				out = typesubst(t->types, out->typevars, out);
		}
		return out;
	}
	else if(t->op == TypeSubname && isptrast(typeptr(t->super))){
		// This could be (G+?).foo, where 'foo' is a symbol in one
		// of G's extensions.
		out = resolveparsetype(tc, t->super);
		if(out){
			out = lookuptag(out, t->name, tc->scope_);
			if(out)
				return out;
		}
	}
	return t;
}

// Resolve named types.  This takes a type that may contain unresolved
// named types and returns a type without named types.  This should be
// called on any types that come directly from parsed user code before
// they appear in the internal representation.
//
// If a name is encountered that can not be resolved (ie, it has not
// been defined yet), it is declared as a TypeUnresolved and this
// unresolved type replaces the named type in the type being
// resolved.  When this type is later declared, the presence of a
// TypeUnresolved binding for the type name will cause the declaration
// to replace the unresolved type with its actual type.  Because this
// is performed in-place, existing pointers to this type do not have
// to be updated.
//
// If no changes need to be made to the input type, it is simply
// returned.
Type*
resolvenamedtype(Typecheck *tc, Line line, Type *t)
{
	ResolveState s;
	Type *out;

	s.tc = tc;
	s.line = line;
	out = resolvenamedtype1(&s, t);
	if(out)
		return out;
	return t;
}

// Declare a variable.
Var*
declare(Typecheck *tc, Line line, Name name, Type *type)
{
	Var *v;
	Frame *f;

	if(isabstracttype(type)){
		error(line, "cannot declare %s: type %T is abstract", name, type);
		return nil;
	}
	f = tc->frame_;
	assert(f);
	if(f->tmpoff != f->off)
		panic("declare");
	v = mkVar(name, type);
	v->off = f->off;
	v->f = f;
	f->off += sizeoftype(type);
	if(f->off > f->size)
		f->size = f->off;
	f->tmpoff = f->off;
	f->vars = listinsert(nil, f->vars, v);
	f->allvars = listinsert(nil, f->allvars, v);
	return v;
}

// Declare a variable which must be unique in the current scope.
Var*
declareunique(Typecheck *tc, Line line, Name name, Type *type)
{
	if(lookup(tc->scope_, name, 1) != nil){
		error(line, "duplicate definition of '%s' is illegal", name);
		return nil;
	}
	return declare(tc, line, name, type);
}

// Declare a named type.
Var*
declaretype(Typecheck *tc, Line line, Name name, Type **type)
{
	Type *existing;
	Frame *f;
	Var *v;

	v = nil;
	assert(type);
	// Do not allow type redeclarations in the current scope.
	if(tc->scope_ && (v = _lookup(tc->scope_, name, 1, 1)) != nil){
		existing = v->type;
		if(existing == *type)
			return v;
		if(existing->op != TypeUnresolved){
			error(line, "%s already defined in this scope as type %T", name, existing);
			return nil;
		}
	}else{
		existing = lookuptype(tc->scope_, name, 0);
	}
	if(existing && existing->op == TypeUnresolved &&
	   existing->declscope == (*type)->declscope){
		// We have resolved the type by this name in this
		// scope.  We have to update it in place so pointers
		// to the unresolved type will now point to the resolved
		// type.  In order to maintain memoization, we also
		// have to update the pointer to the type being declared
		// so it points to the previously unresolved type.
		*existing = **type;
		*type = existing;
		if(v)
			return v;
	}
	
	// This type declaration is new to this scope.
	f = tc->frame_;
	v = mkVar(name, *type);
	v->istype = 1;
	f->vars = listinsert(nil, f->vars, v);
	f->allvars = listinsert(nil, f->allvars, v);
	return v;
}

void
escapef(Frame *f)
{
	List<Var*> *l;
	Var *vv;

	// Add slot for EP in local frame if needed.
	// Code.c knows that EP = 0(FP).
	if(f->escsize == 0){
		for(l=f->allvars; l; l=l->next){
			vv = l->x;
			if(!vv->escapes && vv->off >= 0)
				vv->off += sizeof(void*);
		}
		f->size += sizeof(void*);
		f->off += sizeof(void*);
		f->tmpoff += sizeof(void*);
		f->escsize = sizeof(void*);	// for enclosing EP
	}
}

Var*
escapev(Var *v)
{
	List<Var*> *l;
	int n;
	Var *vv;

	if(v->escapes)
		return v;
	v->escapes = 1;
	
	// Remove v from frame, sliding other vars down
	// if v is not a fn arg.
	n = sizeoftype(v->type);
	if(v->off >= 0){
		for(l=v->f->allvars; l; l=l->next){
			vv = l->x;
			if(!vv->escapes && vv->off > v->off)
				vv->off -= n;
		}
		v->f->size -= n;
		v->argoff = 1;
	}else
		v->argoff = v->off;
	
	escapef(v->f);

	// Add v to esc frame.
	v->off = v->f->escsize;
	v->f->escsize += sizeoftype(v->type);

	return v;
}	

static void
markvarsconst(Expr *expr)
{
	if(expr == nil)
		return;
	ExprName *en = expr->exprname();
	ExprTuple *et = expr->exprtuple();
	if(en){
		if(en->var_)
			en->var_->isconst = 1;
	}else if(et){
		for(int i=0; i<et->exprs_.size(); i++)
			markvarsconst(et->exprs_[i]);
	}
}

// Declare expr as type.
int
declareexpr(Typecheck *tc, Expr *expr, Type *type)
{
	List<Type*> *lt;

	if(expr == nil)
		panic("declareexpr");
	if(type == nil){
		error(expr->line_, "no type for decl");
		return 0;
	}
	ExprName *en = expr->exprname();
	ExprTuple *et = expr->exprtuple();
	if(en){
		if(strcmp(en->name_, "_") != 0 && !declare(tc, en->line_, en->name_, type))
			return 0;
	}else if(et){
		if(type->op != TypeTuple){
			error(expr->line_, "cannot declare %#E as %T", expr, type);
			return 0;
		}
		if(et->exprs_.size() != listlen(type->types)){
			error(expr->line_, "cannot declare %#E as %T [count mismatch]", expr, type);
			return 0;
		}
		int i;
		for(i=0, lt=type->types; i<et->exprs_.size() && lt; i++, lt=lt->next)
			if(!declareexpr(tc, et->exprs_[i], lt->x))
				return 0;
	}else{
		error(expr->line_, "cannot declare %#E", expr);
		return 0;
	}
	return 1;
}

Fn*
typecheckfn(Typecheck *tc, Line line, Fn *fn, Var *fnvar, List<Name> *ntypevars, int old)
{
	Frame *f;
	Fnarg *a;
	List<Type*> *subst;
	Var *v;

	pushscope(tc);
	subst = ZList;
	List<Type*> *typevars = names2typevars(ntypevars);
	for(List<Type*> *l=typevars; l; l=l->next){
		declaretype(tc, line, l->x->name, &l->x);
		subst = listinsert((Arena*)nil, subst, (Type*)nil);
	}

	for(List<Fnarg*> *l=fn->fnargs; l; l=l->next){
		a = l->x;
		a->type = resolvenamedtype(tc, line, a->type);
	}
	fn->rettype = resolvenamedtype(tc, line, fn->rettype);
	fn->type = typefnargs(fn->fnargs, fn->rettype);
	if(typevars)
		fn->type = typeforall(typevars, fn->type);

	if(fnvar){
		// FIXME: not exactly right!
		if(old && !canunifytypes(line, fnvar->type, fn->type, typevars, subst))
			error(line, "incompatible redeclaration of %s: was [%T], now [%T]",
			            fnvar->name, fnvar->type, fn->type);
		fnvar->type = fn->type;
	}
	if(checkunresolved(line, fn->type)){
		popscope(tc);
		return nil;
	}

	f = mkFrame();
	f->outer = tc->frame_;
	f->depth = tc->frame_->depth + 1;
	assert(fn->cs == nil);
	fn->tc = new Typecheck(f, mkScope(f, tc->scope_, nil));
	fn->tc->scope_->canonicals = tc->scope_->canonicals;  // XXX is this all we need to copy?
	fn->frame = f;
	fn->outer = tc->fn_;
	fn->tc->fn_ = fn;
	fn->tc->reachable_ = 1;

	// Declare return slot to take up space in frame.
	declare(fn->tc, line, nameof(".ret"), typeptr(fn->rettype));

	for(List<Fnarg*> *l=fn->fnargs; l; l=l->next){
		a = l->x;
		declare(fn->tc, line, a->name, a->type);
	}
	f->argsize = f->size;
	for(List<Var*> *l=f->allvars; l; l=l->next){
		v = l->x;
		v->off = -v->off - sizeoftype(v->type);
	}
	f->size = FrameReserved;
	f->off = f->size;
	f->tmpoff = f->off;

	for(List<Stmt*> *l=fn->stmts; l; l=l->next)
		typecheckstmt(fn->tc, l->x);
	checklabels(fn->tc);
	popscope(tc);
	return fn;
}

static Var*
mknewstructvar(Typecheck *tc, Line line, Name name, Type *type, List<Type*> *lt)
{
	List<Type*> *l;

	// Attempt to add a field 'name' to a struct.
	type = resolvenamedtype(tc, line, type);
	if(checkunresolved(line, type))
		return nil;
	if(isabstracttype(type)){
		error(line, "field '%s' has abstract type %T", name, type);
		return nil;
	}
	// Duplicate name checking
	// Should not clash with any struct fields from structs in 'lt'.
	for(l=lt; l; l=l->next){
		// FIXME Should probably also check subtype fields of l->x
		if(lookupstruct(l->x, name, tc->scope_, nil) != nil){
			error(line, "field '%s' already defined in type %T", name, l->x);
			return nil;
		}
	}
	return mkVar(name, type);
}

// FIXME: extending Asts
// FIXME: this is a mess!
Type*
typecheckstruct(Typecheck *tc, Line line, Name name, List<Type*> *typevars,
                List<Structelem*> *decls, Type *super, Type *extends, int isext, int istagged)
{
	List<Type*> *lt, *l, *tocheck;
	List<Var*> *lv;
	Type *t, *base, *fulltype;
	List<Structelem*> *tagelems;
	Structelem *e;
	Name fieldname;
	Name tagname;
	int off;
	Var *v;

	assert(!extends || !super);

	// Declare non-extension structs unresolved, before checking the body.
	// This will maintain type canonicalization for recursive structs.
	if(!super && !extends){
		t = typeunresolved(tc->scope_, name);
		fulltype = t;
		if(typevars){
			fulltype->op = TypeForall;
			fulltype->typevars = typevars;
			fulltype->left = mkTypeStruct(tc->scope_, name, super);
			t = fulltype->left;
		}
	}
	// Declare tag structs as a TypeSubnype.
	// This will automatically resolve nested type names like Sym.Var.
	else if(!extends){
		t = typesubname(rootstruct(super), name);
		fulltype = t;
	}
	// Extension structures are not resolvable.
	else{
		t = mkTypeStruct(tc->scope_, name, super);
		fulltype = t;
	}

	t->declscope = tc->scope_;
	t->tagshash  = mkhash<Name,List<Type*>*>();
	if(!extends){
		t->op = TypeStruct;
		t->super = super;
	}
	assert(t->op == TypeStruct);

	lt = ZList;
	lv = ZList;
	tagelems = ZList;
	// Compute the initial offset.  Assume that builtin structs
	// do not use this function, and therefore t->reserved is 0.
	if(extends)
		off = offsetof(ExtStruct, v);
	else if(super)
		off = sizeoftype(super);
	else{
		off = 0;
		if(istagged)
			off += sizeof(void*);
		if(isext)
			off += sizeof(void*);
	}

	// Types we should search for name clashes.
	tocheck = ZList;
	if(super)
		tocheck = listinsert(asyntax, tocheck, super);
	if(extends)
		tocheck = listinsert(asyntax, tocheck, extends);
	// Typecheck the base fields and gather tag decls
	for(List<Structelem*> *ll=decls; ll; ll=ll->next){
		e = ll->x;
		switch(e->op){
		case StructelemDecl:
			// Duplicate name checking (other fields in this new struct).
			fieldname = e->decl->name;
			for(List<Var*> *l=lv; l; l=l->next){
				v = l->x;
				if(v->name == fieldname)
					goto Dup;
			}
			// Try to make this variable.
			v = mknewstructvar(tc, e->line, fieldname, e->decl->type, tocheck);
			if(v == nil)
				goto Bad;
			v->off = off;
			v->host = t;
			off += sizeoftype(v->type);
			lt = listappend(nil, lt, v->type);
			lv = listappend(nil, lv, v);
			break;
		case StructelemTag:
			tagelems = listappend(nil, tagelems, e);
			break;
		}
	}

	// Finish constructing this type
	if(tagelems && !istagged){
		error(line, "struct %s has tags but is not discriminated", name);
		goto Bad;
	}
	t->types = lt;
	t->vars  = lv;
	t->ctorargs = lt;
	t->extensible = isext;
	if(!super && !extends && istagged)
		t->discriminated = 1;
	// Prepend the super field types to the constructor type
	if(super){
		l = listcopy(nil, super->ctorargs);
		t->ctorargs = listconcat(l, lt);
	}
	// Link extensions with their parent struct
	if(extends){
		t->extends = extends;
		extends->extensions = listappend(nil, extends->extensions, t);
		t = extends;
	}
	// Link tags with the base struct
	if(isverticalsubtype(typeast()->structrep, t)){
		if(super)
			base = lookuptagowner(t);
		else
			base = t;
	}
	else{
		base = rootstruct(t);
		if(super)
			addstructtag(base, t);
	}

	// Construct tag types
	for(List<Structelem*> *ll=tagelems; ll; ll=ll->next){
		e = ll->x;
		if((tagname = e->tagname) == nil)
			continue;
		fieldname = tagname;
		extends = nil;
		super = nil;
		// Duplicate name checking
		Type *tag = lookuptag(base, tagname, tc->scope_);
		if(tag){
			if(isext)
				extends = tag;
			else
				goto Dup;
		}
		// Subtags of Ast cannot be new
		if(isverticalsubtype(typeast()->structrep, t) && !extends){
			error(e->line, "syntax element '%T.%s' does not exist", t, tagname);
			goto Bad;
		}
		// Check the super tag (if any)
		if(extends && e->supername){
			error(e->line, "illegal subtyping; tag %s already declared", tagname);
			goto Bad;
		}
		if(e->supername){
			Name supername;
			if((supername = e->supername) == nil)
				continue;
			super = lookuptag(base, supername, tc->scope_);
			if(!super){
				error(e->line, "undeclared tag %s", supername);
				goto Bad;
			}
		}
		else if(!extends)
			super = t;
		// Rename extension tags
		if(extends)
			tagname = nameprint("%s.X%d", tagname, listlen(extends->extensions));
		// Construct this tag type
		tag = typecheckstruct(tc, e->line, tagname, ZList, e->tagdecls, super,
		                      extends, isext, 1);
	}
	return fulltype;

Dup:
	error(e->line, "struct element %s already defined", fieldname);
Bad:
	listfree(lt);
	listfree(lv);
	return nil;
}

// Attribute checks

// Construct the function type.
static Type*
attrfntype(Typecheck *tc, StmtDeclareAttribute *stmt)
{
	assert(stmt->names_);
	Name n = stmt->names_->x;
	Type *host = lookuptype(tc->scope_, n, 0);
	if(!host){
		error(stmt->line_, "no type '%s'\n", n);
		return nil;
	}
	if(!issubtypesyntax(gramwildtype(host->grammar), host)){
		error(stmt->line_, "not a grammar family: [%T]", host);
		return nil;
	}
	for(List<Name> *l=stmt->names_->next; l; l=l->next){
		Type *t;
		if((t = lookuptag(host, l->x, tc->scope_)) == nil){
			error(stmt->line_, "no type %T.%s\n", host, l->x);
			return nil;
		}
		host = t;
	}
	return typefn(list1(nil, typeptr(host)), stmt->type_);
}

// Check the function parameters: allow (G?), (G?.sym), and (G?.sym.rule)
// It should be G? and not G because this attribute holds for the entire
// grammar family; the function should handle all syntax in that family.
static Type*
attrfntype(Typecheck *tc, StmtFndef *stmt)
{
	assert(stmt->flags_ & FnAttribute);
	return typefnargs(stmt->fn_->fnargs, stmt->fn_->rettype);
}

static int
attrtypesplit(Typecheck *tc, Line line, Type *tfn, Type **phost, Type **ptret)
{
	if(tfn == nil)
		return -1;
	assert(tfn->op == TypeFn);
	tfn = resolvenamedtype(tc, line, tfn);
	if(checkunresolved(line, tfn))
		return -1;
	if(listlen(tfn->types) != 1){
	Badargs:
		error(line, "bad arguments for attribute function, expected (ptr G?) -> _,"
		                  " found %T", tfn);
		return -1;
	}
	Type *arg = listpeek(tfn->types);
	assert(arg);
	if(!isptrast(arg))
		goto Badargs;
	arg = arg->left;
	Type *host = arg;
	if(!arg->grammar)
		goto Badargs;
	for(; arg->super; arg=arg->super){
		if(isreservedsymname(arg->name)){
			error(line, "cannot attach attributes to '%s'", arg->name);
			return -1;
		}
	}
	if(!issubtypesyntax(gramwildtype(arg->grammar), arg))
		goto Badargs;
	if(ptret)
		*ptret = tfn->left;
	if(phost)
		*phost = host->structrep;
	return 0;
}

static Var*
declareattribute(Typecheck *tc, Line line, Name name, Type *tfn)
{
	Type *host, *tret;	
	if(attrtypesplit(tc, line, tfn, &host, &tret) < 0)
		return nil;

	if(isabstracttype(tret)){
		error(line, "attribute '%s' has abstract type %T", name, tret);
		return nil;
	}

	// Check for (and allow) redeclaration.
	Var *attr;
	if((attr = lookupattr(host, name, tc->scope_, 0)) != nil){
		if(attr->type != tret){
			error(line, "bad type for attribute '%s' - found [%T], expected [%T]",
			                  name, attr->type, tret);
			return nil;
		}
		if(attr->isbuiltinattr){
			error(line, "cannot override builtin attribute '%s'", name);
			return nil;
		}
		assert(attr->host == host);
		assert(attr->attrfn);
		return attr;
	}
	
	// Duplicate name checking
	// FIXME Should probably also check subtype fields of host
	if(lookupstruct(host, name, tc->scope_, nil) != nil){
		error(line, "attribute '%s' already defined in type %T", name, host);
		return nil;
	}

	// Create new attribute.
	attr = addstructattr(name, tret, host, 0);
	attr->declscope = tc->scope_;
	attr->mangled = nameprint("attr@%T.%s", host, name);
	return attr;
}

static Var*
lookupattribute(Typecheck *tc, Line line, Name name, Type *tfn)
{
	Type *host, *tret;
	if(attrtypesplit(tc, line, tfn, &host, &tret) < 0)
		return nil;

	// The attribute must exist on this host with the same type.
	Var *attr = lookupattr(host, name, tc->scope_, 0);
	if(attr == nil){
		if(lookupattr(host, name, tc->scope_, 1))
			error(line, "cannot extend attribute '%s' from [%T]", name, host);
		else
			error(line, "no attribute '%s' to extend on [%T]", name, host);
		return nil;
	}
	if(attr->isbuiltinattr){
		error(line, "cannot extend builtin attribute '%s'", name);
		return nil;
	}
	if(attr->type != tret){
		error(line, "bad type for attribute '%s', found [%T] expecting [%T]",
			attr->name, tret, attr->type);
		return nil;
	}
	return attr;
}
